using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.InteropServices;
using c_tt;

namespace BearWare
{
    /**
     * @defgroup initclient Client Initialization
     *
     * @brief This section explains how to instantiate a new client
     * instance and query its current state.
     *
     * The TeamTalk4.TeamTalk4() constructor has the parameter @a poll_events
     * which should be 'false' for Windows Forms applications and 'true'
     * for console applications. The @a poll_events parameter tells how the 
     * client instance should post its events. Since a Windows Forms 
     * application has its own message-loop it does not need to "poll" for 
     * events. This is, however, needed by console applications which should
     * call TeamTalk4.GetMessage() to get events.
     *
     * When a new client instance is created a call to TeamTalk4.GetFlags() will
     * show that the client initially is in state #ClientFlag ::CLIENT_CLOSED. This
     * means that no operation has been performed on the client.
     *
     * Querying TeamTalk4.GetFlags() call can be used by a user application as a
     * state-machine to show which actions are currently possible for the
     * user application. The #ClientFlag enumeration shows the information
     * which can be retrived though TeamTalk4.GetFlags().
     * 
     * 
     * @defgroup sounddevices Sound Capture and Playback
     * @brief This section explains how to record and playback audio
     * in the client instance.
     *
     * Before being able to initialize the sound devices to use for
     * recording and playback the computer's available sound devices must
     * first be queried. This is done using the functions
     * TeamTalk4.GetSoundInputDevices() and TeamTalk4.GetSoundOutputDevices(). These two
     * functions return arrays of #BearWare.SoundDevice-structs which contains a
     * description of the sound device. In the #BearWare.SoundDevice-struct there's
     * a member variable called @a nDeviceID. This ID should be passed to
     * the client instance's two sound initialization functions
     * TeamTalk4.InitSoundInputDevice() and TeamTalk4.InitSoundOutputDevice(). Once this
     * has been done the TeamTalk4.GetFlags() call will return a value containing
     * the mask bits #ClientFlag ::CLIENT_SNDINPUT_READY and #ClientFlag ::CLIENT_SNDOUTPUT_READY.
     *
     * A computer's default sound devices can be queried using
     * TeamTalk4.GetDefaultSoundDevices(). A loop-back test of the selected sound
     * devices can be performed using TeamTalk4.StartSoundLoopbackTest().
     * 
     * Be aware that the sound devices might fail if e.g. a USB sound
     * device is unplugged while the client is talking in a channel. In
     * this case ensure the application is processing the errors
     * #ClientError ::INTERR_SNDINPUT_FAILURE and #ClientError ::INTERR_SNDOUTPUT_FAILURE in the
     * TeamTalk4.OnInternalError() event.
     *
     * Read section @ref transmission to see how to transmit recorded
     * audio to other users.
     *
     *
     * @defgroup videocapture Video Capture and Image Display
     *
     * @brief This section explains how to detect and configure video
     * capture devices.
     *
     * The client is able to capture video and present them to the user
     * application in RGB32-format and transmit the image in encoded
     * format to other users.
     *
     * @section vidcapinit Initialize Video Capture Device
     *
     * To capture video the user application must first query the
     * available capture devices by calling TeamTalk4.GetVideoCaptureDevices(). A
     * #BearWare.VideoCaptureDevice supports a certain number of capture formats
     * each described in the @a captureFormats member of #BearWare.CaptureFormat.
     *
     * Once a device has been chosen the TeamTalk4.InitVideoCaptureDevice() must
     * be called for the client instance to start capturing video
     * frames. Use the @a szDevice member of #BearWare.VideoCaptureDevice as the
     * device identifier for the video capture device and pass a
     * #BearWare.CaptureFormat from the @a captureFormats array of
     * #BearWare.VideoCaptureDevice. The @a lpVideoCodec parameter of
     * TeamTalk4.InitVideoCaptureDevice() can be passed as NULL if the captured
     * video frames do not need to be transmitted. Check out section @ref
     * codecs on how to configure the video codec.
     *
     * @section vidcapdisplay Display Captured Video
     *
     * When a video frame becomes available the event
     * TeamTalk4.OnUserVideoFrame() is posted to the application and
     * TeamTalk4.GetUserVideoFrame() can be used to extract the RGB32 image. On
     * Windows it's also possible to call TeamTalk4.PaintVideoFrame() to make the
     * client instance paint on a @c HWND by getting its HDC.
     *
     *
     * @defgroup codecs Audio and Video Codecs
     *
     * @brief This section explains how to configure audio and video
     * codecs.
     *
     * The client is able to encode audio in Speex and CELT
     * format whereas video can be encoded in <a
     * href="http://www.theora.org">Theora</a> format. Speex is
     * recommended for voice and CELT for music.
     *
     * Choosing the right codec settings in an end-user application is
     * very important and proper settings depend entirely on the user
     * scenario. Always ensure that the codec settings do not require too
     * much bandwidth and thereby resulting in packetloss causing
     * inaudible conversations and poor video quality. Detecting
     * packetloss can be done using #BearWare.UserStatistics and #BearWare.ClientStatistics.
     *
     * Note that bandwidth usage will be much higher when running in peer
     * to peer mode since each client must then broadcast data to all
     * users instead of having the server forward the audio and video
     * packets. Checkout the section @ref connectivity on the different
     * types of connections.
     * 
     * Every channel must configure which audio codec to use in order for
     * users to be able to talk to each other. The @a codec member of
     * #BearWare.Channel specifies which audio codec (#BearWare.AudioCodec) should be
     * used. A channel does not restrict the video codec (#BearWare.VideoCodec)
     * users are using.
     *
     *
     * @defgroup desktopshare Desktop Sharing
     *
     * @brief This section explains how to use the desktop sharing
     * feature where users can share their desktop applications.
     *
     * A user can transmit a desktop window to other users in a
     * channel by passing the handle of a window to the TeamTalk client
     * instance. The TeamTalk client then converts the window to a bitmap
     * image which is transmitted to the server. The server then forwards
     * the bitmap image to all other users in the channel.
     *
     * @section desktoptx Send Desktop Window (or bitmap)
     *
     * Before sending a desktop window to a channel the handle
     * (identifier) of the window to share must first be found. It is,
     * however, possible to simply send a raw bitmap by calling
     * TeamTalk4.SendDesktopWindow().
     *
     * @subsection desktopwin Windows Desktop Sharing
     *
     * TeamTalk for Windows provides the following functions for
     * obtaining different @c HWNDs:
     *
     * - WindowsHelper.GetDesktopActiveHWND()
     *   - Get the @c HWND of the window which has focus.
     * - WindowsHelper.GetDesktopHWND()
     *   - Get the @c HWND of the Windows desktop.
     * - WindowsHelper.GetDesktopWindowHWND()
     *   - Enumerate all visible windows.
     * - WindowsHelper.GetWindow()
     *   - Get information about a window, e.g. window title, size, etc.
     *
     * Once the @c HWND of the window to share has been found use the
     * following function for sending the window to the channel:
     * 
     * - TeamTalk4.SendDesktopWindowFromHWND()
     *
     * @section desktopshow Displaying Shared Desktop Window (or bitmap)
     *
     * When a shared desktop window is received the event
     * TeamTalk4.OnUserDesktopWindow() is posted to the local client
     * instance. TeamTalk4.GetUserDesktopWindow() can then be called to obtain
     * a bitmap image of the shared window.
     *
     * @subsection desktopcursor Desktop Cursor Sharing
     *
     * It is also possible to share the position of the mouse cursor when
     * sharing a desktop window. Use TeamTalk4.SendDesktopCursorPosition() to
     * transmit the position of the mouse cursor. When the position is
     * received the event #BearWare.TeamTalk4.OnUserDesktopCursorPosition is posted to the
     * local client instance. TeamTalk4.GetUserDesktopCursor() can then be used
     * to obtain the cursor position.
     * 
     * @section desktopinput Remote Desktop Access
     *
     * If a user has shared a desktop window it's possible for other users
     * in the same channel to take over control of mouse and keyboard on
     * the computer sharing the desktop window.
     * 
     * @subsection rxdesktopinput Receive Desktop Input
     *
     * In order for a client instance to allow remote desktop access it is
     * required to first subscribe to desktop input from the user who
     * wants access to the shared desktop window. This is done by calling
     * TeamTalk4.DoSubscribe() along with the user-id and subscription
     * ::SUBSCRIBE_DESKTOPINPUT. Once desktop input (mouse or keyboard
     * input) is received from a remote user the
     * TeamTalk4.OnUserDesktopInput() event will be posted to the client
     * instance. The actual mouse or keyboard input can then be obtained
     * by the client instance using TeamTalk4.GetUserDesktopInput() which returns
     * a struct of type #BearWare.DesktopInput. Afterwards
     * WindowsHelper.Execute() can be used to execute the mouse or
     * keyboard input.
     *
     * @subsection txdesktopinput Transmit Desktop Input
     * 
     * The remote user who wants to transmit mouse or keyboard input to
     * the user sharing a desktop window can use
     * TeamTalk4.SendDesktopInput(). Remember that the user sharing the desktop
     * window must have enabled the subscription #SUBSCRIBE_DESKTOPINPUT.
     *
     * @subsection transdesktopinput Desktop Input and Keyboard Layouts
     *
     * It can be quite troublesome to handle keyboard input since each
     * key-code depends on the OS and the regional settings on the
     * OS. E.g. on a German keyboard the Z key is located where the Y key
     * is on a US keyboard. The German keyboard also has letters which
     * don't even appear on a US keyboard.
     *
     * Because of the issues with keyboard layouts and regional settings
     * the TeamTalk API provides #BearWare.WindowsHelper.KeyTranslate which can
     * be used to translate a keyboard's scan-code to an intermediate
     * format. If e.g. a client instance is running Windows then
     * #BearWare.WindowsHelper.KeyTranslate can be called with
     * ::TTKEY_WINKEYCODE_TO_TTKEYCODE which converts the scan-code on a
     * Windows keyboard to TeamTalk's intermediate format (TTKEYCODE). To
     * be able to execute the key-code once it's received it must be
     * converted back again from TeamTalk's intermediate format to the
     * platform where the application is running. I.e. if the TTKEYCODE is
     * received on a Mac then #BearWare.WindowsHelper.KeyTranslate must be
     * called with ::TTKEY_TTKEYCODE_TO_MACKEYCODE.
     *
     * @defgroup events Client Event Handling
     *
     * @brief This section explains how to handle events generated by the
     * client instance.
     *
     * When events occur in the client instance, like e.g. if a new
     * user joins a channel, the client instance invokes an event which
     * the user application must act upon.
     *
     * Note that when an event occurs the TeamTalk client instance doesn't
     * wait for the user application to process the event. So if e.g. a
     * user sends a text-message and immediately after disconnects from
     * the server, then the text-message cannot be retrieved since the
     * user is no longer available when the user application starts
     * processing the new text-message event. This is, of course, annoying
     * when designing the user application, but the reason for this design
     * choice it that the client instance is a realtime component which
     * cannot wait for the UI to process data, since audio playback and
     * recording would then be halted.
     *
     * The section @ref stepbystep gives a good idea of how events are
     * processed in a user application.
     *
     *
     * @defgroup errorhandling Client Error Handling
     *
     * @brief This section explains how to handle errors occuring in the
     * client instance or as a result of server commands.
     *
     * There are two types errors which can occur in the client,
     * either server command error or internal errors. Section @ref
     * commands describes all the commands a client can issue to a
     * server. If a server commands fails the client instance notifies the
     * user application through the event TeamTalk4.OnCmdError(). An
     * example of a server command error could be to issue the TeamTalk4.DoLogin()
     * command with an incorrect server password. The server will in this
     * case respond with the error #ClientError ::CMDERR_INCORRECT_SERVER_PASSWORD. The
     * user application must be designed to process these errors so
     * application users can be notified of errors.
     *
     * Internal errors are errors due to failing devices. Currently only
     * two such errors exist #ClientError ::INTERR_SNDINPUT_FAILURE and
     * #ClientError ::INTERR_SNDOUTPUT_FAILURE.
     *
     *
     * @defgroup connectivity Client/Server Connectivity
     *
     * @brief This section explains how to connect to a server and how the
     * client should transmit voice and video data.
     * 
     * To communicate with a server the TeamTalk client creates
     * both a TCP and UDP connection to the server. Commands, i.e. the
     * @c BearWare.TeamTalk4.Do*-functions, are sent on TCP whereas audio and video are sent
     * on UDP.
     *
     * To connect to a server the user application must call
     * TeamTalk4.Connect(). Once connected the event TeamTalk4.OnConnectSuccess() is
     * posted to the user application and the TeamTalk4.DoLogin() command can be
     * issued. Always ensure to call TeamTalk4.Disconnect() before attempting to
     * create a new connection with TeamTalk4.Connect().
     *
     * When the client instance has joined a channel and wants to transmit
     * audio or video data to other users this can be done in two ways
     * depending on how the user application configures the client. One
     * way is forward through server and the other is peer to peer mode
     * which are explained in the following two sections.
     *
     * @section txforward Forward Through Server Transmission Mode
     *
     * By default the client instance is sending its audio and video
     * packets to the server and the server will then broadcast the
     * packets to the other users on behalf of the client. In other words
     * the client puts the bandwidth load onto the server. This approach
     * has its advantages and disadvantages. Since most internet
     * connections nowadays have limited upstream they cannot broadcast
     * audio and video packets to numerous users at the same time, so
     * therefore the default behaviour of the TeamTalk client is to have
     * the server do the broadcasting. This means that the server must
     * have sufficient bandwidth available to handle data transmission
     * from and to all the connected users. One disadvantage by having the
     * server forward the audio and video packets is that it doubles
     * latency, since the client doesn't send directly to other clients.
     *
     * If the server should not allow clients to forward audio and video
     * packets the @a uUserRights member of #BearWare.ServerProperties must disable
     * #UserRight ::USERRIGHT_FORWARD_AUDIO and #UserRight ::USERRIGHT_FORWARD_VIDEO. Doing so means
     * that clients connecting must use peer to peer connections in order
     * to communicate. Note that the client doesn't automatically switch
     * to peer to peer mode, but relies on the user application to call
     * TeamTalk4.EnablePeerToPeer().
     *
     * @section txp2p Peer to Peer Transmission Mode
     *
     * By calling TeamTalk4.EnablePeerToPeer() the client instance will attempt to
     * create peer to peer connections to all users it's communicating
     * with. This reduces latency since the client will then broadcast the
     * audio and video packets itself, so they do not have to be forwarded
     * through the server.
     *
     * The event TeamTalk4.OnConnectionP2P() is posted to the user application if
     * a peer to peer connection either fails or succeeds. If a peer to
     * peer connection fails the client will send through the server to
     * that user given that the server allows it, i.e. if
     * #UserRight ::USERRIGHT_FORWARD_AUDIO or #UserRight ::USERRIGHT_FORWARD_VIDEO is enabled.
     *
     * If the server does not allow users to forward audio and video
     * packets and the peer to peer connection to a user fails, then that
     * user will be unavailable for audio and video data. So beware of
     * this when configuring the server and client.
     *
     *
     * @defgroup commands Client/Server Commands
     *
     * @brief This section contains the list of commands which can be
     * issued by the client instance to the server.
     *
     * @section cmdprocessing Client/Server Command Processing
     *
     * Functions with the prefix @c Do* are commands which the client can
     * issue to the server. Every @c Do* function returns a command
     * identifier which can user application can use to check when the
     * server has finished processing the issued command. Once the client
     * receives a response to a command the client instance posts the
     * event TeamTalk4.OnCmdProcessing() to the user application
     * containing the command identifier and whether the command is being
     * processed or it has completed.
     *
     * As an example, say the user application wants to issue the
     * TeamTalk4.DoLogin() command. When the application calls TeamTalk4.DoLogin() the
     * returned command ID is stored in a variable. The application then
     * waits for the TeamTalk4.OnCmdProcessing() event to be posted with
     * the stored command ID. The first time TeamTalk4.OnCmdProcessing()
     * is posted to the user application it is to say that processing has
     * begun. The second time TeamTalk4.OnCmdProcessing() is called it is
     * to say that the command has completed. In between the command
     * starting and completing several other events may take place. If
     * e.g. the TeamTalk4.DoLogin() fails the user application would receive the
     * event TeamTalk4.OnCmdError(). If on the other hand the command was
     * successful all the channels and user would be posted as events to
     * the application before the login-command completed processing.
     *
     *
     * @defgroup transmission Audio and Video Transmission
     *
     * @brief This section explains how to transmit audio and video to
     * users in a channel.
     *
     * Once the client instance has joined a channel it can
     * transmit audio and video to other users in the channel. When
     * transmitting audio and video it is important to be aware of how the
     * client is configured to do this. Sections @ref txforward and @ref
     * txp2p explains the two supported data transmission modes.
     *
     * To transmit audio the client must have the flag
     * #ClientFlag ::CLIENT_SNDINPUT_READY enabled which is done in the function
     * TeamTalk4.InitSoundInputDevice(). To transmit video requires the flag
     * #ClientFlag ::CLIENT_VIDEO_READY which is enabled by the function
     * TeamTalk4.InitVideoCaptureDevice(). To hear what others users are saying a
     * sound output device must have been configured using
     * TeamTalk4.InitSoundOutputDevice() and thereby have enabled the flag
     * #ClientFlag ::CLIENT_SNDOUTPUT_READY.
     *
     * Calling TeamTalk4.EnableTransmission() will make the client instance start
     * transmitting either audio or video data. Note that audio
     * transmission can also be activated automatically using voice
     * activation. This is done by called TeamTalk4.EnableVoiceActivation() and
     * setting a voice activation level using TeamTalk4.SetVoiceActivationLevel().
     *
     *
     * @defgroup server Server Properties
     *
     * @brief This section explains how to configure a server and setup
     * user accounts.
     *
     * The server keeps track of which users are in which channels
     * and ensures that users in the same channel can communicate with
     * each other. It is also the job of the server to provide user
     * authentication so only users with the proper credentials are
     * allowed to do certain operations.
     *
     * It is a good idea to check out section @ref serversetup to learn
     * how to configure the TeamTalk server.
     *
     * The server can be configured in a number of ways using the
     * #BearWare.ServerProperties-struct. Only users who are logged on to the
     * server as administrators can change a server's properties while
     * it's running. This is done using the command TeamTalk4.DoUpdateServer().
     *
     * The @a uUserRights bitmask in #BearWare.ServerProperties specifies what
     * users are allowed to do. The server can e.g. be configured to only
     * allow users who have an account to log on by disabling the flag
     * #UserRight ::USERRIGHT_GUEST_LOGIN. If default users (without administrator
     * rights) shouldn't be allowed to create channels this can be
     * disabled by disabling the flag #UserRight ::USERRIGHT_CHANNEL_CREATION.
     *
     * @section useradmin User Administration
     *
     * Two types of users exists on a server, default users and
     * administrator users. The #UserType-enum can be used to see who is
     * what. To be administrator the user must have a user account on the
     * server with the #UserType ::USERTYPE_ADMIN flag set.
     *
     * As administrator it is possible to list all users who have an
     * account on the server using TeamTalk4.DoListUserAccounts(). To create a new
     * user account call the command TeamTalk4.DoNewUserAccount() and to delete an
     * account call TeamTalk4.DoDeleteUserAccount().
     *
     * @section userban Kicking and Banning Users
     *
     * Sometimes it may be necessary to kick and ban users from a
     * server. As administrator it is possible to use the command
     * TeamTalk4.DoKickUser() to kick a user off the server. A channel operator
     * can also kick a user from a channel (but not off a server).
     *
     * As administrator it is also possible to ban users from the server,
     * so they can no longer log in. This can be done using
     * TeamTalk4.DoBanUser(). To list who are currently banned call TeamTalk4.DoListBans()
     * and to remove a ban call TeamTalk4.DoUnBanUser().
     *
     *
     * @defgroup channels Channel Properties
     *
     * @brief This section explains the concept of channels where users
     * can interact.
     *
     * Users are arranged in a tree structure consisting of
     * channels where each channel can hold a number of users. While
     * in a channel users can transmit audio and video to each other
     * as well as sending channel messages. On a server there will
     * always be a root channel which cannot be deleted.
     *
     * In other conferencing tools channels are also refered to as
     * "rooms".
     *
     * @section chanadmin Channel Administration
     *
     * To create a new channel on a server an administrator can call
     * TeamTalk4.DoMakeChannel() whereas a default user (without administrator
     * rights) has to use TeamTalk4.DoJoinChannel() to create a new channel. Using
     * TeamTalk4.DoJoinChannel() creates a dynamic channel which disappears again
     * from the server when the last user leaves the channel. The first
     * user to join a dynamic channel will become operator of the channel,
     * meaning that he can kick user and make other operators as well.
     *
     * Updating a channel's properties can only be done by administrators
     * and operators of the channel and this is done by using
     * TeamTalk4.DoUpdateChannel().
     *
     * To remove a channel a user must be administrator and can do so by
     * calling TeamTalk4.DoRemoveChannel().
     * 
     * @section filesharing File Sharing
     *
     * While in a channel users can upload and download files. To upload a
     * file to a channel the channel needs to have a disk quota. The disk
     * quota is specified by @a nDiskQuota in the #BearWare.Channel-struct. The
     * file being uploaded must have a file size which is less than the
     * disk quota and the sum of sizes of existing files. Once a file is
     * uploaded only administrators, channel operators and the file's
     * owner can delete a file. Note that the file's owner must have an
     * account on the server to delete the file.
     *
     * Call TeamTalk4.DoSendFile() to upload a file and
     * TeamTalk4.DoRecvFile() to download a file. Only users who have
     * a #BearWare.UserAccount on the server are allowed to upload
     * files. There is no limit on the maximum number of file
     * transfers but it is advised to queue file transfers so the file
     * transfers do no affect server performance.
     *
     * @section voicelog Storing Conversations to Audio Files
     *
     * In some applications it may be required to be able to save all
     * audio data received by the client instance to disk. This can be
     * archived by calling TeamTalk4.SetUserAudioFolder() which
     * will then save received audio data in the following format:
     * "YYYYMMDD-HHMMSS \#USERID USERNAME.wav". USERNAME is the @a
     * szUsername from #BearWare.User.
     *
     * To store audio data from outside the local client instance's
     * channel, please read section @ref spying.
     *
     *
     * @defgroup users User Properties
     *
     * @brief This section explains users interact and how to configure
     * user settings.
     *
     * Users can be seen on the server after a successful call
     * to TeamTalk4.DoLogin(). Once logged in a user can send user to user
     * text-messages using TeamTalk4.DoTextMessage() as well as receive
     * broadcast messages. A user cannot send audio and video data to
     * other users until they have joined the same channel.
     *
     * @section usertypes User Types
     *
     * A user can either be a default user #UserType ::USERTYPE_DEFAULT or an
     * administrator #UserType ::USERTYPE_ADMIN. A default user has limited rights on the
     * server whereas an administrator can change server properties,
     * create, delete and remove channels as well as move, kick and ban
     * users.
     *
     * @section userinteract User Interaction
     *
     * Once a user has joined a channel it is possible to transmit audio
     * and video data to other users. If a user starts talking in the
     * channel the event TeamTalk4.OnUserTalking() is posted to the user
     * application and if a video frame is received the event
     * TeamTalk4.OnUserVideoFrame() is sent to the user application.
     *
     * @section uservolume User Audio Settings
     *
     * While in the same channel the user application can change different
     * settings on a user. If e.g. a user's volume is too low the user
     * application can call TeamTalk4.SetUserVolume() to increase the volume. If
     * it's still not loud enough TeamTalk4.SetUserGainLevel() can be called to
     * use software to amplify the volume level.
     *
     * If on the other hand the user application wants to mute a user
     * TeamTalk4.SetUserMute() can be used for this. Note that muting a user
     * doesn't mean that the client instance will stop receiving audio
     * from that user, it simply means it will not be played. To stop
     * receiving audio from a user the local client instance must ask the
     * server to unsubscribe audio data from the user. This is explained
     * in the next section.
     *
     * @section subscriptions User Subscriptions
     * 
     * When logging on to a server the local client instance will by
     * default subscribe to user messages, channel messages, broadcast
     * messages, audio data and video data from all users. If,
     * however, a client wants to stop receiving e.g. audio from a
     * user, he can call TeamTalk4.DoUnsubscribe() along with the user
     * ID and the #Subscription ::SUBSCRIBE_AUDIO-flag to tell the
     * server that he no longer wants to receive audio from that
     * user. The server will then respond with the event
     * TeamTalk4.OnCmdUserUpdate() and the @a uLocalSubscriptions
     * member of #BearWare.User will have the
     * #Subscription ::SUBSCRIBE_AUDIO-flag removed. At the remote user
     * the \a uPeerSubscriptions member will be
     * changed. Subscribe/unsubscribe can also be done for user,
     * channel and broadcast messages and video data. The function
     * TeamTalk4.DoSubscribe() can be used to revert unsubscriptions.
     *
     * @subsection spying Spying on Users
     *
     * Previously it has been said that users can only receive audio and
     * video from users when they are in the same channel, but actually an
     * administrator user can call TeamTalk4.DoSubscribe() with the flags prefixed
     * SUBSCRIBE_INTERCEPT_* to spy on users outside his own channel. In
     * other words it's possible hear and see video data outside ones
     * channel. Also all user and channel messages sent by a user can also
     * be intercepted in this way.
     *
     * Having the ability to intercept all data sent from users in any
     * channel means that it's possible to log everything that is
     * happening on the server. Both audio and video transfers as well
     * as text messaging. Checkout TeamTalk4.SetUserAudioFolder()
     * on how to store voice data to audio files.
     *
     *
     * @defgroup hotkey Windows Hotkeys
     *
     * @brief This section explains how to setup hot-keys on Windows.
     *
     * Hotkeys can be used to e.g. enable push-to-talk.
     *
     * Windows supports intercepting key strokes globally, i.e. without
     * having the user application's window focused. To investigate which
     * keys are currently being pressed the function
     * TeamTalk4.HotKey_InstallTestHook() can be used. Once the desired
     * key-combination has been found the function TeamTalk4.HotKey_Register() can
     * be used to register the combination as a hotkey and have the
     * TeamTalk4.OnHotKeyToggle() event posted whenever the key combination
     * becomes active.
     *
     * Note that it's not advised to have a hotkey installed while
     * debugging an application in Visual Studio. It slows down the
     * debugger dramatically.
     * 
     *
     * @defgroup mixer Windows Mixer
     *
     * @brief This section explains how to control the Windows mixer.
     *
     * The Windows mixer can also be manipulated so e.g. Line In
     * can be chosen instead of Microphone for recording.
     *
     * To find the mixer which is associated with the current sound input
     * or output device the @a nWaveDeviceID member of #BearWare.SoundDevice must
     * be used when querying the mixer.
     *
     *
     * @defgroup firewall Windows Firewall
     *
     * @brief This section explains how to configure the Windows firewall
     * available in Windows XP SP2 and later.
     *
     * The Windows Firewall can be modified so applications can be
     * added to the firewall's exception list.
     *
     * The Windows Firewall was introduced in Windows XP SP2. Modifying
     * the Windows Firewall requires administrator rights. On Windows XP
     * the user running the application, which calls the DLL, is assumed
     * to have administrator rights. On Windows Vista/7 the DLL will
     * automatically call User Account Control (UAC) to obtain
     * administrator rights.
     *
     * Check out #BearWare.WindowsFirewall class on how to add
     * application executables to the Windows Firewall exception list.
     *
     */

    /** @addtogroup sounddevices
     * @{ */

    /**
     * @brief The supported sound systems.
     *
     * @see SoundDevice
     * @see TeamTalk4.InitSoundInputDevice()
     * @see TeamTalk4.InitSoundOutputDevice()
     * @see TeamTalk4.InitSoundDuplexDevices() */
    public enum SoundSystem : uint
    {
        /** @brief Sound system denoting invalid or not found. */
        SOUNDSYSTEM_NONE = 0,
        /** @brief Windows legacy audio system. Should be used on Windows Mobile. */
        SOUNDSYSTEM_WINMM = 1,
        /** @brief DirectSound audio system. Should be used on Windows 2K/XP. */
        SOUNDSYSTEM_DSOUND = 2,
        /**
         * @brief Advanced Linux Sound Architecture (ALSA). Should be used on Linux.
         *
         * Often ALSA sound devices only support a limited number of
         * sample rates so TeamTalk internally use software filters to
         * resample the audio to the sample rate used by the selected
         * audio codecs. */
        SOUNDSYSTEM_ALSA = 3,
        /** @brief Core Audio. Should be used on MacOS. */
        SOUNDSYSTEM_COREAUDIO = 4,
        /** @brief Windows Audio Session API (WASAPI). Should be used
         * on Windows Vista/7.
         *
         * WASAPI audio devices typically only support a single sample
         * rate so internally TeamTalk uses software filters to
         * resample audio to the sample rate used by the selected
         * audio codecs.
         * 
         * Check @c supportedSampleRates and @c nDefaultSampleRate of
         * #BearWare.SoundDevice to see which sample rates are supported. */
        SOUNDSYSTEM_WASAPI = 5
    }

    /**
     * @brief A struct containing the properties of a sound device
     * for either playback or recording.
     *
     * Use @a nDeviceID to pass to TeamTalk4.InitSoundInputDevice() or
     * TeamTalk4.InitSoundOutputDevice().
     *
     * Note that the @a nDeviceID may change if the user application
     * is restarted and a new sound device is added or removed from
     * the computer.
     * 
     * @see TeamTalk4.GetSoundInputDevices
     * @see TeamTalk4.GetSoundOutputDevices */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct SoundDevice
    {
        /** @brief The ID of the sound device. Used for passing to
         * TeamTalk4.InitSoundInputDevice() and
         * TeamTalk4.InitSoundOutputDevice(). Note that @a nDeviceID might change
         * if USB sound devices are plugged in or unplugged, therefore
         * use @a szDeviceID to ensure proper device is used.  */
        public int nDeviceID;
        /** @brief The sound system used by the sound device. */
        public SoundSystem nSoundSystem;
        /** @brief The name of the sound device. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szDeviceName;
        /** @brief An identifier uniquely identifying the sound device
         * even when new sound devices are being added and removed.  In
         * DirectSound, WASAPI and WinMM it would be the GUID of the sound
         * device. Note that it may not always be available. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szDeviceID;
        /** 
         * @brief The ID of the device used in Win32's
         * waveInGetDevCaps and waveOutGetDevCaps.
         *
         * Value will be -1 if no ID could be found This ID can also
         * be used to find the corresponding mixer on Windows passing
         * it as @a nWaveDeviceID.  Note that this ID applies both to
         * DirectSound and WinMM.
         *
         * @see WindowsMixer.GetWaveInName
         * @see WindowsMixer.GetWaveOutName
         * @see WindowsMixer.GetMixerCount */
        public int nWaveDeviceID;
        /** @brief Whether the sound device supports 3D-sound
         * effects. */
        public bool bSupports3D;
        /** @brief The maximum number of input channels. */
        public int nMaxInputChannels;
        /** @brief The maximum number of output channels. */
        public int nMaxOutputChannels;
        /** @brief Supported sample rates by device. A zero value terminates
         * the list of supported sample rates or its maximum size of 16. 
         * Investigating the support sample rates is usually only required
         * on Linux since sound devices often don't numerous sample rates. */
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = TeamTalk4.TT_SAMPLERATES_MAX)]
        public int[] supportedSampleRates;
        /** @brief The default sample rate for the sound device. */
        public int nDefaultSampleRate;
    }

    /**
     * @brief An enum encapsulation the minimum, maximum and default sound
     * levels for input and output sound devices. */
    public struct SoundLevel
    {
        /**
         * @brief The maximum value of recorded audio.
         * @see TeamTalk4.GetSoundInputLevel
         * @see TeamTalk4.SetVoiceActivationLevel
         * @see TeamTalk4.GetVoiceActivationLevel */
        public const int SOUND_VU_MAX = 20;
        /**
         * @brief The minimum value of recorded audio.
         * @see TeamTalk4.GetSoundInputLevel
         * @see TeamTalk4.SetVoiceActivationLevel
         * @see TeamTalk4.GetVoiceActivationLevel */
        public const int SOUND_VU_MIN = 0;
        /**
         * @brief The maximum volume for master volume. To gain the
         * volume level using software call TeamTalk4.SetUserGainLevel().
         *
         * @see BearWare.TeamTalk4.SetSoundOutputVolume
         * @see BearWare.TeamTalk4.GetSoundOutputVolume
         * @see BearWare.TeamTalk4.SetUserVolume
         * @see BearWare.TeamTalk4.GetUserVolume */
        public const int SOUND_VOLUME_MAX = 255;
        /**
         * @brief The minimum volume for master volume.
         * @see BearWare.TeamTalk4.SetSoundOutputVolume
         * @see BearWare.TeamTalk4.GetSoundOutputVolume
         * @see BearWare.TeamTalk4.SetUserVolume
         * @see BearWare.TeamTalk4.GetUserVolume */
        public const int SOUND_VOLUME_MIN = 0;
        /**
         * @brief The maximum gain level. 
         *
         * A gain level of 32000 gains the volume by a factor 32.  A gain
         * level of #SOUND_GAIN_DEFAULT means no gain.
         *
         * @see BearWare.TeamTalk4.SetSoundInputGainLevel
         * @see BearWare.TeamTalk4.GetSoundInputGainLevel
         * @see BearWare.TeamTalk4.SetUserGainLevel
         * @see BearWare.TeamTalk4.GetUserGainLevel */
        public const int SOUND_GAIN_MAX = 32000;
        /**
         * @brief The default gain level.
         *
         * A gain level of 1000 means no gain. Check #SOUND_GAIN_MAX
         * and #SOUND_GAIN_MIN to see how to increase and lower gain
         * level.
         *
         * @see BearWare.TeamTalk4.SetSoundInputGainLevel
         * @see BearWare.TeamTalk4.GetSoundInputGainLevel
         * @see BearWare.TeamTalk4.SetUserGainLevel
         * @see BearWare.TeamTalk4.GetUserGainLevel */
        public const int SOUND_GAIN_DEFAULT = 1000;
        /**
         * @brief The minimum gain level (since it's zero it means
         * silence).
         *
         * A gain level of 100 is 1/10 of the default volume.
         *
         * @see BearWare.TeamTalk4.SetSoundInputGainLevel
         * @see BearWare.TeamTalk4.GetSoundInputGainLevel
         * @see BearWare.TeamTalk4.SetUserGainLevel
         * @see BearWare.TeamTalk4.GetUserGainLevel */
        public const int SOUND_GAIN_MIN = 0;
    }

    /**
     * @brief Status of audio files being written to disk.
     * @see OnUserAudioFile */
    public enum AudioFileStatus : uint
    {
        /** @brief Error while processing audio file. */
        AFS_ERROR = 0,
        /** @brief Started processing audio file. */
        AFS_STARTED = 1,
        /** @brief Finished processing audio file. */
        AFS_FINISHED = 2,
        /** @brief Aborted processing of audio file. */
        AFS_ABORTED = 3
    }

    /**
     * @brief Audio file formats supported for muxed audio recordings.
     * @see TeamTalk4.StartRecordingMuxedAudioFile() */
    public enum AudioFileFormat : uint
    {
        /** @brief Used to denote nothing selected. */
        AFF_NONE = 0,
        /** @brief Store in 16-bit wave format. */
        AFF_WAVE_FORMAT = 2,
        /** @brief Store in MP3-format. This requires lame_enc.dll to
         * be in the same directory as the application's
         * execuable. The LAME DLLs can be obtained from
         * http://lame.sourceforge.net. Note that the MP3-format is
         * subject to licensing by Fraunhofer and Thomson
         * Multimedia. */
        AFF_MP3_64KBIT_FORMAT = 3,
        /** @see #AudioFileFormat ::AFF_MP3_64KBIT_FORMAT */
        AFF_MP3_128KBIT_FORMAT = 4,
        /** @see #AudioFileFormat ::AFF_MP3_64KBIT_FORMAT */
        AFF_MP3_256KBIT_FORMAT = 5,
        /** @see #AudioFileFormat ::AFF_MP3_64KBIT_FORMAT */
        AFF_MP3_16KBIT_FORMAT  = 6,
        /** @see #AudioFileFormat ::AFF_MP3_64KBIT_FORMAT */
        AFF_MP3_32KBIT_FORMAT  = 7
    }


    /**
     * @brief An audio block containing the raw audio from a user who
     * was talking.
     *
     * To enable audio blocks first call TT_EnableAudioBlockEvent()
     * then whenever new audio is played the event
     * TeamTalk4.OnUserAudioBlock() is generated. Use
     * TeamTalk4.AcquireUserAudioBlock() to retrieve the audio block.
     *
     * Note that each user is limited to 128 kbytes of audio data.
     *
     * @see TeamTalk4.EnableAudioBlockEvent()
     * @see TeamTalk4.AcquireUserAudioBlock()
     * @see TeamTalk4.ReleaseUserAudioBlock() */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct AudioBlock
    {
        /** @brief The sample rate of the raw audio. */
        public int nSampleRate;
        /** @brief The number of channels used (1 for mono, 2 for stereo). */
        public int nChannels;
        /** @brief The raw audio in 16-bit integer format array. The
         * size of the array in bytes is @c sizeof(short) * @c
         * nSamples * @c nChannels. */
        public System.IntPtr lpRawAudio;
        /** @brief The number of samples in the raw audio array. */
        public int nSamples;
        /** @brief The index of the first sample in @c lpRawAudio. Its
         * value will be a multiple of @c nSamples. The sample index
         * can be used to detect overflows of the internal
         * buffer. When a user initially starts talking the @c
         * nSampleIndex will be 0 and while the user is talking @c
         * nSampleIndex will be greater than 0. When the user stops
         * talking @c nSampleIndex will be reset to 0 again. */
        public uint uSampleIndex;
    }

    /**
     * @brief Struct describing the audio format used by a
     * media file.
     *
     * @see TeamTalk4.GetMediaFileInfo()
     * @see MediaFileInfo
     */
    public struct AudioFormat
    {
        /** @brief The audio file format, e.g. wave or MP3. */
        public AudioFileFormat nAudioFmt;
        /** @brief Sample rate of media file. */
        public int nSampleRate;
        /** @brief Channels used by media file, mono = 1, stereo = 2. */
        public int nChannels;
    }

    /** @} */

    /** @addtogroup videocapture
     * @{ */

    /** 
     * @brief The picture format used by a capture device. 
     *
     * @see CaptureFormat
     * @see VideoCaptureDevice */
    public enum FourCC : uint
    {
        /** @brief Internal use to denote no supported formats. */
        FOURCC_NONE     = 0,
        /** @brief Prefered image format with the lowest bandwidth
         * usage. A 640x480 pixel image takes up 460.800 bytes. */
        FOURCC_I420     = 100,
        /** @brief Image format where a 640x480 pixel images takes up
         * 614.400 bytes. */
        FOURCC_YUY2     = 101,
        /** @brief The image format with the highest bandwidth
         * usage. A 640x480 pixel images takes up 1.228.880 bytes. */
        FOURCC_RGB32    = 102
    }

    /**
     * @brief A struct containing the properties of a video capture
     * format.
     *
     * A struct for holding a supported video capture format by a 
     * #BearWare.VideoCaptureDevice. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct CaptureFormat
    {
        /** @brief The width in pixels of the video device supported
         * video format. */
        public int nWidth;
        /** @brief The height in pixels of the video device supported
         * video format. */
        public int nHeight;
        /** @brief The numerator of the video capture device's video
         * format. Divinding @a nFPS_Numerator with @a
         * nFPS_Denominator gives the frame-rate. */
        public int nFPS_Numerator;
        /** @brief The denominator of the video capture device's video
         * format. Divinding @a nFPS_Numerator with @a
         * nFPS_Denominator gives the frame-rate.*/
        public int nFPS_Denominator;
        /** @brief Picture format for capturing. */
        public FourCC picFourCC;
    }

    /**
     * @brief A RGB32 image where the pixels can be accessed directly
     * in an allocated @a imageBuffer.
     *
     * Use TeamTalk4.AcquireUserVideoFrame() to acquire a user's image and
     * remember to call TeamTalk4.ReleaseUserVideoFrame() when the image has
     * been processed so TeamTalk can release its resources. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct VideoFrame
    {
        /** @brief The width in pixels of the image contained in @a
         * frameBuffer. */
        public int nWidth;
        /** @brief The height in pixels of the image contained in @a
         * imageBuffer. */
        public int nHeight;
        /** @brief A unique identifier for the frames which are part of the
         * same video sequence. If the stream ID changes it means the
         * frames which are being received are part of a new video sequence
         * and @a nWidth and @a nHeight may have changed. The @a nStreamID
         * will always be a positive integer value.*/
        public int nStreamID;
        /** @brief Whether the image acquired is a key-frame. If it is
         * not a key-frame and there has been packet loss or a
         * key-frame has not been acquired prior then the image may
         * look blurred. */
        public bool bKeyFrame;
        /** @brief A buffer allocated internally by TeamTalk. */
        public System.IntPtr frameBuffer;
        /** @brief The size in bytes of the buffer allocate in @a
         * frameBuffer. */
        public int nFrameBufferSize;
    }

    /** 
     * @brief A struct containing the properties of a video capture
     * device.
     *
     * The information retrieved from the video capture device is used
     * to initialize the video capture device using the
     * TeamTalk4.InitVideoCaptureDevice() function.
     * 
     * @see TeamTalk4.GetVideoCaptureDevices */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct VideoCaptureDevice
    {
        /** @brief A string identifying the device. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szDeviceID;
        /** @brief The name of the capture device. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szDeviceName;
        /** @brief The name of the API used to capture video. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szCaptureAPI;
        /** @brief The supported capture formats. */
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = TeamTalk4.TT_CAPTUREFORMATS_MAX)]
        public CaptureFormat[] captureFormats;
        /** @brief The number of capture formats available in @a
         * captureFormats array. */
        public int nCaptureFormatsCount;
    }

    /**
     * @brief Struct describing the audio and video format used by a
     * media file.
     *
     * @see TT_GetMediaFile() */
    public struct MediaFileInfo
    {
        /** @brief The audio properties of the media file. */
        public AudioFormat audioFmt;
        /** @brief The video properties of the media file. */
        public CaptureFormat videoFmt;
        /** @brief The duration of the media file in miliseconds. */
        public uint uDurationMSec;
    }


    /** @} */

    /** @addtogroup desktopshare
     * @{ */

    /**
     * @brief The bitmap format used for a #BearWare.DesktopWindow. */
    public enum BitmapFormat : uint
    {
        /** @brief Used to denote nothing selected. */
        BMP_NONE            = 0,
        /** @brief The bitmap is a 256-colored bitmap requiring a
         * palette. The default 256 colored palette is the Netscape
         * browser-safe palette. Use TeamTalk4.Palette_GetColorTable() to
         * access or change the palette. The maximum size of a 
         * 8-bit bitmap is 4095 blocks of 120 by 34 pixels. */
        BMP_RGB8_PALETTE    = 1,
        /** @brief The bitmap is a 16-bit colored bitmap. The maximum
         * pixels. */
        BMP_RGB16_555       = 2,
        /** @brief The bitmap is a 24-bit colored bitmap. The maximum
         * size of a 24-bit bitmap is 4095 blocks of 85 by 16
         * pixels. */
        BMP_RGB24           = 3,
        /** @brief The bitmap is a 32-bit colored bitmap. The maximum
         * size of a 32-bit bitmap is 4095 blocks of 51 by 20
         * pixels. */
        BMP_RGB32           = 4
    }

    /** @brief The protocols supported for transferring a
     * #BearWare.DesktopWindow.
     *
     * So far only one, UDP-based, protocol is supported. */
    public enum DesktopProtocol : uint
    {
        /** @brief Desktop protocol based on ZLIB for image
         * compression and UDP for data transmission. */
        DESKTOPPROTOCOL_ZLIB_1  = 1
    }

    /**
     * @brief A struct containing the properties of a shared desktop window.
     *
     * The desktop window is a description of the bitmap which can be retrieved using 
     * TeamTalk4.GetUserDesktopWindow() or the bitmap which should be transmitted using
     * TeamTalk4.SendDesktopWindow(). */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct DesktopWindow
    {
        /** @brief The width in pixels of the bitmap. */
        public int nWidth;
        /** @brief The height in pixels of the bitmap. */
        public int nHeight;
        /** @brief The format of the bitmap. */
        public BitmapFormat bmpFormat;
        /** @brief The number of bytes for each scan-line in the
         * bitmap. Zero means 4-byte aligned. */
        public int nBytesPerLine;
        /** @brief The ID of the session which the bitmap belongs
         * to. If the session ID changes it means the user has started
         * a new session. This e.g. happens if the desktop session has
         * been closed and restart or if the bitmap has been
         * resized. Set @c nSessionID to 0 if the desktop window is
         * used with TT_SendDesktopWindow(). */
        public int nSessionID;
        /** @brief The desktop protocol used for transmitting the desktop window. */
        public DesktopProtocol nProtocol;
    }

    /**
     * @brief The state of a key (or mouse button), i.e. if it's
     * pressed or released. @see DesktopInput */
    public enum DesktopKeyState : uint
    {
        /** @brief The key is ignored. */
        DESKTOPKEYSTATE_NONE       = 0x00000000,
        /** @brief The key is pressed. */
        DESKTOPKEYSTATE_DOWN       = 0x00000001,
        /** @brief The key is released. */
        DESKTOPKEYSTATE_UP         = 0x00000002,
    }

    /**
     * @brief A struct containing a mouse or keyboard event.
     *
     * The DesktopInput struct is used for desktop access where a
     * remote user can control mouse or keybaord on a shared
     * desktop. Check out section @ref desktopinput on how to use
     * remote desktop access. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct DesktopInput
    {
        /** @brief The X coordinate of the mouse. If used with
         * WindowsHelper.Execute() and the mouse position should be
         * ignored then set to #BearWare.TeamTalk4.TT_DESKTOPINPUT_MOUSEPOS_IGNORE. */
        public ushort uMousePosX;
        /** @brief The Y coordinate of the mouse. If used with
         * TT_DesktopInput_Execute() and the mouse position should be
         * ignored then set to #BearWare.TeamTalk4.TT_DESKTOPINPUT_MOUSEPOS_IGNORE. */
        public ushort uMousePosY;
        /** @brief The key-code (or mouse button) pressed. If used
         * with TT_DesktopInput_Execute() and no key (or mouse button)
         * is pressed then set to #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_IGNORE.
         * Read section @ref transdesktopinput on issues with
         * key-codes and keyboard settings. */
        public uint uKeyCode;
        /** @brief The state of the key (or mouse button) pressed,
         * i.e. if it's up or down. */
        public DesktopKeyState uKeyState;
    }

    /** @} */


    /** @addtogroup codecs
     * @{ */

    /**
     * @brief Speex audio codec settings. The Speex codec is
     * recommended for voice and uses less bandwidth than
     * #BearWare.CELTCodec. */
    [StructLayout(LayoutKind.Explicit, CharSet = CharSet.Unicode)] //Windows Mobile tries to save space, so need to be explicit
    public struct SpeexCodec
    {
        /** @brief 0 means 8 KHz, 1 means 16 KHz, 2 means 32 KHz */
        [FieldOffset(0)]
        public int nBandmode;
        /** @brief A value from 1-10. */
        [FieldOffset(4)]
        public int nQuality;
        /** @brief Milliseconds of audio data in each
         * packet. Recommended is 40 ms. Max is 1000. */
        [FieldOffset(8)]
        public int nMSecPerPacket;
        /** @brief Use Speex' jitter buffer for playback. Recommended
         * is FALSE. */
        [FieldOffset(12)]
        public bool bUseJitterBuffer;
        /** @brief Playback should be done in stereo. Doing so will
         * disable 3d-positioning.
         *
         * @see TeamTalk4.SetUserPosition
         * @see TeamTalk4.SetUserStereo */
        [FieldOffset(16)]
        public bool bStereoPlayback;
    }

    /** @brief Speex audio codec settings for Variable Bitrate mode
     * (VBR). The Speex codec is recommended for voice and uses less
     * bandwidth than #BearWare.CELTCodec. The Speex VBR codec was
     * introduced in version 4.2. */
    [StructLayout(LayoutKind.Explicit, CharSet = CharSet.Unicode)] //Windows Mobile tries to save space, so need to be explicit
    public struct SpeexVBRCodec
    {
        /** @brief Set to 0 for 8 KHz (narrow band), set to 1 for 16 KHz 
         * (wide band), set to 2 for 32 KHz (ultra-wide band). */
        [FieldOffset(0)]
        public int nBandmode;
        /** @brief A value from 0-10. If @c nBitRate is non-zero it
         * will override this value. */
        [FieldOffset(4)]
        public int nQualityVBR;
        /** @brief The bitrate at which the audio codec should output
         * encoded audio data. Dividing it by 8 gives roughly the
         * number of bytes per second used for transmitting the
         * encoded data. For limits check out
         * #BearWare.SpeexConstants. Note that specifying @c nBitRate
         * will override nQualityVBR. */
        [FieldOffset(8)]
        public int nBitRate;
        /** @brief The maximum bitrate at which the audio codec is
         * allowed to output audio. Set to zero if it should be
         * ignored. */
        [FieldOffset(12)]
        public int nMaxBitRate;
        /** @brief Enable/disable discontinuous transmission. When
         * enabled Speex will ignore silence, so the bitrate will
         * become very low. */
        [FieldOffset(16)]
        public bool bDTX;
        /** @brief Milliseconds of audio data in each packet. Speex
         * uses 20 msec frame sizes. Recommended is 40 ms. Min is 20,
         * max is 1000. */
        [FieldOffset(20)]
        public int nMSecPerPacket;
        /** @brief Use Speex' jitter buffer for playback. Recommended
         * is FALSE. */
        [FieldOffset(24)]
        public bool bUseJitterBuffer;
        /** @brief Playback should be done in stereo. Doing so will
         * disable 3d-positioning.
         *
         * @see TeamTalk4.SetUserPosition
         * @see TeamTalk4.SetUserStereo */
        [FieldOffset(28)]
        public bool bStereoPlayback; 
    }

    /** @brief Speex constants */
    public struct SpeexConstants
    {
        /** @brief The minimum bitrate for Speex codec in 8 KHz
         * mode, i.e. quality set to 0. */
        public const int SPEEX_NB_MIN_BITRATE = 2150;
        /** @brief The maximum bitrate for Speex codec in 8 KHz
         * mode, i.e. quality set to 10. */
        public const int SPEEX_NB_MAX_BITRATE = 24600;
        /** @brief The minimum bitrate for Speex codec in 16 KHz
         * mode, i.e. quality set to 0. */
        public const int SPEEX_WB_MIN_BITRATE = 3950;
        /** @brief The maximum bitrate for Speex codec in 16 KHz
         * mode, i.e. quality set to 10. */
        public const int SPEEX_WB_MAX_BITRATE = 42200;
        /** @brief The minimum bitrate for Speex codec in 32 KHz
         * mode, i.e. quality set to 0. */
        public const int SPEEX_UWB_MIN_BITRATE = 4150;
        /** @brief The maximum bitrate for Speex codec in 32 KHz
         * mode, i.e. quality set to 10. */
        public const int SPEEX_UWB_MAX_BITRATE = 44000;
    }

    /**
     * @brief CELT audio codec settings. The CELT codec is
     * recommended for music and speech. It has a higher bandwidth
     * usage than #BearWare.SpeexCodec. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct CELTCodec
    {
        /** @brief The sample rate to use. Sample rate must be in the
         * range 32000 - 96000 Hz. */
        public int nSampleRate;
        /** @brief Mono = 1 or stereo = 2. Note that echo
         * cancellation, denoising and AGC is not supported when using
         * stereo.
         *
         * @see TeamTalk4.EnableEchoCancellation()
         * @see TeamTalk4.EnableDenoising()
         * @see TeamTalk4.EnableAGC() */
        public int nChannels;
        /** @brief The bitrate at which the audio codec should output
         * encoded audio data. Dividing it by 8 gives roughly the
         * number of bytes per second used for transmitting the
         * encoded data. Check out #BearWare.CELTConstants for
         * bitrate limits. */
        public int nBitRate;
        /** @brief Milliseconds of audio data in each
         * packet. Recommended is 40 ms. Min is 20, max is 1000. */
        public int nMSecPerPacket;
    }

    /** @brief CELT audio codec settings. The CELT codec is
     * recommended for music and speech. It has a higher bandwidth
     * usage than #BearWare.SpeexCodec. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct CELTVBRCodec
    {
        /** @brief The sample rate to use. Sample rate must be in the
         * range 32000 - 96000 Hz. */
        public int nSampleRate;
        /** @brief Mono = 1 or stereo = 2. Note that echo
         * cancellation, denoising and AGC is not support when using
         * stereo.
         *
         * @see TeamTalk4.EnableEchoCancellation()
         * @see TeamTalk4.EnableDenoising()
         * @see TeamTalk4.EnableAGC() */
        public int nChannels;
        /** @brief The bitrate at which the audio codec can output
         * encoded audio data. Dividing it by 8 gives roughly the
         * number of bytes per second used for transmitting the
         * encoded data. Check out #BearWare.CELTConstants for
         * bitrate limits. */
        public int nBitRate;
        /** @brief Milliseconds of audio data in each
         * packet. Recommended is 40 ms. Min is 20, max is 1000. */
        public int nMSecPerPacket;
    }

    /** @brief CELT constants */
    public struct CELTConstants
    {
        /** @brief The minimum bitrate supported for CELT. */
        public const int CELT_MIN_BITRATE = 35000;
        /** @brief The maximum bitrate supported for CELT. */
        public const int CELT_MAX_BITRATE = 3000000;
    }

    /** 
     * @brief Theora video codec settings.
     *
     * Configuring the Theora codec so it fits an end-user's domain
     * can be trickly. A high resolution in #BearWare.CaptureFormat will result
     * in much high bandwidth usage. Also using a high frame-rate will
     * cause much higher bandwidth usage.
     *
     * Note that width and height for a video frame encoded by Theora
     * must be a multiple of 16. If it is not then part of the picture
     * will be "cut" out. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct TheoraCodec
    {
        /** @brief The target bitrate in bits per second. If set to
         * zero the Theora codec will use variable bitrate (VBR). If
         * non-zero it will use constant bitrate (CBR). */
        public int nTargetBitRate;
        /** @brief A value from 0 - 63. Higher value gives higher
         * quality. */
        public int nQuality;
    }

    /**
     * @brief The codecs supported.
     * @see AudioCodec
     * @see VideoCodec */
    public enum Codec : uint
    {
        /** @brief No codec specified. */
        NO_CODEC                    = 0,
        /** @brief Speex audio codec, http://www.speex.org @see
         * SpeexCodec */
        SPEEX_CODEC                 = 1,
        /** @brief CELT audio codec version 0.5.2 used in version 4.1 and 
         * prior, http://www.celt-codec.org @see CELTCodec */
        CELT_0_5_2_OBSOLETE_CODEC   = 2,
        /** @brief Theora video codec, http://www.theora.org @see
         * TheoraCodec */
        THEORA_CODEC                = 3,
        /** @brief Speex audio codec in VBR mode, http://www.speex.org
         * @see SpeexVBRCodec */
        SPEEX_VBR_CODEC             = 4,
        /** @brief CELT audio codec version 0.11.1, 
         * http://www.celt-codec.org @see CELTCodec */
        CELT_CODEC                  = 5,
        /** @brief CELT audio codec version 0.11.1 in VBR mode, 
         * http://www.celt-codec.org @see CELTVBRCodec */
        CELT_VBR_CODEC              = 6
    }

    /**
     * @brief Struct used for specifying which audio codec a channel
     * uses. */
    [StructLayout(LayoutKind.Explicit)]
    public struct AudioCodec
    {
        /** @brief Specifies whether the member @a speex or @a celt holds
         * the codec settings. */
        [FieldOffset(0)]
        public Codec nCodec;
        /** @brief Speex codec settings if @a nCodec is
         * #Codec ::SPEEX_CODEC. */
        [FieldOffset(4)]
        public SpeexCodec speex;
        /** @brief CELT codec settings if @a nCodec is
         * #Codec ::CELT_CODEC */
        [FieldOffset(4)]
        public CELTCodec celt;
        /** @brief Speex codec settings if @a nCodec is
         * #Codec ::SPEEX_VBR_CODEC */
        [FieldOffset(4)]
        public SpeexVBRCodec speex_vbr;
        /** @brief CELT codec settings if @a nCodec is
         * #Codec ::CELT_VBR_CODEC */
        [FieldOffset(4)]
        public CELTVBRCodec celt_vbr;
    }

    /** @brief Common audio configuration which should be used by users
     * in the same #BearWare.Channel.
     *
     * Users' audio levels may be diffent due to how their microphone
     * is configured in their OS. Automatic Gain Control (AGC) can be used
     * to ensure all users in the same channel have the same audio level.
     *
     * @see TeamTalk4.DoMakeChannel()
     * @see TeamTalk4.DoUpdateChannel()
     * @see TeamTalk4.EnableAGC() */
    [StructLayout(LayoutKind.Explicit)]
    public struct AudioConfig
    {
        /** @brief Whether clients who join a #BearWare.Channel should
         * automatically enable AGC with the settings specified @a
         * bGainLevel, @a @a nMaxIncDBSec, @a nMaxDecDBSec and @a
         * nMaxGainDB. If the local client instance has already
         * enabled the flag #ClientFlag ::CLIENT_SNDINPUT_AGC it will
         * not enable AGC automatically when joining the channel. */
        [FieldOffset(0)]
        public bool bEnableAGC;
        /** @brief A value from 0 to 32000. Default is 8000. */
        [FieldOffset(4)]
        public int nGainLevel;
        /** @brief Used so volume should not be amplified too quickly 
         * (maximal gain increase in dB/second). Default is 12. */
        [FieldOffset(8)]
        public int nMaxIncDBSec;
        /** @brief Negative value! Used so volume should not be attenuated
         * too quickly (maximal gain decrease in dB/second). Default is 
         * -40. */
        [FieldOffset(12)]
        public int nMaxDecDBSec;
        /** @brief Ensure volume doesn't become too loud (maximal gain
         * in dB). Default is 30. */
        [FieldOffset(16)]
        public int nMaxGainDB;
        /** @brief Whether clients who join the channel should automatically
         * enable denoising. If the local client instance has already
         * enabled the flag #ClientFlag ::CLIENT_SNDINPUT_DENOISING it will not
         * enable denoising automatically when joining a channel. */
        [FieldOffset(20)]
        public bool bEnableDenoise;
        /** @brief Negative value! Maximum attenuation of the noise in dB.
         * Default value is -30. */
        [FieldOffset(24)]
        public int nMaxNoiseSuppressDB;
    }

    /** @brief Struct used for specifying the video codec to use. */
    [StructLayout(LayoutKind.Explicit)]
    public struct VideoCodec
    {
        /** @brief Specifies member holds the codec settings. So far
         * there is only one video codec to choose from, namely @a
         * theora. */
        [FieldOffset(0)]
        public Codec nCodec;
        /** @brief Theora codec settings if @a nCodec is
         * #Codec ::THEORA_CODEC. */
        [FieldOffset(4)]
        public TheoraCodec theora;
    }
    /** @} */

    /** @addtogroup transmission
     * @{ */

    /**
     * @brief Enum specifying data transmission types. @see
     * TeamTalk4.EnableTransmission @see TeamTalk4.IsTransmitting */
    [Flags]
    public enum TransmitType : uint
    {
        /** @brief Transmitting nothing. */
        TRANSMIT_NONE   = 0x0,
        /** @brief Transmit audio. */
        TRANSMIT_AUDIO  = 0x1,
        /** @brief Transmit video. */
        TRANSMIT_VIDEO  = 0x2
    }
    /** @} */

    /** @addtogroup server
     * @{ */

    /** 
     * @brief The rights users have once they have logged on to the
     * server.
     *
     * #BearWare.ServerProperties holds the user rights in its \a uUserRights
     * member variable and is retrieved by calling
     * TeamTalk4.GetServerProperties() once connected to the server.
     *
     * @see BearWare.ServerProperties
     * @see BearWare.TeamTalk4.GetServerProperties */
    [Flags]
    public enum UserRight : uint
    {
        /** @brief Users who log onto the server has none of the
          * rights below. */
        USERRIGHT_NONE                  = 0x00000000,
        /** @brief Users can log on without an account and by only
         * specifying the server password. */
        USERRIGHT_GUEST_LOGIN           = 0x00000001,
        /** @brief Users can see users in all other channels. This
         * option cannot be changed in a call to TeamTalk4.DoUpdateServer(). */
        USERRIGHT_VIEW_ALL_USERS        = 0x00000002,
        /** @brief Users are allowed to create channels. */
        USERRIGHT_CHANNEL_CREATION      = 0x00000004,
        /** @brief Users can become operators of channels. */
        USERRIGHT_CHANNEL_OPERATORS     = 0x00000008,
        /** @brief Users can use channel commands (text messages
         * prefixed with '/'. */
        USERRIGHT_CHANNEL_COMMANDS      = 0x00000010,
        /** @brief None-admins are allowed to broadcast messages to
         * all users. */
        USERRIGHT_CLIENT_BROADCAST      = 0x00000020,
        /** @brief Users are allowed to change subscriptions to other
         * users. */
        USERRIGHT_SUBSCRIPTIONS         = 0x00000040,
        /** @brief Users are allowed to forward audio packets through
         * server. */
        USERRIGHT_FORWARD_AUDIO         = 0x00000080,
        /** @brief Users are allowed to forward video packets through
         * server. */
        USERRIGHT_FORWARD_VIDEO         = 0x00000100,
        /** @brief Allow multiple users to log on to the server with
         * the same #BearWare.UserAccount. */
        USERRIGHT_DOUBLE_LOGIN          = 0x00000200,
        /** @brief Users are allowed to forward desktop packets through
         * server.
         * @note Requires server version 4.3.0.1490 or later.*/
        USERRIGHT_FORWARD_DESKTOP       = 0x00000400,
        /** @brief Users are only allowed to use valid UTF-8 strings.
         * If a non-UTF-8 string is passed in a command the server will
         * respond with the command error #ClientError ::CMDERR_SYNTAX_ERROR.
         * @note Requires server version 4.3.1.1940 or later. */
        USERRIGHT_STRICT_UTF8           = 0x00000800,
        /** @brief Users are allowed to forward desktop input packets through
         * server. Requires server version 4.6.0.2551 or later.*/
        USERRIGHT_FORWARD_DESKTOPINPUT  = 0x00001000,
    }

    /** 
     * @brief A struct containing the properties of the server's
     * settings.
     *
     * The server properties is available after a successful call to
     * TeamTalk4.DoLogin()
     *
     * @see TeamTalk4.DoUpdateServer
     * @see TeamTalk4.GetServerProperties 
     * @see TeamTalk4.DoLogin
     * @see UserRight */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct ServerProperties
    {
        /** @brief The server's name. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szServerName;
        /** @brief The server's password to login. Users must provide
         * this in the TeamTalk4.DoLogin() command.  When extracted
         * through TeamTalk4.GetServerProperties() the password will
         * only be set for users of user-type
         * #UserType ::USERTYPE_ADMIN. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szServerPasswd;
        /** @brief The message of the day. When updating the MOTD an admin
         * can use the variables %users% (number of users), %admins% (number
         * of admins), %uptime% (hours, minutes and seconds the server has
         * been online), %voicetx% (KBytes transmitted), %voicerx% (KBytes
         * received) and %lastuser% (nickname of last user to log on to the
         * server) as part of the MOTD. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szMOTD;
        /** @brief The message of the day including variables. This property
         * is only set for #UserType ::USERTYPE_ADMIN users. Read-only property. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szMOTDRaw;
        /** @brief A bitmask based on #UserRight which specifies the rights 
         * a user have who logs onto the server. */
        public UserRight uUserRights;
        /** @brief The maximum number of users allowed on the server.  A user
         * with admin account can ignore this */
        public int nMaxUsers;
        /** @brief The maximum number of logins with wrong password before
         * banning user's IP-address. */
        public int nMaxLoginAttempts;
        /** @brief The maximum number of users allowed to log in with the same
         * IP-address. 0 means disabled. */
        public int nMaxLoginsPerIPAddress;
        /** @brief Bandwidth restriction for audio codecs created by 
         * non-administrators. This value will hold the highest bitrate which 
         * is allowed for audio codecs. 0 = no limit. @see AudioCodec */
        public int nAudioCodecBpsLimit;
        /** @brief The maximum number of bytes per second which the server 
         * will allow for audio packets. If this value is exceeded the server
         * will start dropping audio packets. 0 = disabled. */
        public int nMaxAudioTxPerSecond;
        /** @brief The maximum number of bytes per second which the server 
         * will allow for audio packets. If this value is exceeded the server
         * will start dropping video packets. 0 = disabled. */
        public int nMaxVideoTxPerSecond;
        /** @brief The maximum number of bytes per second which the server 
         * will allow for desktop packets. If this value is exceeded the server
         * will start dropping desktop packets. 0 = disabled. */
        public int nMaxDesktopTxPerSecond;
        /** @brief The amount of bytes per second which the server 
         * will allow for packet forwarding.  If this value is exceeded the server
         * will start dropping packets. 0 = disabled. */
        public int nMaxTotalTxPerSecond;
        /** @brief Whether the server automatically saves changes */
        public bool bAutoSave;
        /** @brief The server's TCP port. */
        public int nTcpPort;
        /** @brief The server's UDP port. */
        public int nUdpPort;
        /** @brief The number of seconds before a user who hasn't
         * responded to keepalives will be kicked off the server. @see
         * TeamTalk4.SetKeepAliveInterval. */
        public int nUserTimeout;
        /** @brief The server version. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szServerVersion;
        /** @brief The version of the server's protocol. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szServerProtocolVersion;
    }

    /**
     * @brief A struct containing the server's statistics,
     * i.e. bandwidth usage and user activity.
     *
     * Use BearWare.TeamTalk4.DoQueryServerStats() to query the server's statistics
     * and when the command completes use BearWare.TeamTalk4.GetServerStatistics() to
     * extract the statistics. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct ServerStatistics
    {
        /** @brief The number of bytes sent from the server to
         * clients. */
        public long nTotalBytesTX;
        /** @brief The number of bytes received by the server from
         * clients. */
        public long nTotalBytesRX;
        /** @brief The number of bytes in audio packets sent from the
         *  server to clients. */
        public long nVoiceBytesTX;
        /** @brief The number of bytes in audio packets received by
         *  the server from clients. */
        public long nVoiceBytesRX;
        /** @brief The number of bytes in video packets sent from the
         *  server to clients. */
        public long nVideoBytesTX;
        /** @brief The number of bytes in video packets received by
         *  the server from clients. */
        public long nVideoBytesRX;
        /** @brief The number of bytes in desktop packets sent from the
         *  server to clients. */
        public long nDesktopBytesTX;
        /** @brief The number of bytes in desktop packets received by
         *  the server from clients. */
        public long nDesktopBytesRX;
        /** @brief The server's uptime in msec. */
        public long nUptimeMSec;
    }

    /**
     * @brief A struct containing the properties of a banned user.
     * This struct is used by TeamTalk4.GetBannedUsers(). */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct BannedUser
    {
        /** @brief IP-address of banned user. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szIpAddress;
        /** @brief Channel where user was located when banned. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szChannelPath;
        /** @brief Date and time when user was banned. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szBanTime;
        /** @brief Nickname of banned user. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szNickname;
        /** @brief Username of banned user. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szUsername;
    }

    /** @ingroup users
     * @brief The types of users supported. 
     * @see User */
    [Flags]
    public enum UserType : uint
    {
        /** @brief Used internally to denote an unauthenticated
         * user. */
        USERTYPE_NONE       = 0x0,
        /** @brief A default user who can join channels. */
        USERTYPE_DEFAULT    = 0x01,
        /** @brief A user with administrator privileges. */
        USERTYPE_ADMIN      = 0x02
    }

    /** 
     * @brief A struct containing the properties of a user account.
     *
     * A registered user is one that has a user account on the server.
     *
     * @see TeamTalk4.DoListUserAccounts
     * @see TeamTalk4.DoNewUserAccount
     * @see TeamTalk4.DoDeleteUserAccount */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct UserAccount
    {
        /** @brief The account's username */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szUsername;
        /** @brief The account's password. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szPassword;
        /** @brief A bitmask of the type of user based on #UserType. */
        public UserType uUserType;
        /** @brief A user data field which can be used for additional
         * information. The @a nUserData field of the #BearWare.User struct will
         * contain this value when a user who logs in with this account. */
        public int nUserData;
        /** @brief Additional notes about this user. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szNote;
        /** @brief User should (manually) join this channel after login.
         * If an initial channel is specified in the user's account then
         * no password is required for the user to join the channel.
         * @see TeamTalk4.DoJoinChannel() */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szInitChannel;
        /** @brief Channels where this user will automatically become channel
         * operator when joining. @see TeamTalk4.DoChannelOp() */
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = TeamTalk4.TT_CHANNELS_OPERATOR_MAX)]
        public int[] autoOperatorChannels;
    }
    /** @} */

    /** @addtogroup users
     * @{ */

    /** 
     * @brief A user by default accepts audio, video and text messages
     * from all users. Using subscribtions can, however, change what
     * the local client instance is willing to accept from other
     * users.
     *
     * By calling TeamTalk4.DoSubscribe() and TeamTalk4.DoUnsubscribe() the local
     * client instance can tell the server (and thereby remote users)
     * what he is willing to accept from other users.
     *
     * To check what a user subscribes to check out the #BearWare.User struct's
     * @a uLocalSubscriptions. The subscriptions with the prefix
     * SUBSCRIBE_INTERCEPT_* options can be used to spy on users and
     * receive data from them even if one is not participating in the
     * same channel as they are.
     *
     * @see TeamTalk4.DoSubscribe
     * @see TeamTalk4.DoUnsubscribe */
    [Flags]
    public enum Subscription : uint
    {
        /** @brief No subscriptions. */
        SUBSCRIBE_NONE                      = 0x0000,
        /** @brief Subscribing to user messages. */
        SUBSCRIBE_USER_MSG                  = 0x0001,
        /** @brief Subscribing to channel messages. */
        SUBSCRIBE_CHANNEL_MSG               = 0x0002,
        /** @brief Subscribing to broadcast messsages. */
        SUBSCRIBE_BROADCAST_MSG             = 0x0004,
        /** @brief Subscribing to audio. */
        SUBSCRIBE_AUDIO                     = 0x0008,
        /** @brief Subscribing to video. */
        SUBSCRIBE_VIDEO                     = 0x0010,
        /** @brief Subscribing to desktop sharing. */
        SUBSCRIBE_DESKTOP                   = 0x0020,
        /** @brief Subscribing to custom user messages. 
         * @see #TextMsgType ::MSGTYPE_CUSTOM. */
        SUBSCRIBE_CUSTOM_MSG                = 0x0040,
        /** @brief Subscribing to desktop input.
         * @see TeamTalk4.GetUserDesktopInput()
         * @see TeamTalk4.SendDesktopInput() */
        SUBSCRIBE_DESKTOPINPUT              = 0x0080,
        /** @brief Intercept all user text messages sent by a
        * user. Only user-type #UserType ::USERTYPE_ADMIN can do this. */
        SUBSCRIBE_INTERCEPT_USER_MSG        = 0x0100,
        /** @brief Intercept all channel messages sent by a user. Only
        * user-type #UserType ::USERTYPE_ADMIN can do this. */
        SUBSCRIBE_INTERCEPT_CHANNEL_MSG     = 0x0200,
        /** SUBSCRIBE_INTERCEPT_BROADCAST_MSG = 0x0400, */
        /** @brief Intercept all audio sent by a user. Only user-type
         * #UserType ::USERTYPE_ADMIN can do this. By enabling this subscription an
         * administrator can listen to audio sent by users outside his
         * own channel. */
        SUBSCRIBE_INTERCEPT_AUDIO           = 0x0800,
        /** @brief Intercept all video sent by a user. Only user-type
         * #UserType ::USERTYPE_ADMIN can do this. By enabling this subscription an
         * administrator can receive video frames sent by users
         * outside his own channel. */
        SUBSCRIBE_INTERCEPT_VIDEO           = 0x1000,
        /** @brief Intercept all desktop data sent by a user. Only
         * user-type #UserType ::USERTYPE_ADMIN can do this. By enabling this
         * subscription an administrator can views desktops
         * sent by users outside his own channel. */
        SUBSCRIBE_INTERCEPT_DESKTOP         = 0x2000,
        /** @brief Intercept all custom text messages sent by user. 
         * Only user-type #UserType ::USERTYPE_ADMIN can do this.  */
        SUBSCRIBE_INTERCEPT_CUSTOM_MSG      = 0x4000
    }

    /**
     * @brief The possible states for a user. Used for #BearWare.User's @a
     * uUserState variable. */
    [Flags]
    public enum UserState : uint
    {
        /** @brief The user is in initial state. */
        USERSTATE_NONE          = 0x00,
        /** @brief If set the user is currently talking. Basically
         * means the event TeamTalk4.OnUserTalking() has been posted for
         * this user. This flag doesn't apply to "myself". For
         * "myself" (client instance) use TeamTalk4.IsTransmitting() */
        USERSTATE_TALKING       = 0x01,
        /** @brief If set the user is muted. @see TeamTalk4.SetUserMute */
        USERSTATE_MUTE          = 0x02,
        /** @brief Whether a peer to peer connection has been
         * established to this user.  @see OnConnectionP2P */
        USERSTATE_P2P_CONNECTED = 0x04,
        /** @brief If set the user currently has an active desktop session
         * @see TeamTalk4.SendDesktopWindow(). */
        USERSTATE_DESKTOP       = 0x08,
        /** @brief If set the user currently has an active video stream
         * @see TeamTalk4.GetUserVideoFrame(). */
        USERSTATE_VIDEO         = 0x10,
    }

    /**
     * @brief A struct containing the properties of a user.
     * @see BearWare.UserType
     * @see BearWare.TeamTalk4.GetUser */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct User
    {
        /** @brief The user's ID. A value from 1 - 65535. */
        public int nUserID;
        /** @brief The user's nickname. @see BearWare.TeamTalk4.DoChangeNickname */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szNickname;
        /** @brief Username of user's account. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szUsername;
        /** @brief The user's currently status mode. @see
         * BearWare.TeamTalk4.DoChangeStatusMode */
        public int nStatusMode;
        /** @brief The user's current status message. @see
         * BearWare.TeamTalk4.DoChangeStatusMode */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szStatusMsg;
        /** @brief The channel the user is currently participating
         * in. 0 if none. */
        public int nChannelID;
        /** @brief The user's IP-address. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szIPAddress;
        /** @brief The user's client version. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szVersion;
        /** @brief A bitmask of the type of user based on #UserType. */
        public UserType uUserType;
        /** @brief A bitmask of the user's current state, e.g. talking, muted,
         * etc. */
        public UserState uUserState;
        /** @brief A bitmask of what the local user subscribes to from
         * this user.  @see BearWare.TeamTalk4.DoSubscribe */
        public Subscription uLocalSubscriptions;
        /** @brief A bitmask of what this user subscribes to from local
         * user.  @see BearWare.TeamTalk4.DoSubscribe */
        public Subscription uPeerSubscriptions;
        /** @brief The @a nUserData of the #BearWare.UserAccount used
         * by the user to log in. This field can be use to denote
         * e.g. a database ID. */
        public int nUserData;
        /** @brief Store audio received from this user to this
         * folder. @see TeamTalk4.SetUserAudioFolder */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szAudioFolder;
        /** @brief The #AudioFileFormat used for audio files. 
         * @see TeamTalk4.SetUserAudioFolder() */
        public AudioFileFormat uAFF;
        /** @brief Name of audio file currently being recorded. 
         * @see TeamTalk4.SetUserAudioFolder() */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szAudioFileName;
    }

    /**
     * @brief Packet reception and data statistics for a user.
     * 
     * @see BearWare.TeamTalk4.GetUserStatistics */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct UserStatistics
    {
        /** @brief Number of audio packets received from user. */
        public long nAudioPacketsRecv;
        /** @brief Number of audio packets lost from user. */
        public long nAudioPacketsLost;
        /** @brief Number of video packets received from user. A video 
         * frame can consist of several video packets. */
        public long nVideoPacketsRecv;
        /** @brief Number of video frames received from user. */
        public long nVideoFramesRecv;
        /** @brief Video frames which couldn't be shown because packets were
         * lost. */
        public long nVideoFramesLost;
        /** @brief Number of video frames dropped because user application  
         * didn't retrieve video frames in time. */
        public long nVideoFramesDropped;
    }

    /** 
     * @brief Text message types.
     * 
     * The types of messages which can be passed to TeamTalk4.DoTextMessage()().
     *
     * @see BearWare.TextMessage
     * @see BearWare.TeamTalk4.DoTextMessage
     * @see BearWare.TeamTalk4.GetTextMessage */
    public enum TextMsgType : uint
    {
        /** @brief A User to user text message. A message of this
         * type can be sent across channels. */
        MSGTYPE_USER        = 1,
        /** @brief A User to channel text message. Users of type
         * ::USERTYPE_DEFAULT can only send this text message to the
         * channel they're participating in, whereas users of type
         * ::USERTYPE_ADMIN can send to any channel. */
        MSGTYPE_CHANNEL     = 2,
        /** @brief A broadcast message. Only admins can send
         * broadcast messages unless #UserRight ::USERRIGHT_CLIENT_BROADCAST is
         * enabled. */
        MSGTYPE_BROADCAST   = 3,
        /** @brief A custom user to user text message. Works the same
         * way as #TextMsgType ::MSGTYPE_USER. */
        MSGTYPE_CUSTOM = 4
    }

    /** 
     * @brief A struct containing the properties of a text message
     * sent by a user.
     *
     * @see OnCmdUserTextMessage
     * @see TeamTalk4.DoTextMessage
     * @see TeamTalk4.GetTextMessage */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct TextMessage
    {
        /** @brief The type of text message */
        public TextMsgType nMsgType;
        /** @brief Will be set automatically on outgoing message. */
        public int nFromUserID;
        /** @brief The originators username. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szFromUsername;
        /** @brief Set to zero if channel message. */
        public int nToUserID;
        /** @brief Set to zero if @a nMsgType is #TextMsgType ::MSGTYPE_USER or
         * #TextMsgType ::MSGTYPE_BROADCAST. */
        public int nChannelID;
        /** @brief The actual text message. The message can be
         * multi-line (include EOL)  */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szMessage;
    }
    /** @} */

    /** @addtogroup channels
     * @{ */

    /**
     * @brief The types of channels supported. @see Channel */
    [Flags]
    public enum ChannelType : uint
    {
        /** @brief A default channel is a channel which disappears
         * after the last user leaves the channel. */
        CHANNEL_DEFAULT             = 0x0000,
        /** @brief A channel which persists even when the last user
         * leaves the channel. */
        CHANNEL_STATIC              = 0x0001,
        /** @brief Only one user can transmit at a time. Note that
         * this option doesn't apply if clients are running in peer to
         * peer mode. @see TeamTalk4.EnablePeerToPeer() */
        CHANNEL_SOLO_TRANSMIT       = 0x0002,
        /** @brief Audio transmitted to the channel by the client
         * instance is also sent back and played by the client
         * instance. */
        CHANNEL_ECHO                = 0x0004,
        /** @brief Same as #ChannelType ::CHANNEL_ECHO. */
        CHANNEL_ECHO_AUDIO          = 0x0004,
        /** @brief Voice and video transmission in the channel is controlled 
         * by a channel operator (or an administrator). For a user to
         * transmit audio or video to this type of channel the channel operator must
         * add the user's ID to either @a voiceUsers or @a videoUsers in the
         * #BearWare.Channel struct and call TeamTalk4.DoUpdateChannel().
         *
         * @note
         * Requires server version 4.1.0.994 or later.
         *
         * @see TeamTalk4.IsChannelOperator()
         * @see #UserType ::USERTYPE_ADMIN */
        CHANNEL_CLASSROOM           = 0x0008,
        /** @brief Video sent to the channel should also be sent back
         * to the local client instance. */
        CHANNEL_ECHO_VIDEO          = 0x0010,
        /** @brief Desktop session sent to the channel should also be
         * sent back to the local client instance. */
        CHANNEL_ECHO_DESKTOP        = 0x0020,
        /** @brief Only channel operators (and administrators) will receive 
         * audio/video/desktop transmissions. Default channel users 
         * will only see transmissions from operators and/or 
         * administrators. */
        CHANNEL_OPERATOR_RECVONLY   = 0x0040
    }

    /**
     * @brief A struct containing the properties of a channel.
     * @see TeamTalk4.GetChannel
     * @see ChannelType
     * @see AudioCodec */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct Channel
    {
        /** @brief Parent channel ID. 0 means no parent channel,
         * i.e. it's the root channel. */
        public int nParentID;
        /** @brief The channel's ID. A value from 1 - 65535. */
        public int nChannelID;
        /** @brief Name of the channel. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szName;
        /** @brief Topic of the channel. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szTopic;
        /** @brief Password to join the channel.  When extracted
         * through TeamTalk4.GetChannel() the password will only be set for
         * users of user-type #UserType ::USERTYPE_ADMIN. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szPassword;
        /** @brief Whether password is required to join channel */
        public bool bPassword;
        /** @brief A bitmask of the type of channel based on #ChannelType. */
        public ChannelType uChannelType;
        /** @brief Number of bytes available for file storage. */
        public long nDiskQuota;
        /** @brief Operator password, i.e. for use with '/opme'
         * command. @see UserRight ::USERRIGHT_CHANNEL_COMMANDS */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szOpPassword;
        /** @brief Max number of users in channel. */
        public int nMaxUsers;
        /** @brief The audio codec used by users in the channel. */
        public AudioCodec codec;
        /** @brief The audio configuration which users who join the channel
         * should use. If the audio configuration forces some options, e.g. 
         * AGC and denoising then these options will automatically be enabled
         * and override what is currently set by TeamTalk4.EnableDenoising() and 
         * TeamTalk4.EnableAGC().
         *
         * @note
         * Requires server and client version 4.1.0.1127 or later. */
        public AudioConfig audiocfg;
        /** @brief The IDs of users who are allowed to transmit voice
         * data to the channel. This setting only applies to channels
         * of type #ChannelType ::CHANNEL_CLASSROOM. Only channel
         * operators and administrators are allowed to change the
         * users who are allowed to transmit data to a channel. Call
         * BearWare.TeamTalk4.DoUpdateChannel() to update the list of users who are
         * allowed to transmit data to the channel.
         *
         * @note
         * Requires server version 4.1.0.994 or later.
         *
         * @see TeamTalk4.IsChannelOperator
         * @see TeamTalk4.DoChannelOp
         * @see TransmitType.TRANSMIT_AUDIO
         * @see TeamTalk.TT_CLASSROOM_FREEFORALL */
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = TeamTalk4.TT_VOICEUSERS_MAX)]
        public int[] voiceUsers;
        /** @brief The IDs of users who are allowed to transmit video
         * data to the channel. This setting only applies to channels
         * of type #ChannelType ::CHANNEL_CLASSROOM. Only channel
         * operators and administrators are allowed to change the
         * users who are allowed to transmit data to a channel. Call
         * BearWare.TeamTalk4.DoUpdateChannel() to update the list of
         * users who are allowed to transmit data to the channel.
         *
         * @note
         * Requires server version 4.1.0.994 or later.
         *
         * @see TeamTalk4.IsChannelOperator
         * @see TeamTalk4.DoChannelOp
         * @see TransmitType.TRANSMIT_VIDEO
         * @see TeamTalk.TT_CLASSROOM_FREEFORALL */
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = TeamTalk4.TT_VIDEOUSERS_MAX)]
        public int[] videoUsers;
        /** @brief The IDs of users who are allowed to share their
         * desktop to the channel. This setting only applies to
         * channels of type #ChannelType ::CHANNEL_CLASSROOM. Only channel operators
         * and administrators are allowed to change the users who are
         * allowed to transmit data to a channel. Call
         * TeamTalk4.DoUpdateChannel() to update the list of users who are
         * allowed to transmit data to the channel.
         *
         * @note
         * Requires server version 4.3.0.1490 or later.
         *
         * @see TeamTalk4.IsChannelOperator
         * @see TeamTalk4.DoChannelOp
         * @see TeamTalk4.SendDesktopWindow()
         * @see TeamTalk.TT_CLASSROOM_FREEFORALL */
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = TeamTalk4.TT_DESKTOPUSERS_MAX)]
        public int[] desktopUsers;
    }

    /** 
     * @brief A struct containing the properties of a file transfer.
     * @see TeamTalk4.GetFileTransferInfo */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct FileTransfer
    {
        /** @brief The ID identifying the file transfer. */
        public int nTransferID;
        /** @brief The channel where the file is/will be located. */
        public int nChannelID;
        /** @brief The file path on local disk. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szLocalFilePath;
        /** @brief The filename in the channel. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szRemoteFileName;
        /** @brief The size of the file being transferred. */
        public long nFileSize;
        /** @brief The number of bytes transferred so far. */
        public long nTransferred;
        /** @brief TRUE if download and FALSE if upload. */
        public bool bInbound;
    }

    /** @brief Status of a file transfer.
     * @see OnFileTransfer */
    public enum FileTransferStatus : uint
    {
        /** @brief Error during file transfer. */
        FILETRANSFER_ERROR      = 0,
        /** @brief File transfer started. */
        FILETRANSFER_STARTED    = 1,
        /** @brief File transfer finished. */
        FILETRANSFER_FINISHED   = 2
    }

    /**
     * @brief A struct containing the properties of a file in a #BearWare.Channel.
     * @see TeamTalk4.GetChannelFileInfo */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct FileInfo
    {
        /** @brief The ID identifying the file. */
        public int nFileID;
        /** @brief The name of the file. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szFileName;
        /** @brief The size of the file. */
        public long nFileSize;
        /** @brief Username of the person who uploaded the files. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szUsername;
    }
    /** @} */

    /** @ingroup connectivity
     * @brief Statistics of bandwidth usage and ping times in the local 
     * client instance.
     * @see TeamTalk4.GetStatistics */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct ClientStatistics
    {
        /** @brief Bytes sent on UDP. */
        public long nUdpBytesSent;
        /** @brief Bytes received on UDP. */
        public long nUdpBytesRecv;
        /** @brief Voice data sent (on UDP). */
        public long nVoiceBytesSent;
        /** @brief Voice data received (on UDP). */
        public long nVoiceBytesRecv;
        /** @brief Video data sent (on UDP). */
        public long nVideoBytesSent;
        /** @brief Video data received (on UDP). */
        public long nVideoBytesRecv;
        /** @brief Desktop data sent (on UDP). */
        public long nDesktopBytesSent;
        /** @brief Desktop data received (on UDP). */
        public long nDesktopBytesRecv;
        /** @brief Response time to server on UDP (based on ping/pong
         * sent at a specified interval. Set to -1 if not currently
         * available. @see TeamTalk4.SetKeepAliveInterval */
        public int nUdpPingTimeMs;
        /** @brief Response time to server on TCP (based on ping/pong
         * sent at a specified interval. Set to -1 if not currently
         * available.  @see TeamTalk4.SetKeepAliveInterval */
        public int nTcpPingTimeMs;
        /** @brief The number of seconds nothing has been received by
         * the client on TCP. */
        public int nTcpServerSilenceSec;
        /** @brief The number of seconds nothing has been received by
         * the client on UDP. */
        public int nUdpServerSilenceSec;
    }

    /** @addtogroup errorhandling
     * @{ */

    /**
     * @brief Errors which can occur either as a result of client
     * commands or as a result of internal errors.
     *
     * Use TeamTalk4.GetErrorMessage() to get a text-description of the
     * error. */
    public enum ClientError : uint
    {
        /** @brief Command indicating success. Only used internally. */
        SUCCESS = 0,

        /* COMMAND ERRORS 1000-1999 ARE DUE TO INVALID OR UNSUPPORTED
         * COMMANDS */

        /** @brief Command has syntax error. Only used internally. */
        CMDERR_SYNTAX_ERROR                     = 1000,
        /** @brief The server doesn't support the issued command.
         *
         * This error may occur if the server is an older version than
         * the client instance. */
        CMDERR_UNKNOWN_COMMAND                  = 1001,
        /** @brief Command cannot be performed due to missing
         * parameter. Only used internally. */
        CMDERR_MISSING_PARAMETER                = 1002,
        /** @brief The server uses a protocol which is incompatible
         * with the client instance. */
        CMDERR_INCOMPATIBLE_PROTOCOLS           = 1003,
        /** @brief The server does not support the audio codec specified
         * by the client. Introduced in version 4.1.0.1264. 
         * @see TeamTalk4.DoMakeChannel()
         * @see TeamTalk4.DoJoinChannel() */
        CMDERR_UNKNOWN_AUDIOCODEC               = 1004,

        /* COMMAND ERRORS 2000-2999 ARE DUE TO INSUFFICIENT RIGHTS */

        /** @brief Invalid server password. 
         *
         * The TeamTalk4.DoLogin() command passed a server password which was
         * invalid.  @see TeamTalk4.DoLogin */
        CMDERR_INCORRECT_SERVER_PASSWORD        = 2000,
        /** @brief Invalid channel password. 
         *
         * The TeamTalk4.DoJoinChannel() or TeamTalk4.DoJoinChannel()ByID passed an
         * invalid channel password. TeamTalk4.DoMakeChannel() can also cause
         * a this error if the password is longer than #BearWare.TeamTalk4.TT_STRLEN. */
        CMDERR_INCORRECT_CHANNEL_PASSWORD       = 2001,
        /** @brief Invalid username or password for account.
         *
         * The TeamTalk4.DoLogin() command was issued with invalid account
         * properties. This error can also occur by
         * TeamTalk4.DoNewUserAccount() if username is empty. */
        CMDERR_INVALID_ACCOUNT                  = 2002,
        /** @brief Login failed due to maximum number of users on
         * server.
         *
         * TeamTalk4.DoLogin() failed because the server does not allow any
         * more users. */
        CMDERR_MAX_SERVER_USERS_EXCEEDED        = 2003,
        /** @brief Cannot join channel because it has maximum number
         * of users.
         *
         * TeamTalk4.DoJoinChannel() or TeamTalk4.DoJoinChannel()ByID failed because
         * no more users are allowed in the channel. */
        CMDERR_MAX_CHANNEL_USERS_EXCEEDED       = 2004,
        /** @brief IP-address has been banned from server.
         *
         * TeamTalk4.DoLogin() failed because the local client's IP-address
         * has been banned on the server. */
        CMDERR_SERVER_BANNED                    = 2005,
        /** @brief Command not authorized.
         *
         * The command cannot be performed because the client instance
         * has insufficient rights.
         *
         * @see TeamTalk4.DoDeleteFile
         * @see TeamTalk4.DoJoinChannel
         * @see TeamTalk4.DoJoinChannelByID
         * @see TeamTalk4.DoLeaveChannel
         * @see TeamTalk4.DoChannelOp
         * @see TeamTalk4.DoChannelOpEx
         * @see TeamTalk4.DoKickUser
         * @see TeamTalk4.DoUpdateChannel
         * @see TeamTalk4.DoChangeNickname
         * @see TeamTalk4.DoChangeStatus
         * @see TeamTalk4.DoTextMessage
         * @see TeamTalk4.DoSubscribe
         * @see TeamTalk4.DoUnsubscribe
         * @see TeamTalk4.DoMakeChannel
         * @see TeamTalk4.DoRemoveChannel
         * @see TeamTalk4.DoMoveUser
         * @see TeamTalk4.DoUpdateServer
         * @see TeamTalk4.DoSaveConfig
         * @see TeamTalk4.DoSendFile 
         * @see TeamTalk4.DoRecvFile 
         * @see TeamTalk4.DoBanUser
         * @see TeamTalk4.DoUnBanUser
         * @see TeamTalk4.DoListBans
         * @see TeamTalk4.DoListUserAccounts
         * @see TeamTalk4.DoNewUserAccount
         * @see TeamTalk4.DoDeleteUserAccount */
        CMDERR_NOT_AUTHORIZED                   = 2006,
        /** @brief Server doesn't allow users to create channels.
         *
         * TeamTalk4.DoJoinChannel() or TeamTalk4.DoJoinChannel()ByID failed because
         * #UserRight ::USERRIGHT_CHANNEL_CREATION is not enabled.  
         * @see ServerProperties */
        CMDERR_CANNOT_CREATE_CHANNELS           = 2007,
        /** @brief Cannot upload file because disk quota will be exceeded.
         *
         * TeamTalk4.DoSendFile() was not allowed because there's not enough
         * disk space available for upload.
         *
         * @see Channel */
        CMDERR_MAX_DISKUSAGE_EXCEEDED           = 2008,
        /** @brief Modifying subscriptions not enabled.
         *
         * TeamTalk4.DoSubscribe() or TeamTalk4.DoUnsubscribe() failed because the
         * server does not allow users to change subscriptions. The
         * #UserRight ::USERRIGHT_SUBSCRIPTIONS is not enabled in the server's
         * #BearWare.ServerProperties.
         *
         * @see ServerProperties */
        CMDERR_SUBSCRIPTIONS_DISABLED           = 2009,

        /** @brief Invalid password for becoming channel operator.
         * 
         * The password specified in TeamTalk4.DoChannelOpEx() is not correct.
         * The operator password is the @a szOpPassword of the 
         * #BearWare.Channel-struct. */
        CMDERR_INCORRECT_OP_PASSWORD            = 2010,

        /** @brief The selected #BearWare.AudioCodec exceeds what the
         * server allows.
         *
         * A server can limit the vitrate of audio codecs if @c 
         * nAudioCodecBpsLimit of #BearWare.ServerProperties is specified. */
        CMDERR_AUDIOCODEC_BITRATE_LIMIT_EXCEEDED = 2011,

        /** @brief The maximum number of logins allowed per IP-address has
         * been exceeded.
         * 
         * @see ServerProperties
         * @see TeamTalk4.DoLogin() */
        CMDERR_MAX_LOGINS_PER_IPADDRESS_EXCEEDED = 2012,


        /* COMMAND ERRORS 3000-3999 ARE DUE TO INVALID STATE OF CLIENT INSTANCE */

        /** @brief Client instance has not been authenticated.
         * 
         * TeamTalk4.DoLogin() has not been issued successfully or
         * TeamTalk4.DoLogout() could not be performed because client
         * instance is already logged in.*/
        CMDERR_NOT_LOGGEDIN                     = 3000,

        /** @brief Already logged in.
         *
         * TeamTalk4.DoLogin() cannot be performed twice. */
        CMDERR_ALREADY_LOGGEDIN                 = 3001,
        /** @brief Cannot leave channel because not in channel.
         *
         * TeamTalk4.DoLeaveChannel() failed because user is not in a channel. */
        CMDERR_NOT_IN_CHANNEL                   = 3002,
        /** @brief Cannot join same channel twice.
         * 
         * TeamTalk4.DoJoinChannel() or TeamTalk4.DoJoinChannel()ByID failed because
         * client instance is already in the specified channel. */
        CMDERR_ALREADY_IN_CHANNEL               = 3003,
        /** @brief Channel already exists.
         *
         * TeamTalk4.DoMakeChannel() failed because channel already exists. */
        CMDERR_CHANNEL_ALREADY_EXISTS           = 3004,
        /** @brief Channel does not exist.
         *
         * Command failed because channel does not exists.
         * @see TeamTalk4.DoRemoveChannel
         * @see TeamTalk4.DoUpdateChannel
         * @see TeamTalk4.DoMakeChannel Due to invalid channel name
         * @see TeamTalk4.DoSendFile
         * @see TeamTalk4.DoRecvFile
         * @see TeamTalk4.DoDeleteFile
         * @see TeamTalk4.DoJoinChannel
         * @see TeamTalk4.DoJoinChannelByID
         * @see TeamTalk4.DoLeaveChannel
         * @see TeamTalk4.DoChannelOp
         * @see TeamTalk4.DoKickUser
         * @see TeamTalk4.DoBanUser
         * @see TeamTalk4.DoMoveUser
         * @see TeamTalk4.DoTextMessage */
        CMDERR_CHANNEL_NOT_FOUND                = 3005,
        /** @brief User not found.
         * 
         * Command failed because user does not exists.
         * @see TeamTalk4.DoChannelOp
         * @see TeamTalk4.DoKickUser
         * @see TeamTalk4.DoBanUser
         * @see TeamTalk4.DoMoveUser
         * @see TeamTalk4.DoTextMessage
         * @see TeamTalk4.DoSubscribe
         * @see TeamTalk4.DoUnsubscribe */
        CMDERR_USER_NOT_FOUND                   = 3006,
        /** @brief Banned IP-address does not exist.
         * 
         * TeamTalk4.DoUnBanUser() failed because there is no banned
         * IP-address which matches what was specified. */
        CMDERR_BAN_NOT_FOUND                    = 3007,
        /** @brief File transfer doesn't exists.
         *
         * TeamTalk4.DoSendFile() or TeamTalk4.DoRecvFile() failed because the server
         * cannot process the file transfer. */
        CMDERR_FILETRANSFER_NOT_FOUND           = 3008,
        /** @brief Server failed to open file.
         *
         * TeamTalk4.DoSendFile() or TeamTalk4.DoRecvFile() failed because the server
         * cannot open the specified file (possible file lock). */
        CMDERR_OPENFILE_FAILED                  = 3009,
        /** @brief Cannot find user account.
         * 
         * TeamTalk4.DoDeleteUserAccount() failed because the specified user
         * account does not exists. */
        CMDERR_ACCOUNT_NOT_FOUND                = 3010,
        /** @brief File does not exist.
         *
         * TeamTalk4.DoSendFile(), TeamTalk4.DoRecvFile() or TeamTalk4.DoDeleteFile() failed
         * because the server cannot find the specified file. */
        CMDERR_FILE_NOT_FOUND                   = 3011,
        /** @brief File already exist.
         *
         * TeamTalk4.DoSendFile() failed because the file already exists in
         * the channel. */
        CMDERR_FILE_ALREADY_EXISTS              = 3012,
        /** @brief Server does not allow file transfers.
         *
         * TeamTalk4.DoSendFile() or TeamTalk4.DoRecvFile() failed because the server
         * does not allow file transfers. */
        CMDERR_FILESHARING_DISABLED             = 3013,

        /** @brief Cannot process command since server is not empty.
         * @see TeamTalk4.DoUpdateServer() Ensure not to update 
         * #UserRight ::USERRIGHT_VIEW_ALL_USERS. */
        CMDERR_SERVER_HAS_USERS                 = 3014,
        /** @brief Cannot process command since channel is not empty.
         * @see TeamTalk4.DoUpdateChannel() #BearWare.AudioCodec cannot be changed while
         * there are users in a channel. */
        CMDERR_CHANNEL_HAS_USERS                = 3015,

        /* ERRORS 10000-10999 ARE NOT COMMAND ERRORS BUT INSTEAD
         * ERRORS IN THE CLIENT INSTANCE. */

        /** @brief A sound input device failed. 
         *
         * This can e.g. happen when joining a channel and the sound
         * device has been unplugged. */
        INTERR_SNDINPUT_FAILURE                 = 10000,
        /** @brief A sound output device failed.
         *
         * This can e.g. happen when joining a channel and the sound
         * device has been unplugged. */
        INTERR_SNDOUTPUT_FAILURE                = 10001,
    }
    /** @} */

    /** @addtogroup initclient
     * @{ */

    /**
     * @brief Flags used to describe the the client instance current
     * state.
     *
     * The client's state is a bitmask of the flags in #ClientFlag.
     *
     * The state of the client instance can be retrieved by calling
     * TeamTalk4.GetFlags() This enables the user application to display the
     * possible options to the end user. If e.g. the flag
     * #ClientFlag ::CLIENT_AUTHORIZED is not set it will not be possible to
     * perform any other commands except TeamTalk4.DoLogin(). Doing so will
     * make the server return an error message to the client. */
    [Flags]
    public enum ClientFlag : uint
    {
        /** @brief If no bits are set the client instance is in closed state,
         * in other words no operations has been performed on
         * it. */
        CLIENT_CLOSED                               = 0x00000000,
        /** @brief If set the client instance's sound input device has
         * been initialized, i.e. TeamTalk4.InitSoundInputDevice() has been
         * called successfully. */
        CLIENT_SNDINPUT_READY                       = 0x00000001,
        /** @brief If set the client instance's sound output device
         * has been initialized, i.e. TeamTalk4.InitSoundOutputDevice() has
         * been called successfully. */
        CLIENT_SNDOUTPUT_READY                      = 0x00000002,
        /** @brief If set the client instance's video device has been
         * initialized, i.e. TeamTalk4.InitVideoCaptureDevice() has been
         * called successfuly. */
        CLIENT_VIDEO_READY                          = 0x00000004,
        /** @brief If set the client instance current have an active
         * desktop session, i.e. TeamTalk4.SendDesktopWindow() has been
         * called. Call TeamTalk4.CloseDesktopWindow() to close the desktop
         * session. */
        CLIENT_DESKTOP_ACTIVE                       = 0x00000008,
        /** @brief If set the client instance will start transmitting
         * audio if the sound level is above the voice activation
         * level. The event TeamTalk4.OnVoiceActivation() is posted
         * when voice activation initiates transmission.
         * @see TeamTalk4.SetVoiceActivationLevel
         * @see TeamTalk4.EnableVoiceActivation */
        CLIENT_SNDINPUT_VOICEACTIVATED              = 0x00000010,
        /** @brief If set the client instance will try to remove noise
         * from recorded audio. @see TeamTalk4.EnableDenoising */
        CLIENT_SNDINPUT_DENOISING                   = 0x00000020,
        /** @brief If set the client instance is using automatic gain 
         * control. @see TeamTalk4.EnableAGC */
        CLIENT_SNDINPUT_AGC                         = 0x00000040,
        /** @brief If set the client instance has muted all users.
        * @see TeamTalk4.SetSoundOutputMute */
        CLIENT_SNDOUTPUT_MUTE                       = 0x00000080,
        /** @brief If set the client instance will auto position users
        * in a 180 degree circle using 3D-sound. This option is only
        * available with #SoundSystem ::SOUNDSYSTEM_DSOUND.
        * @see TeamTalk4.SetUserPosition 
        * @see TeamTalk4.Enable3DSoundPositioning */
        CLIENT_SNDOUTPUT_AUTO3DPOSITION             = 0x00000100,
        /** @brief If set the client instance will try to eliminate
         * echo from speakers. To enable echo cancellation first make
         * the client run on sound duplex mode by calling
         * TeamTalk4.InitSoundDuplexDevices() and afterwards call
         * TeamTalk4.EnableEchoCancellation(). */
        CLIENT_SNDINPUT_AEC                         = 0x00000200,
        /** @brief If set the client instance is running in sound
         * duplex mode where multiple audio output streams are mixed
         * into a single stream. This option must be enabled to
         * support echo cancellation (#ClientFlag ::CLIENT_SNDINPUT_AEC). Call
         * TeamTalk4.InitSoundDuplexDevices() to enable duplex mode.*/
        CLIENT_SNDINOUTPUT_DUPLEX                   = 0x00000400,
        /** @brief If set the client instance is currently transmitting
         * audio.  @see TeamTalk4.EnableTransmission */
        CLIENT_TX_AUDIO                             = 0x00001000,
        /** @brief If set the client instance is currently transmitting video.
         * @see TeamTalk4.EnableTransmission */
        CLIENT_TX_VIDEO                             = 0x00002000,
        /** @brief If set the client instance is currently muxing
         * audio streams into a single file. This is enabled by calling
         * TeamTalk4.StartRecordingMuxedAudioFile(). */
        CLIENT_MUX_AUDIOFILE                        = 0x00004000,
        /** @brief If set the client instance is currently transmitting
         * a desktop window. A desktop window update is issued by calling
         * TeamTalk4.SendDesktopWindow(). The event 
         * OnDesktopTransferUpdate() is triggered when a desktop
         * window transmission completes. */
        CLIENT_TX_DESKTOP                           = 0x00008000,
        /** @brief If set the client instance is currently try to
         * connect to a server, i.e. TeamTalk4.Connect() has been called. */
        CLIENT_CONNECTING                           = 0x00010000,
        /** @brief If set the client instance is connected to a server,
         * i.e. TeamTalk4.OnConnectSuccess() event has been issued after
         * doing a TeamTalk4.Connect(). Valid commands in this state:
         * TeamTalk4.DoLogin() */
        CLIENT_CONNECTED                            = 0x00020000,
        /** @brief Helper for #ClientFlag ::CLIENT_CONNECTING and 
         * #ClientFlag ::CLIENT_CONNECTED to see if 
         * TeamTalk4.Disconnect() should be called. */
        CLIENT_CONNECTION = CLIENT_CONNECTING | CLIENT_CONNECTED,
        /** @brief If set the client instance is logged on to a
         * server, i.e. got TeamTalk4.OnCmdMyselfLoggedIn() event
         * after issueing TeamTalk4.DoLogin(). */
        CLIENT_AUTHORIZED                           = 0x00040000,
        /** @brief If set the client instance will try and connect to
         * other users using peer to peer connections. Audio will be
         * broadcast to users instead of forwarded through server
         * (thereby increasing the bandwith usage).  Note that if
         * #UserRight ::USERRIGHT_FORWARD_AUDIO is disabled and no peer to peer
         * connection could be established, i.e. event
         * TeamTalk4.OnConnectionP2P() was posted with failure, then data
         * cannot be transferred to a user. */
        CLIENT_P2P_AUDIO                            = 0x00100000,
        /** @brief If set the client instance will try and connect to
         * other users using peer to peer connections. Video will be
         * broadcast to users instead of forwarded through server
         * (thereby increasing the bandwith usage).  Note that if
         * #UserRight ::USERRIGHT_FORWARD_VIDEO is disabled and no peer to peer
         * connection could be established, i.e. event
         * TeamTalk4.OnConnectionP2P() was posted with failure, then data
         * cannot be transferred to a user. */
        CLIENT_P2P_VIDEO                            = 0x00200000,
        /** @brief Helper for #ClientFlag ::CLIENT_P2P_AUDIO and #ClientFlag ::CLIENT_P2P_VIDEO to see
         * if the client instance is currently attempting P2P connections. */
        CLIENT_P2P = CLIENT_P2P_AUDIO | CLIENT_P2P_VIDEO,
        /** @brief If set the client is currently streaming the audio
         * of a media file. When streaming a video file the
         * #CLIENT_STREAM_VIDEO flag is also typically set.
         * @see TeamTalk4.StartStreamingMediaFileToChannel() */
        CLIENT_STREAM_AUDIO                         = 0x00400000,
        /** @brief If set the client is currently streaming the video
         *of a media file. When streaming a video file the
         *#CLIENT_STREAM_AUDIO flag is also typically set.
         * @see TeamTalk4.StartStreamingMediaFileToChannel() */
        CLIENT_STREAM_VIDEO                         = 0x00800000
    }

    /** @} */

    /** @ingroup initclient
     * @brief The class which encapsulates the TeamTalk 4 client. */
    public class TeamTalk4 : IDisposable
    {
        /** @addtogroup initclient
         * @{ */

        /** @brief The maximum length of all strings used in
            TeamTalk. This value includes the zero terminator, so @b
            511 characters. Note that Unicode characters are converted
            to UTF-8 internally and limited once again to 512 if the
            string gets longer. */
        public const int TT_STRLEN = 512;

        /** @ingroup videocapture
         * The maximum number of video formats which will be queried for a 
         * #BearWare.VideoCaptureDevice. */
        public const int TT_CAPTUREFORMATS_MAX = 128;

        /** @ingroup channels 
         * The maximum number of users allowed to
         * transmit audio when a #BearWare.Channel is configured with
         * #ChannelType ::CHANNEL_CLASSROOM. */
        public const int TT_VOICEUSERS_MAX = 16;

        /** @ingroup channels
         * The maximum number of users allowed to transmit video when a #BearWare.Channel is configured
         * with #ChannelType ::CHANNEL_CLASSROOM. */
        public const int TT_VIDEOUSERS_MAX = 16;

        /** @ingroup channels
         * The maximum number of users allowed to transmit when a #BearWare.Channel
         * is configured with #ChannelType ::CHANNEL_CLASSROOM. */
        public const int TT_DESKTOPUSERS_MAX = 16;

        /** @ingroup channels
         * If a #BearWare.Channel is configured with #ChannelType ::CHANNEL_CLASSROOM then only users
         * certain user IDs are allowed to transmit. If, however, @c 
         * TT_CLASSROOM_FREEFORALL is put in either @c voiceUsers, @c videoUsers 
         * and @c desktopUsers then everyone in the channel are allowed to 
         * transmit. */
        public const int TT_CLASSROOM_FREEFORALL = 0xFFFF;

        /** @ingroup users
         * The maximum number of channels where a user can automatically become
         * channel operator.
         * @see #BearWare.UserAccount */
        public const int TT_CHANNELS_OPERATOR_MAX = 16;

        /** @ingroup sounddevices
         * The maximum number of sample rates supported by a #BearWare.SoundDevice. */
        public const int TT_SAMPLERATES_MAX = 16;

        /** @ingroup desktopshare
         *
         * The maximum number #BearWare.DesktopInput instances which can be sent by
         * BearWare.TeamTalk4.SendDesktopInput or received by #BearWare.TeamTalk4.GetUserDesktopInput */
        public const int TT_DESKTOPINPUT_MAX = 16;

        /** @ingroup desktopshare
         *
         * If @c uKeyCode in #BearWare.DesktopInput is set to
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_IGNORE it means no key (or mouse button)
         * was pressed in the desktop input event and
         * TT_DesktopInput_Execute() will ignore the value. */
        public const uint TT_DESKTOPINPUT_KEYCODE_IGNORE = 0xFFFFFFFF;

        /** @ingroup desktopshare
         *
         * If @c uMousePosX or @c uMousePosY in #BearWare.DesktopInput are set to
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_MOUSEPOS_IGNORE it means the mouse position is
         * ignored when calling WindowsHelper.Execute(). */
        public const ushort TT_DESKTOPINPUT_MOUSEPOS_IGNORE = 0xFFFF;

        /** @ingroup desktopshare
         *
         * If @c uKeyCode of #BearWare.DesktopInput is set to
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_LMOUSEBTN then WindowsDesktopInput.Execute()
         * will see the key-code as a left mouse button click. */
        public const uint TT_DESKTOPINPUT_KEYCODE_LMOUSEBTN = 0x1000;

        /** @ingroup desktopshare
         *
         * If @c uKeyCode of #BearWare.DesktopInput is set to
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_RMOUSEBTN then WindowsDesktopInput.Execute()
         * will see the key-code as a right mouse button click. */
        public const uint TT_DESKTOPINPUT_KEYCODE_RMOUSEBTN = 0x1001;

        /** @ingroup desktopshare
         *
         * If @c uKeyCode of #BearWare.DesktopInput is set to
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_MMOUSEBTN then WindowsDesktopInput.Execute()
         * will see the key-code as a middle mouse button click. */
        public const uint TT_DESKTOPINPUT_KEYCODE_MMOUSEBTN = 0x1002;


#if ENABLE_ENCRYPTION
        public const int DEFAULT_TCPPORT = 10443;
        public const int DEFAULT_UDPPORT = 10443;
#else
        public const int DEFAULT_TCPPORT = 10333;
        public const int DEFAULT_UDPPORT = 10333;
#endif
        public const int DEFAULT_AGC_GAINLEVEL      = 8000;
        public const int DEFAULT_AGC_INC_MAXDB      = 12;
        public const int DEFAULT_AGC_DEC_MAXDB      = -40;
        public const int DEFAULT_AGC_GAINMAXDB      = 30;
        public const int DEFAULT_DENOISE_SUPPRESS   = -30;

        /** 
         * @brief Create a new TeamTalk client instance.
         *
         * @param poll_based If the application using this class is a
         * Windows Forms application @a poll_based should be 'false'
         * since events will be posted using the Windows message
         * loop. In Console applications on the other hand, the user
         * application will have to 'poll' for events using
         * GetMessage(). Remember to put the @c [STAThread] macro on
         * the @c Main method if building a console application. */
        public TeamTalk4(bool poll_based)
        {
            Assembly assembly = Assembly.GetExecutingAssembly();
            AssemblyName name = assembly.GetName();
            Version dllversion = new Version(Marshal.PtrToStringUni((c_tt.TT4.TT_GetVersion())));
            if (!name.Version.Equals(dllversion))
            {
                string errmsg = String.Format("Invalid {2} version loaded. {2} is version {0} and {3} is version {1}",
                    dllversion.ToString(), name.Version.ToString(), c_tt.TT4.dllname, c_tt.TT4.mgtdllname);
                throw new Exception(errmsg);
            }

            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__AUDIOCODEC) == Marshal.SizeOf(new AudioCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__BANNEDUSER) == Marshal.SizeOf(new BannedUser()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__CAPTUREFORMAT) == Marshal.SizeOf(new CaptureFormat()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__CELTCODEC) == Marshal.SizeOf(new CELTCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__CHANNEL) == Marshal.SizeOf(new Channel()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__CLIENTSTATISTICS) == Marshal.SizeOf(new ClientStatistics()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__FILEINFO) == Marshal.SizeOf(new FileInfo()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__FILETRANSFER) == Marshal.SizeOf(new FileTransfer()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__SERVERPROPERTIES) == Marshal.SizeOf(new ServerProperties()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__SERVERSTATISTICS) == Marshal.SizeOf(new ServerStatistics()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__SOUNDDEVICE) == Marshal.SizeOf(new SoundDevice()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__SPEEXCODEC) == Marshal.SizeOf(new SpeexCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__TEXTMESSAGE) == Marshal.SizeOf(new TextMessage()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__THEORACODEC) == Marshal.SizeOf(new TheoraCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__TTMESSAGE) == Marshal.SizeOf(new TTMessage()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__USER) == Marshal.SizeOf(new User()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__USERACCOUNT) == Marshal.SizeOf(new UserAccount()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__USERSTATISTICS) == Marshal.SizeOf(new UserStatistics()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__VIDEOCAPTUREDEVICE) == Marshal.SizeOf(new VideoCaptureDevice()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__VIDEOCODEC) == Marshal.SizeOf(new VideoCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__AUDIOCONFIG) == Marshal.SizeOf(new AudioConfig()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__CELTVBRCODEC) == Marshal.SizeOf(new CELTVBRCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__SPEEXVBRCODEC) == Marshal.SizeOf(new SpeexVBRCodec()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__VIDEOFRAME) == Marshal.SizeOf(new VideoFrame()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__AUDIOBLOCK) == Marshal.SizeOf(new AudioBlock()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__AUDIOFORMAT) == Marshal.SizeOf(new AudioFormat()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__MEDIAFILEINFO) == Marshal.SizeOf(new MediaFileInfo()));
            Debug.Assert(TT4.TT_DBG_SIZEOF(TTStructType.__DESKTOPINPUT) == Marshal.SizeOf(new DesktopInput()));

            Debug.Assert(TT4.TT_DBG_EVENT_LAST() == ClientEvent.WM_TEAMTALK_DUMMY_LAST);

            if (poll_based)
                m_ttInst = TT4.TT_InitTeamTalkPoll();
            else
            {
                m_eventhandler = new MyEventHandler(this);
#if PocketPC
                IntPtr hWnd = m_eventhandler.Hwnd;
#else
                m_eventhandler.CreateControl();
                IntPtr hWnd = m_eventhandler.Handle;
#endif
                m_ttInst = TT4.TT_InitTeamTalk(hWnd);
            }
        }

        /** 
         * @brief Close the TeamTalk client instance and release its
         * resources.
         */
        ~TeamTalk4()
        {
            DeleteMe();
        }


        #region IDisposable Members

        public void Dispose()
        {
            DeleteMe();
        }

        private void DeleteMe()
        {
            if (m_ttInst != IntPtr.Zero)
            {
                TT4.TT_CloseTeamTalk(m_ttInst);
                m_ttInst = IntPtr.Zero;
            }
        }

        #endregion

        /**
         * @brief Poll for events in the client instance.
         * 
         * In a Windows Forms application events can be processed using
         * the application message-loop, but in Console application the
         * application must process the events itself. The most common
         * way to do this is to start a timer which calls #GetMessage
         * every now and then to ensure the UI is updated with the latest
         * events.
         *
         * @param nWaitMs The amount of time to wait for the event. If -1 the 
         * function will block forever or until the next event occurs.
         * @return Returns TRUE if an event has occured otherwise FALSE.
         * @see TeamTalk4()
         * @see ClientEvent */
        public bool GetMessage(int nWaitMs)
        {
            TTMessage msg;
            if (TT4.TT_GetMessage(m_ttInst, out msg, ref nWaitMs))
            {
                processMsg(ref msg);
                return true;
            }
            return false;
        }

        /**
         * @brief Get a bitmask describing the client's current state.
         *
         * Checks whether the client is connecting, connected, authorized,
         * etc. The current state can be checked by and'ing the returned
         * bitmask which is based on #ClientFlag.
         *
         * @return A bitmask describing the current state. */
        public BearWare.ClientFlag GetFlags()
        {
            return TT4.TT_GetFlags(m_ttInst);
        }

        /**
         * @brief Same as GetFlags().
         */
        public ClientFlag Flags
        {
            get { return GetFlags(); }
        }

        /**
         * @brief Set license information to disable trial mode.
         *
         * This function must be called before instantiating the TeamTalk4-class.
         *
         * @param szRegName The registration name provided by BearWare.dk.
         * @param nRegKey The registration key provided by BearWare.dk.
         * @return True if the provided registration is acceptable. */
        public static bool SetLicenseInformation(string szRegName, long nRegKey)
        {
            return TT4.TT_SetLicenseInformation(szRegName, nRegKey);
        }

        public void processMsg(ref TTMessage msg)
        {
            switch (msg.wmMsg)
            {
                case ClientEvent.WM_TEAMTALK_CON_SUCCESS:
                    if(OnConnectSuccess != null)
                        OnConnectSuccess();
                    break;
                case ClientEvent.WM_TEAMTALK_CON_FAILED:
                    if (OnConnectFailed != null)
                    OnConnectFailed();
                    break;
                case ClientEvent.WM_TEAMTALK_CON_LOST:
                    if (OnConnectionLost != null)
                        OnConnectionLost();
                    break;
                case ClientEvent.WM_TEAMTALK_CON_P2P:
                    if (OnConnectionP2P != null)
                        OnConnectionP2P((int)msg.wParam, msg.lParam != 0);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_PROCESSING:
                    if (OnCmdProcessing != null)
                        OnCmdProcessing((int)msg.wParam, msg.lParam != 0);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_MYSELF_LOGGEDIN:
                    if (OnCmdMyselfLoggedIn != null)
                        OnCmdMyselfLoggedIn((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_MYSELF_LOGGEDOUT:
                    if (OnCmdMyselfLoggedOut != null)
                        OnCmdMyselfLoggedOut();
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_MYSELF_JOINED:
                    if (OnCmdMyselfJoinedChannel != null)
                        OnCmdMyselfJoinedChannel((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_MYSELF_LEFT:
                    if (OnCmdMyselfLeftChannel != null)
                        OnCmdMyselfLeftChannel((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_MYSELF_KICKED:
                    if (OnCmdMyselfKickedFromChannel != null)
                        OnCmdMyselfKickedFromChannel((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_USER_LOGGEDIN:
                    if (OnCmdUserLoggedIn != null)
                        OnCmdUserLoggedIn((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_USER_LOGGEDOUT:
                    if (OnCmdUserLoggedOut != null)
                        OnCmdUserLoggedOut((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_USER_UPDATE:
                    if (OnCmdUserUpdate != null)
                        OnCmdUserUpdate((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_USER_JOINED:
                    if (OnCmdUserJoinedChannel != null)
                        OnCmdUserJoinedChannel((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_USER_LEFT:
                    if (OnCmdUserLeftChannel != null)
                        OnCmdUserLeftChannel((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_USER_TEXTMSG:
                    if (OnCmdUserTextMessage != null)
                        OnCmdUserTextMessage((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_CHANNEL_NEW:
                    if (OnCmdChannelNew != null)
                        OnCmdChannelNew((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_CHANNEL_UPDATE:
                    if (OnCmdChannelUpdate != null)
                        OnCmdChannelUpdate((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_CHANNEL_REMOVE:
                    if (OnCmdChannelRemove != null)
                        OnCmdChannelRemove((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_SERVER_UPDATE:
                    if (OnCmdServerUpdate != null)
                        OnCmdServerUpdate();
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_FILE_NEW:
                    if (OnCmdFileNew != null)
                        OnCmdFileNew((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_FILE_REMOVE:
                    if (OnCmdFileRemove != null)
                        OnCmdFileRemove((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_ERROR:
                    if (OnCmdError != null)
                        OnCmdError((ClientError)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CMD_SUCCESS :
                    if (OnCmdSuccess != null)
                        OnCmdSuccess((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_TALKING:
                    if (OnUserTalking != null)
                        OnUserTalking((int)msg.wParam, msg.lParam != 0);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_VIDEOFRAME:
                    if (OnUserVideoFrame != null)
                        OnUserVideoFrame((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_INTERNAL_ERROR:
                    if (OnInternalError != null)
                        OnInternalError((ClientError)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_VOICE_ACTIVATION:
                    if (OnVoiceActivation != null)
                        OnVoiceActivation(msg.wParam != 0);
                    break;
                case ClientEvent.WM_TEAMTALK_STREAM_AUDIOFILE_USER:
                    if (OnStreamAudioFileUser != null)
                        OnStreamAudioFileUser((int)msg.wParam, (AudioFileStatus)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_STREAM_AUDIOFILE_CHANNEL:
                    if (OnStreamAudioFileChannel != null)
                        OnStreamAudioFileChannel((int)msg.wParam, (AudioFileStatus)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_HOTKEY:
                    if (OnHotKeyToggle != null)
                        OnHotKeyToggle((int)msg.wParam, msg.lParam != 0);
                    break;
                case ClientEvent.WM_TEAMTALK_HOTKEY_TEST:
                    if (OnHotKeyTest != null)
                        OnHotKeyTest((int)msg.wParam, msg.lParam != 0);
                    break;
                case ClientEvent.WM_TEAMTALK_FILETRANSFER:
                    if (OnFileTransfer != null)
                        OnFileTransfer((int)msg.wParam, (FileTransferStatus)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_AUDIOFILE:
                    if (OnUserAudioFile != null)
                        OnUserAudioFile((int)msg.wParam, (AudioFileStatus)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_AUDIOBLOCK :
                    if (OnUserAudioBlock != null)
                        OnUserAudioBlock((int)msg.wParam);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_DESKTOPWINDOW :
                    if (OnUserDesktopWindow != null)
                        OnUserDesktopWindow((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_DESKTOPWINDOW_TRANSFER :
                    if (OnDesktopTransferUpdate != null)
                        OnDesktopTransferUpdate((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_DESKTOPCURSOR:
                    if (OnUserDesktopCursorPosition != null)
                        OnUserDesktopCursorPosition((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_CON_MAX_PAYLOAD_UPDATED :
                    if (OnMaxPayloadUpdate != null)
                        OnMaxPayloadUpdate((int)msg.wParam, (int)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_STREAM_MEDIAFILE_CHANNEL :
                    if (OnStreamMediaFileChannel != null)
                        OnStreamMediaFileChannel((AudioFileStatus)msg.lParam);
                    break;
                case ClientEvent.WM_TEAMTALK_USER_DESKTOPINPUT :
                    if (OnUserDesktopInput != null)
                        OnUserDesktopInput((int)msg.wParam);
                    break;
            }
        }

        /** @} */

        /** @addtogroup sounddevices
         * @{ */

        /**
         * @brief Get the default sound devices. 
         *
         * @param lpnInputDeviceID The ID of the default input device.
         * @param lpnOutputDeviceID The ID of the default output device.
         * @see TeamTalk4.InitSoundInputDevice
         * @see TeamTalk4.InitSoundOutputDevice */
        public bool GetDefaultSoundDevices(out int lpnInputDeviceID,
                                           out int lpnOutputDeviceID)
        {
            return TT4.TT_GetDefaultSoundDevices(m_ttInst, out lpnInputDeviceID, out lpnOutputDeviceID);
        }
        /**
         * @brief Get the default sound devices for the specified sound system.
         *
         * @see TT_GetDefaultSoundDevices() */
        public bool GetDefaultSoundDevicesEx(SoundSystem nSndSystem,
                                             out int lpnInputDeviceID,
                                             out int lpnOutputDeviceID)
        {
            return TT4.TT_GetDefaultSoundDevicesEx(nSndSystem, out lpnInputDeviceID, out lpnOutputDeviceID);
        }
        /**
         * @brief Get information about input devices for audio recording. 
         *
         * @param lpSoundDevices An output array of #BearWare.SoundDevice-structs which
         * will receive the available sound input devices.
         * @see TeamTalk4.GetDefaultSoundDevices
         * @see TeamTalk4.InitSoundInputDevice
         * @see TeamTalk4.InitSoundOutputDevice */
        public bool GetSoundInputDevices(out SoundDevice[] lpSoundDevices)
        {
            int count = 0;
            bool b = TT4.TT_GetSoundInputDevices_NULL(m_ttInst, IntPtr.Zero, ref count);
            SoundDevice[] devs = new SoundDevice[count];
            b = TT4.TT_GetSoundInputDevices(m_ttInst, devs, ref count);
            lpSoundDevices = b ? devs : null;
            return b;
        }
        /** 
         * @brief Get the list of sound output devices for playback. 
         *
         * The nDeviceID of the #BearWare.SoundDevice struct should be passed to
         * #InitSoundOutputDevice.
         *
         * @param lpSoundDevices An output array of #BearWare.SoundDevice-structs which
         * will receive the available sound output devices.
         * @see TeamTalk4.GetDefaultSoundDevices 
         * @see TeamTalk4.InitSoundOutputDevice */
        public bool GetSoundOutputDevices(out SoundDevice[] lpSoundDevices)
        {
            int count = 0;
            bool b = TT4.TT_GetSoundOutputDevices_NULL(m_ttInst, IntPtr.Zero, ref count);

            SoundDevice[] devs = new SoundDevice[count];
            b = TT4.TT_GetSoundOutputDevices(m_ttInst, devs, ref count);
            lpSoundDevices = b ? devs : null;
            return b;
        }

        /**
         * @brief Initialize the sound input devices (for recording audio).
         *
         * The @a nDeviceID of the #BearWare.SoundDevice should be used as @a 
         * nInputDeviceID.
         * 
         * Calling this function will set the flag #ClientFlag ::CLIENT_SNDINPUT_READY.
         *
         * @param nInputDeviceID The @a nDeviceID of #BearWare.SoundDevice extracted 
         * through #GetSoundInputDevices.
         * @see BearWare.SoundDevice
         * @see GetDefaultSoundDevices
         * @see GetSoundInputDevices
         * @see CloseSoundInputDevice
         * @see GetSoundInputLevel */
        public bool InitSoundInputDevice(int nInputDeviceID)
        {
            return TT4.TT_InitSoundInputDevice(m_ttInst, nInputDeviceID);
        }

        /** 
         * @brief Initialize the sound output devices (for sound playback).
         *
         * The @a nDeviceID of the #BearWare.SoundDevice should be used as @a 
         * nOutputDeviceID.
         *
         * Callling this function will set the flag
         * #ClientFlag ::CLIENT_SNDOUTPUT_READY.
         *
         * @param nOutputDeviceID Should be the @a nDeviceID of 
         * #BearWare.SoundDevice extracted through GetSoundOutputDevices().
         * @see BearWare.SoundDevice
         * @see TeamTalk4.GetDefaultSoundDevices
         * @see TeamTalk4.GetSoundOutputDevices
         * @see TeamTalk4.CloseSoundOutputDevice */
        public bool InitSoundOutputDevice(int nOutputDeviceID)
        {
            return TT4.TT_InitSoundOutputDevice(m_ttInst, nOutputDeviceID);
        }
        /**
         * @brief Enable duplex mode where multiple audio streams are
         * mixed into a single stream using software.
         *
         * Duplex mode can @b ONLY be enabled on sound devices which
         * support the same sample rate. Sound systems #SoundSystem ::SOUNDSYSTEM_WASAPI
         * and #SoundSystem ::SOUNDSYSTEM_ALSA typically only support a single sample
         * rate.  Check @c supportedSampleRates in #BearWare.SoundDevice to see
         * which sample rates are supported.
         *
         * Sound duplex mode is required for echo cancellation since sound
         * input and output device must be synchronized. Also sound cards
         * which does not support multiple output streams should use
         * duplex mode.
         *
         * If TeamTalk4.InitSoundDuplexDevices() is successful the following
         * flags will be set:
         *
         * - #ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX
         * - #ClientFlag ::CLIENT_SNDOUTPUT_READY
         * - #ClientFlag ::CLIENT_SNDINPUT_READY
         *
         * Sound duplex mode is required for echo cancellation since sound
         * input and output device must be synchronized. Also sound cards
         * which does not support multiple output stream should use
         * duplex mode.
         *
         * Call TeamTalk4.CloseSoundDuplexDevices() to shut down duplex mode.
         *
         * Note that it is only the audio streams from users in the local
         * client instance's current channel which will be mixed. If the
         * local client instance calls TeamTalk4.DoSubscribe() with
         * #Subscription ::SUBSCRIBE_INTERCEPT_AUDIO on a user in another channel then
         * the audio from this user will be started in a separate
         * stream. The reason for this is that the other user may use a
         * different audio codec.
         *
         * @param nInputDeviceID The @a nDeviceID of #BearWare.SoundDevice extracted 
         * through GetSoundInputDevices().
         * @param nOutputDeviceID The @a nDeviceID of #BearWare.SoundDevice extracted
         * through GetSoundOutputDevices().
         * @see TeamTalk4.InitSoundInputDevice()
         * @see TeamTalk4.InitSoundOutputDevice()
         * @see TeamTalk4.EnableEchoCancellation()
         * @see TeamTalk4.CloseSoundDuplexDevices() */
        public bool InitSoundDuplexDevices(int nInputDeviceID, int nOutputDeviceID)
        {
            return TT4.TT_InitSoundDuplexDevices(m_ttInst, nInputDeviceID, nOutputDeviceID);
        }
        /**
         * @brief Shutdown the input sound device.
         *
         * Callling this function will clear the flag
         * #ClientFlag ::CLIENT_SNDINPUT_READY.
         * 
         * If the local client instance is running in duplex mode (flag
         * #ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX is set) then trying to close the
         * sound device will fail since duplex mode require that both
         * input and output sound devices are active at the same
         * time. Therefore in order to close sound devices running in
         * duplex mode call TeamTalk4.CloseSoundDuplexDevices().
         * 
         * @return If running in sound duplex mode (#ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX)
         * then ensure to disable duplex mode prior to closing the sound device.
         * 
         * @see TeamTalk4.InitSoundInputDevice */
        public bool CloseSoundInputDevice()
        {
            return TT4.TT_CloseSoundInputDevice(m_ttInst);
        }
        /**
         * @brief Shutdown the output sound device.
         *
         * Callling this function will clear set the flag
         * #ClientFlag ::CLIENT_SNDOUTPUT_READY.
         *
         * If the local client instance is running in duplex mode (flag
         * #ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX is set) then trying to close the
         * sound device will fail since duplex mode require that both
         * input and output sound devices are active at the same
         * time. Therefore in order to close sound devices running in
         * duplex mode call TeamTalk4.CloseSoundDuplexDevices().
         *
         * @return If running in sound duplex mode (#ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX)
         * then ensure to disable duplex mode prior to closing the sound device.
         * 
         * @see TeamTalk4.InitSoundOutputDevice */
        public bool CloseSoundOutputDevice()
        {
            return TT4.TT_CloseSoundOutputDevice(m_ttInst);
        }
        /**
         * @brief Shut down sound devices running in duplex mode.
         *
         * Calling this function only applies if sound devices has been
         * initialized with TeamTalk4.InitSoundDuplexDevices().
         */
        public bool CloseSoundDuplexDevices()
        {
            return TT4.TT_CloseSoundDuplexDevices(m_ttInst);
        }
        /**
         * @brief Reinitialize sound system (in order to detect
         * new/removed devices).
         *
         * When the TeamTalk client is first initialized all the sound
         * devices are detected and stored in a list inside the client
         * instance. If a user adds or removes e.g. a USB sound device
         * then it's not picked up automatically by the client
         * instance. TT_RestartSoundSystem() can be used to reinitialize
         * the sound system and thereby detect if sound devices have been
         * removed or added.
         *
         * In order to restart the sound system all sound devices in all
         * client instances must be closed using TT_CloseSoundInputDevice(),
         * TT_CloseSoundoutputDevice() and TT_CloseSoundDuplexDevices(). */
        public static bool RestartSoundSystem()
        {
            return TT4.TT_RestartSoundSystem();
        }
        /**
         * @brief Perform a record and playback test of specified sound
         * devices.
         *
         * Call #StopSoundLoopbackTest() to stop the loopback test.
         *
         * @param nInputDeviceID Should be the @a nDeviceID extracted through 
         * #GetSoundInputDevices.
         * @param nOutputDeviceID Should be the @a nDeviceID extracted through 
         * #GetSoundOutputDevices.
         * @param nSampleRate The sample rate the client's recorder should 
         * use.
         * @param nChannels Number of channels to use, i.e. 1 = mono, 2 = stereo.
         * Note that echo cancellation, denoising and AGC is not supported in
         * stereo.
         * @see TeamTalk4.InitSoundInputDevice()
         * @see TeamTalk4.InitSoundOutputDevice()
         * @see TeamTalk4.StopSoundLoopbackTest() */
        public bool StartSoundLoopbackTest(int nInputDeviceID,
                                    int nOutputDeviceID,
                                    int nSampleRate,
                                    int nChannels)
        {
            return TT4.TT_StartSoundLoopbackTest(m_ttInst,
        nInputDeviceID, nOutputDeviceID, nSampleRate, nChannels);
        }
        /**
         * @brief Perform a record and playback test of specified sound
         * devices along with an audio configuration and ability to try
         * echo cancellation.
         *
         * Both input and output devices MUST support the specified sample 
         * rate since this loop back test uses duplex mode 
         * ( @see TeamTalk4.InitSoundDuplexDevices() ). Check out @c 
         * supportedSampleRates of #BearWare.SoundDevice to see which sample rates
         * are supported.
         *
         * Call TeamTalk4.StopSoundLoopbackTest() to stop the loopback
         * test.
         *
         * This function is almost like TeamTalk4.StartSoundLoopbackTest() except
         * that it allows the use of #BearWare.AudioConfig to enable AGC and echo
         * cancellation. Note that AGC and echo cancellation can only be
         * used in mono, i.e. @c nChannels = 1.
         *
         * @param nInputDeviceID Should be the @a nDeviceID extracted through 
         * TeamTalk4.GetSoundInputDevices().
         * @param nOutputDeviceID Should be the @a nDeviceID extracted through 
         * TeamTalk4.GetSoundOutputDevices().
         * @param nSampleRate The sample rate the client's recorder should 
         * use.
         * @param nChannels Number of channels to use, i.e. 1 = mono, 2 = stereo.
         * Note that echo cancellation, denoising and AGC is not supported in
         * stereo.
         * @param lpAudioConfig The audio configuration to use, i.e. AGC 
         * and denoising properties.
         * @param bEchoCancel Whether to enable echo cancellation.
         *
         * @see TeamTalk4.InitSoundInputDevice()
         * @see TeamTalk4.InitSoundOutputDevice()
         * @see TeamTalk4.InitSoundDuplexDevices()
         * @see TeamTalk4.StopSoundLoopbackTest() */
        public bool StartSoundLoopbackTestEx(int nInputDeviceID, int nOutputDeviceID,
                                      int nSampleRate, int nChannels,
                                      AudioConfig lpAudioConfig, bool bEchoCancel)
        {
            return TT4.TT_StartSoundLoopbackTestEx(m_ttInst, nInputDeviceID, nOutputDeviceID, 
                nSampleRate, nChannels, ref lpAudioConfig, bEchoCancel);
        }
        /**
         * @brief Stop recorder and playback test.
         *
         * @see TeamTalk4.InitSoundInputDevice
         * @see TeamTalk4.InitSoundOutputDevice
         * @see TeamTalk4.StartSoundLoopbackTest */
        public bool StopSoundLoopbackTest()
        {
            return TT4.TT_StopSoundLoopbackTest(m_ttInst);
        }
        /**
         * @brief Get the volume level of the current recorded audio.
         *
         * The current level is updated at an interval specified in a channel's
         * #BearWare.AudioCodec.
         *
         * Note that the volume level will not be available until the
         * client instance joins a channel, i.e. it knows what sample rate
         * to use.
         *
         * @return Returns a value between #BearWare.SoundLevel.SOUND_VU_MIN and #BearWare.SoundLevel.SOUND_VU_MAX */
        public int GetSoundInputLevel()
        {
            return TT4.TT_GetSoundInputLevel(m_ttInst);
        }
        /** 
         * @brief Set voice gaining of recorded audio. 
         *
         * The gain level ranges from #BearWare.SoundLevel.SOUND_GAIN_MIN to #BearWare.SoundLevel.SOUND_GAIN_MAX
         * where #BearWare.SoundLevel.SOUND_GAIN_DEFAULT is no gain. So 100 is 1/10 of the
         * original volume and 8000 is 8 times the original volume.
         *
         * @param nLevel A value from #BearWare.SoundLevel.SOUND_GAIN_MIN to #BearWare.SoundLevel.SOUND_GAIN_MAX.
         * @see TeamTalk4.GetSoundInputGainLevel */
        public bool SetSoundInputGainLevel(int nLevel)
        {
            return TT4.TT_SetSoundInputGainLevel(m_ttInst, nLevel);
        }
        /**
         * @brief Get voice gain level of outgoing audio
         *
         * @return A value from #BearWare.SoundLevel.SOUND_GAIN_MIN to #BearWare.SoundLevel.SOUND_GAIN_MAX.
         * @see SetSoundInputGainLevel */
        public int GetSoundInputGainLevel()
        {
            return TT4.TT_GetSoundInputGainLevel(m_ttInst);
        }
        /**
         * @brief Set master volume. 
         *
         * If still not loud enough use SetUserGainLevel().
         *
         * @param nVolume A value from #BearWare.SoundLevel.SOUND_VOLUME_MIN to  #BearWare.SoundLevel.SOUND_VOLUME_MAX.
         * @see TeamTalk4.SetUserVolume */
        public bool SetSoundOutputVolume(int nVolume)
        {
            return TT4.TT_SetSoundOutputVolume(m_ttInst, nVolume);
        }
        /**
         * @brief Get master volume.
         *
         * @return Returns the master volume.
         * @see BearWare.SoundLevel.SOUND_VOLUME_MAX
         * @see BearWare.SoundLevel.SOUND_VOLUME_MIN
         * @see TeamTalk4.GetUserVolume */
        public int GetSoundOutputVolume()
        {
            return TT4.TT_GetSoundOutputVolume(m_ttInst);
        }
        /**
         * @brief Set all users mute.
         *
         * To stop receiving audio from a user call DoUnsubscribe().
         *
         * @param bMuteAll Whether to mute or unmute all users.
         * @see ClientFlag ::CLIENT_SNDOUTPUT_MUTE */
        public bool SetSoundOutputMute(bool bMuteAll)
        {
            return TT4.TT_SetSoundOutputMute(m_ttInst, bMuteAll);
        }
        /**
         * @brief Enable denoising of recorded audio.
         *
         * This call will enable/disable the #ClientFlag ::CLIENT_SNDINPUT_DENOISING
         * flag. Denoising will not be active if the local client instance
         * is participating in a channel which uses a stereo #BearWare.AudioCodec.
         *
         * @param bEnable TRUE to enable, otherwise FALSE.
         * @see TeamTalk4.SetDenoiseLevel() */
        public bool EnableDenoising(bool bEnable)
        {
            return TT4.TT_EnableDenoising(m_ttInst, bEnable);
        }
        /**
         * @brief Set the denoise level of recorded audio.
         *
         * Setting denoise level is only valid if TeamTalk4.GetFlags()
         * contains #ClientFlag ::CLIENT_SNDINPUT_DENOISING.
         *
         * @param nLevel Maximum attenuation of the noise in dB. Negative value!
         * Default value is -15.
         * @see TeamTalk4.EnableDenoising() */
        public bool SetDenoiseLevel(int nLevel)
        {
            return TT4.TT_SetDenoiseLevel(m_ttInst, nLevel);
        }
        /**
         * @brief Set the denoise level of recorded audio.
         *
         * Getting denoise level is only valid if TeamTalk4.GetFlags() contains
         * #ClientFlag ::CLIENT_SNDINPUT_DENOISING.
         *
         * @return Maximum attenuation of the noise in dB. Negative value!
         * @see TeamTalk4.SetDenoiseLevel() */
        public int GetDenoiseLevel()
        {
            return TT4.TT_GetDenoiseLevel(m_ttInst);
        }
        /**
         * @brief Enable/disable acoustic echo cancellation (AEC).
         *
         * In order to enable echo cancellation mode the local client
         * instance must first be set in sound duplex mode by calling
         * TeamTalk4.InitSoundDuplexDevices(). This is because the echo canceller
         * must first mixed all audio streams into a single stream and
         * have then run in synch with the input stream. After calling
         * TeamTalk4.InitSoundDuplexDevices() the flag #ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX
         * will be set.
         *
         * For echo cancellation to work the sound input and output device
         * must be the same sound card since the input and output stream
         * must be completely synchronized. Also it is recommended to also
         * enable denoising and AGC for better echo cancellation by calling
         * TeamTalk4.EnableDenoising() and TeamTalk4.EnableAGC() respectively. Using an 
         * #BearWare.AudioConfig on a #BearWare.Channel can also be use to automatically
         * enable denoising and AGC.
         *
         * Echo cancellation will not be active if the local client
         * instance is participating in a channel which uses a stereo
         * #BearWare.AudioCodec.
         *
         * After calling TeamTalk4.EnableEchoCancellation() the flag
         * #ClientFlag ::CLIENT_SNDINPUT_AEC will be set.
         *
         * Note that Windows Mobile/CE doesn't support echo cancellation.
         *
         * @param bEnable Whether to enable or disable echo callation.
         * @see TeamTalk4.EnableDenoising()
         * @see TeamTalk4.EnableAGC() */
        public bool EnableEchoCancellation(bool bEnable)
        {
            return TT4.TT_EnableEchoCancellation(m_ttInst, bEnable);
        }
        /** 
         * @brief Enable Automatic Gain Control
         *
         * Each #BearWare.Channel can also be configured to automatically enable
         * the local client instance to enable AGC by specifying the @a
         * bEnableAGC member of #BearWare.AudioConfig.
         *
         * AGC will not be active if the local client instance is
         * participating in a channel which uses a stereo #BearWare.AudioCodec.
         *
         * Note that Windows Mobile/CE doesn't support AGC.
         *
         * @param bEnable Whether to enable AGC.
         * @see ClientFlag ::CLIENT_SNDINPUT_AGC
         * @see TeamTalk4.SetAGCSettings() */
        public bool EnableAGC(bool bEnable)
        {
            return TT4.TT_EnableAGC(m_ttInst, bEnable);
        }
        /** 
         * @brief Set Automatic Gain Control (AGC) settings.
         *
         * Since microphone volumes may vary AGC can be used to adjust a signal
         * to a reference volume. That way users will speak at the same volume
         * level.
         *
         * Calling TeamTalk4.SetAGCSettings() is only valid if
         * TeamTalk4.GetFlags() contains
         * #ClientFlag ::CLIENT_SNDINPUT_AGC.
         *
         * @param nGainLevel A value from 0 to 32768. Default is 8000.
         * @param nMaxIncrement Used so volume
         * should not be amplified too quickly (maximal gain increase in dB/second).
         * Default is 12.
         * @param nMaxDecrement Negative value! Used so volume
         * should not be attenuated too quickly (maximal gain decrease in
         * dB/second). Default is -40.
         * @param nMaxGain Ensure volume doesn't become too loud (maximal gain
         * in dB). Default is 30.
         * @see TeamTalk4.EnableAGC() */
        public bool SetAGCSettings(int nGainLevel,
                            int nMaxIncrement,
                            int nMaxDecrement,
                            int nMaxGain)
        {
            return TT4.TT_SetAGCSettings(m_ttInst, nGainLevel, nMaxIncrement,
                                   nMaxDecrement, nMaxGain);
        }
        /** 
         * @brief Get Automatic Gain Control settings.
         *
         * Calling TeamTalk4.GetAGCSettings() is only valid if
         * TeamTalk4.GetFlags() contains #ClientFlag ::CLIENT_SNDINPUT_AGC.
         *
         * @param lpnGainLevel A value from 0 to 32768.
         * @param lpnMaxIncrement Used so volume
         * should not be amplified too quickly (maximal gain increase in dB/second).
         * @param lpnMaxDecrement Negative value! Used so volume
         * should not be attenuated too quickly (maximal gain decrease in 
         * dB/second).
         * @param lpnMaxGain Ensure volume doesn't become too loud (maximal gain
         * in dB).
         * @see TeamTalk4.EnableAGC */
        public bool GetAGCSettings(out int lpnGainLevel,
                            out int lpnMaxIncrement,
                            out int lpnMaxDecrement,
                            out int lpnMaxGain)
        {
            return TT4.TT_GetAGCSettings(m_ttInst, out lpnGainLevel,
                out lpnMaxIncrement, out lpnMaxDecrement, out lpnMaxGain);
        }
        /** 
         * @brief Enable automatically position users using 3D-sound.
         *
         * Note that 3d-sound does not work if sound is running in duplex
         * mode (#ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX).
         *
         * @param bEnable TRUE to enable, otherwise FALSE.
         * @see TeamTalk4.SetUserPosition */
        public bool Enable3DSoundPositioning(bool bEnable)
        {
            return TT4.TT_Enable3DSoundPositioning(m_ttInst, bEnable);
        }
        /** 
         * @brief Automatically position users using 3D-sound.
         *
         * Note that 3d-sound does not work if sound is running in duplex
         * mode (#ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX).
         *
         * @see TeamTalk4.SetUserPosition */
        public bool AutoPositionUsers()
        {
            return TT4.TT_AutoPositionUsers(m_ttInst);
        }

        /**
         * @brief Enable/disable access to user's raw audio.
         *
         * With audio callbacks enabled all audio which has been played
         * will be accessible by calling TeamTalk4.AcquireUserAudioBlock(). Every
         * time a new #BearWare.AudioBlock is available the event
         * OnUserAudioBlock() is generated.
         * 
         * @see TeamTalk4.AcquireUserAudioBlock()
         * @see TeamTalk4.ReleaseUserAudioBlock()
         * @see OnUserAudioBlock() */
        public bool EnableAudioBlockEvent(bool bEnable)
        {
            return TT4.TT_EnableAudioBlockEvent(m_ttInst, bEnable);
        }
        /** @} */

        /** @addtogroup transmission
         * @{ */

        /**
         * @brief Enable voice activation.
         *
         * The client instance will start transmitting audio if the recorded audio
         * level is above or equal to the voice activation level set by
         * #SetVoiceActivationLevel. Once the voice activation level is reached
         * the event #OnVoiceActivation is posted.
         *
         * The current volume level can be queried calling
         * #GetSoundInputLevel.
         *
         * @param bEnable TRUE to enable, otherwise FALSE.
         * @see ClientFlag ::CLIENT_SNDINPUT_VOICEACTIVATED */
        public bool EnableVoiceActivation(bool bEnable)
        {
            return TT4.TT_EnableVoiceActivation(m_ttInst, bEnable);
        }
        /** 
         * @brief Set voice activation level.
         *
         * The current volume level can be queried calling
         * GetSoundInputLevel().
         *
         * @param nLevel Must be between #BearWare.SoundLevel.SOUND_VU_MIN and #BearWare.SoundLevel.SOUND_VU_MAX
         * @see TeamTalk4.EnableVoiceActivation
         * @see TeamTalk4.GetVoiceActivationLevel */
        public bool SetVoiceActivationLevel(int nLevel)
        {
            return TT4.TT_SetVoiceActivationLevel(m_ttInst, nLevel);
        }
        /** 
         * @brief Get voice activation level.
         *
         * @return Returns A value between #BearWare.SoundLevel.SOUND_VU_MIN and #BearWare.SoundLevel.SOUND_VU_MAX
         * @see TeamTalk4.EnableVoiceActivation
         * @see TeamTalk4.SetVoiceActivationLevel */
        public int GetVoiceActivationLevel()
        {
            return TT4.TT_GetVoiceActivationLevel(m_ttInst);
        }
        /**
         * @brief Set the delay of when voice activation should be stopped.
         *
         * When TeamTalk4.GetSoundInputLevel() becomes higher than the specified
         * voice activation level the client instance will start
         * transmitting until TeamTalk4.GetSoundInputLevel() becomes lower than
         * the voice activation level, plus a delay. This delay is by
         * default set to 1500 msec but this value can be changed by
         * calling TeamTalk4.SetVoiceActivationStopDelay().
         *
         * @see EnableVoiceActivation
         * @see SetVoiceActivationLevel */
        public bool SetVoiceActivationStopDelay(int nDelayMSec)
        {
            return TT4.TT_SetVoiceActivationStopDelay(m_ttInst, nDelayMSec);
        }

        /**
         * @brief Get the delay of when voice active state should be disabled.
         *
         * @return The number of miliseconds before voice activated state
         * should be turned back to inactive.
         *
         * @see SetVoiceActivationStopDelay
         * @see EnableVoiceActivation
         * @see SetVoiceActivationLevel */
        public int GetVoiceActivationStopDelay()
        {
            return TT4.TT_GetVoiceActivationStopDelay(m_ttInst);
        }

        /**
         * @brief Store audio conversations to a single file.
         *
         * Unlike TeamTalk4.SetUserAudioFolder(), which stores users' audio
         * streams in separate files, TeamTalk4.StartRecordingMuxedAudioFile()
         * muxes the audio streams into a single file.
         *
         * The audio streams, which should be muxed together, are
         * required to use the same audio codec. In most cases this is
         * the audio codec of the channel where the user is currently
         * participating (i.e. @c codec member of #BearWare.Channel).
         *
         * If the user changes to a channel which uses a different audio
         * codec then the recording will continue but simply be silent
         * until the user again joins a channel with the same audio codec
         * as was used for initializing muxed audio recording.
         *
         * Calling TeamTalk4.StartRecordingMuxedAudioFile() will enable the
         * #ClientFlag ::CLIENT_MUX_AUDIOFILE flag from TeamTalk4.GetFlags().
         *
         * Call TeamTalk4.StopRecordingMuxedAudioFile() to stop recording. Note
         * that only one muxed audio recording can be active at the same
         * time.
         *
         * @param lpAudioCodec The audio codec which should be used as
         * reference for muxing users' audio streams. In most situations
         * this is the #BearWare.AudioCodec of the current channel, i.e.
         * TeamTalk4.GetMyChannelID().
         * @param szAudioFileName The file to store audio to, e.g. 
         * C:\\MyFiles\\Conf.mp3.
         * @param uAFF The audio format which should be used in the recorded
         * file. The muxer will convert to this format.
         *
         * @see SetUserAudioFolder()
         * @see StopRecordingMuxedAudioFile() */
        public bool StartRecordingMuxedAudioFile(AudioCodec lpAudioCodec,
                                                 string szAudioFileName,
                                                 AudioFileFormat uAFF)
        {
            return TT4.TT_StartRecordingMuxedAudioFile(m_ttInst,
                                                       ref lpAudioCodec,
                                                       szAudioFileName,
                                                       uAFF);
        }

        /**
         * @brief Stop an active muxed audio recording.
         *
         * A muxed audio recording started with
         * TeamTalk4.StartRecordingMuxedAudioFile() can be stopped using this
         * function.
         *
         * Calling TeamTalk4.StopRecordingMuxedAudioFile() will clear the
         * #ClientFlag ::CLIENT_MUX_AUDIOFILE flag from TeamTalk4.GetFlags().
         *
         * @see StartRecordingMuxedAudioFile() */
        public bool StopRecordingMuxedAudioFile()
        {
            return TT4.TT_StopRecordingMuxedAudioFile(m_ttInst);
        }
        /** @} */

        /** @addtogroup videocapture
         * @{ */

        /**
         * @brief Get the list of devices available for video capture.
         *
         * @param lpVideoDevices An output array of #BearWare.VideoCaptureDevice-stucts 
         * which will receive the available video capture devices
         * @see TeamTalk4.InitVideoCaptureDevice */
        public bool GetVideoCaptureDevices(out VideoCaptureDevice[] lpVideoDevices)
        {
            //To speed up query we only query for a max of 25.
            //Hopefully no one has more than 25 capture devices. 
            VideoCaptureDevice[] devs = new VideoCaptureDevice[25];
            int count = devs.Length;
            bool b = TT4.TT_GetVideoCaptureDevices(m_ttInst, devs, ref count);
            if (b)
            {
                lpVideoDevices = new VideoCaptureDevice[count];
                for (int i = 0; i < count; i++)
                    lpVideoDevices[i] = devs[i];
            }
            else lpVideoDevices = null;

            return b;
        }
        /**
         * @brief Initialize a video capture device.
         *
         * @param szDeviceID The device idenfier @a szDeviceID of #BearWare.VideoCaptureDevice.
         * @param lpCaptureFormat The capture format to use,
         * i.e. frame-rate, resolution and picture format.
         * @param lpVideoCodec Video codec to use for transmission. Use NULL or set 
         * @a nCodec of #BearWare.VideoCodec to #Codec ::NO_CODEC if capture device should only be 
         * used for testing locally. 
         * @see TeamTalk4.GetVideoCaptureDevices
         * @see TeamTalk4.CloseVideoCaptureDevice
         * @see TeamTalk4.EnableTransmission */
        public bool InitVideoCaptureDevice(string szDeviceID,
                                    CaptureFormat lpCaptureFormat,
                                    VideoCodec lpVideoCodec)
        {
            return TT4.TT_InitVideoCaptureDevice(m_ttInst, szDeviceID, ref lpCaptureFormat, ref lpVideoCodec);
        }
        /**
         * @brief Close a video capture device.
         *
         * @see TeamTalk4.InitVideoCaptureDevice */
        public bool CloseVideoCaptureDevice()
        {
            return TT4.TT_CloseVideoCaptureDevice(m_ttInst);
        }
        /**
         * @brief Paint user's video frame using a Windows' DC (device
         * context).
         *
         * Same as calling TeamTalk4.PaintVideoFrameEx() like this:
         *
           @verbatim
           ttclient.PaintVideoFrameEx(nUserID, hDC, XDest, YDest, 
                                      nDestWidth, nDestHeight, 0, 0, 
                                      src_bmp_width, src_bmp_height);
           @endverbatim
         *
         * @c src_bmp_width and @c src_bmp_height are extracted internally
         * from the source image. */
        public bool PaintVideoFrame(int nUserID,
                             System.IntPtr hDC,
                             int XDest,
                             int YDest,
                             int nDestWidth,
                             int nDestHeight)
        {
            return TT4.TT_PaintVideoFrame(m_ttInst, nUserID, hDC, XDest,
                                    YDest, nDestWidth, nDestHeight);
        }

        /** 
         * @brief Paint user's video frame using a Windows' DC (device
         * context).
         *
         * An application can either paint using TeamTalk4.GetUserVideoFrame()
         * which provides a raw RGB32 array of the image or the
         * application can ask the client instance to paint the image
         * using this function.
         *
         * Typically this paint operation will be called in the WM_PAINT
         * message. Here is how the client instance paints internally:
         *
           @verbatim
           StretchDIBits(hDC, nPosX, nPosY, nWidth, nHeight, XSrc, YSrc, 
                         nSrcWidth, nSrcHeight, frame_buf, &bmi,
                         DIB_RGB_COLORS, SRCCOPY);
           @endverbatim 
         * 
         * @param nUserID The user's ID. 0 for local user.
         * @param hDC The handle to the Windows device context.
         * @param XDest Coordinate of left corner where to start painting.
         * @param YDest Coordinate or top corner where to start painting.
         * @param nDestWidth The width of the image.
         * @param nDestHeight The height of the image.
         * @param XSrc The left coordinate in the source bitmap of where
         * to start reading.
         * @param YSrc The top left coordinate in the source bitmap of where
         * to start reading.
         * @param nSrcWidth The number of width pixels to read from source bitmap.
         * @param nSrcHeight The number of height pixels to read from source bitmap.
         * @see TeamTalk4.GetUserVideoFrame */
        public bool PaintVideoFrameEx(int nUserID,
                              System.IntPtr hDC,
                              int XDest,
                              int YDest,
                              int nDestWidth,
                              int nDestHeight,
                              int XSrc,
                              int YSrc,
                              int nSrcWidth,
                              int nSrcHeight)
        {
            return TT4.TT_PaintVideoFrameEx(m_ttInst, nUserID, hDC, XDest, 
                                            YDest, nDestWidth, nDestHeight, 
                                            XSrc, YSrc, nSrcWidth, nSrcHeight);
        }

        /**
         * @brief Get or query the raw RGB32 bitmap data of a user's video frame.
         *
         * @param nUserID The user's ID. 0 for local client instance.
         * @param lpPicture Reference to a bitmap which will receive the video frame.
         * If the @a lpPicture reference is null or not the proper size a new Bitmap object
         * will be allocated and returned instead. Ensure to ALWAYS call with the 
         * same @a lpPicture reference for each @a nUserID, otherwise a new Bitmap 
         * object will be allocated each time and cause the application to use a 
         * significant amount of memory.
         * @see PaintVideoFrame */
        public bool GetUserVideoFrame(int nUserID,
                               ref System.Drawing.Bitmap lpPicture)
        {
            CaptureFormat cap;
            if (!TT4.TT_GetUserVideoFrame(m_ttInst, nUserID, IntPtr.Zero, 0, out cap))
                return false;

            PixelFormat pixelformat = PixelFormat.Format32bppRgb;

            if (lpPicture == null ||
               lpPicture.Width != cap.nWidth ||
               lpPicture.Height != cap.nHeight
#if PocketPC
                )
#else
               || lpPicture.PixelFormat != pixelformat)
#endif
            {
                lpPicture = new Bitmap(cap.nWidth, cap.nHeight,
                                       pixelformat);
            }

            // Lock the bitmap's bits.  
            System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, 0, lpPicture.Width, lpPicture.Height);
            BitmapData bmpData = lpPicture.LockBits(rect, ImageLockMode.ReadWrite, pixelformat);

            // Get the address of the first line.
            IntPtr ptr = bmpData.Scan0;

            bool b = TT4.TT_GetUserVideoFrame(m_ttInst, nUserID, ptr,
                                              cap.nWidth * cap.nHeight * 4, out cap);
            // Unlock the bits.
            lpPicture.UnlockBits(bmpData);
            return b;
        }

        /** @brief Extract a user's video frame by making TeamTalk
         * allocate the image buffer.
         *
         * Unlike TeamTalk4.GetUserVideoFrame() this function does not require
         * that the user preallocates a buffer which will contain the
         * image data. Instead the image buffer is allocated in
         * internally. REMEMBER, however, to call
         * TeamTalk4.ReleaseUserVideoFrame() when the image has been processed so
         * the resources allocated by TeamTalk can be released.
         *
         * @param nUserID The user's ID. 0 for local user.
         * @param lpVideoFrame The struct which will contain the image data. Note 
         * that it's the @a frameBuffer member of #BearWare.VideoFrame which will contain 
         * the image data allocated internally by TeamTalk.
         * @return Returns TRUE if a video frame was successfully put in the 
         * @a lpVideoFrame parameter.
         * @see GetUserVideoFrame
         * @see ReleaseUserVideoFrame */
        public bool AcquireUserVideoFrame(int nUserID,
                                          ref VideoFrame lpVideoFrame)
        {
            return TT4.TT_AcquireUserVideoFrame(m_ttInst, nUserID, ref lpVideoFrame);
        }
        /** @brief Delete a user's video frame, acquired through
         * TeamTalk4.AcquireUserVideoFrame(), so its allocated resources can be
         * released.
         *
         * @param nUserID The user's ID. 0 for local user.
         * @return Returns TRUE if a video frame was successfully put in the 
         * @a lpVideoFrame parameter.
         * @see AcquireUserVideoFrame
         * @see GetUserVideoFrame */
        public bool ReleaseUserVideoFrame(int nUserID)
        {
            return TT4.TT_ReleaseUserVideoFrame(m_ttInst, nUserID);
        }
        /** @} */

        /** @addtogroup transmission
         * @{ */

        /**
         * @brief Start/stop transmitting audio or video data.
         *
         * To check if transmission of either audio or video is enabled
         * call #GetFlags and check bits #ClientFlag ::CLIENT_TX_AUDIO and
         * #ClientFlag ::CLIENT_TX_VIDEO.
         * 
         * @param uTxType A bitmask of the data types to enable/disable transmission of.
         * @param bEnable Enable/disable transmission of bits in mask. */
        public bool EnableTransmission(TransmitType uTxType,
                                bool bEnable)
        {
            return TT4.TT_EnableTransmission(m_ttInst, uTxType,
                                       bEnable);
        }
        /** 
         * @brief Check if the client instance is currently transmitting.
         *
         * This call also checks if transmission is ongoing due to voice
         * activation.
         *
         * @param uTxType A bitmask specifying whether the client instance is 
         * currently transmitting the given #TransmitType.
         * @see TeamTalk4.EnableTransmission
         * @see TeamTalk4.EnableVoiceActivation */
        public bool IsTransmitting(TransmitType uTxType)
        {
            return TT4.TT_IsTransmitting(m_ttInst, uTxType);
        }

        /** 
         * @brief Stream a wave-file to a user in another channel. Only an
         * administrators can use this function.
         *
         * The event #OnStreamAudioFileUser is called when
         * audio file is started and stopped.
         *
         * @param nUserID The ID of the user to transmit to. The user cannot be
         * in the same channel as the local client instance.
         * @param szAudioFilePath Path to .wav file. The format of the .wav file
         * must be 16-bit PCM uncompressed. Use <a href="http://audacity.sourceforge.net">
         * Audacity</a> to convert to proper file format.
         * @see TeamTalk4.StopStreamingAudioFileToUser */
        public bool StartStreamingAudioFileToUser(int nUserID,
                                           string szAudioFilePath)
        {
            return TT4.TT_StartStreamingAudioFileToUser(m_ttInst, nUserID, szAudioFilePath);
        }
        /**
         * @brief Stop transmitting audio file.
         *
         * @param nUserID The ID of the user being transmitted to.
         * @see TeamTalk4.StartStreamingAudioFileToUser */
        public bool StopStreamingAudioFileToUser(int nUserID)
        {
            return TT4.TT_StopStreamingAudioFileToUser(m_ttInst, nUserID);
        }
        /** 
         * @brief Stream audio file to current channel.
         * 
         * Currently it is only possible to stream to the channel which the 
         * local client instance is participating in.
         *
         * When streaming to the current channel it basically replaces the
         * microphone input with a .wav file. Note that it is not possible
         * to stream a .wav file to a single user in the current channel.
         *
         * The event #OnStreamAudioFileChannel is posted when audio
         * file is being processed.
         *
         * @param nChannelID Pass GetMyChannelID(). Transmitting to other 
         * channels is currently not supported.
         * @param szAudioFilePath Path to .wav file. The format of the .wav file
         * must be 16-bit PCM uncompressed. Use <a href="http://audacity.sourceforge.net">
         * Audacity</a> to convert to proper file format.
         * @see StartStreamingAudioFileToUser
         * @see StopStreamingAudioFileToChannel */
        public bool StartStreamingAudioFileToChannel(int nChannelID,
                                              string szAudioFilePath)
        {
            return TT4.TT_StartStreamingAudioFileToChannel(m_ttInst, nChannelID, szAudioFilePath);
        }
        /**
         * @brief Stop streaming audio file to current channel.
         *
         * @param nChannelID Pass GetMyChannelID(). */
        public bool StopStreamingAudioFileToChannel(int nChannelID)
        {
            return TT4.TT_StopStreamingAudioFileToChannel(m_ttInst, nChannelID);
        }

        /**
         * @brief Stream media file to channel, e.g. avi-, wav- or MP3-file.
         *
         * This function is meant as a replacement for
         * TeamTalk4.StartStreamingAudioFileToChannel() since it can also stream
         * wave-files.
         *
         * Call TeamTalk4.GetMediaFileInfo() to get the properties of a media
         * file, i.e. audio and video format.
         *
         * When a media file is being streamed it will replace the
         * microphone and video input. Therefore calling
         * TeamTalk4.EnableTransmission() will have no effect.
         *
         * The event OnStreamMediaFileToChannel() is called when
         * the media file starts streaming. The flags #ClientFlag #CLIENT_STREAM_AUDIO
         * and/or #CLIENT_STREAM_VIDEO will be set if the call is successful.
         *
         * It is important for the users who are receiving the streamed media file
         * to increase their media playback buffer. This is because a streamed
         * media file will generate bigger chunks of data unlike microphone and 
         * video capture devices which run in realtime. Use 
         * TeamTalk4.SetUserMediaBufferSize() to change the size of a user's media 
         * buffer.
         *
         * @param szMediaFilePath File path to media file.
         * @param lpVideoCodec If video file then specify output codec properties 
         * here. Specify #Codec ::NO_CODEC if video should be ignored.
         * @param uTxType Specify whether audio, video or both should be streamed
         * to channel. If #TRANSMIT_AUDIO is specified then the flag 
         * #CLIENT_STREAM_AUDIO will be set. If #TRANSMIT_VIDEO is specified then
         * the flag #CLIENT_STREAM_VIDEO is set.
         *
         * @see TeamTalk4.StopStreamingMediaFileToChannel() */
        public bool StartStreamingMediaFileToChannel(string szMediaFilePath,
                                              VideoCodec lpVideoCodec,
                                              TransmitType uTxType)
        {
            return TT4.TT_StartStreamingMediaFileToChannel(m_ttInst, szMediaFilePath,
                                                           ref lpVideoCodec, uTxType);
        }

        /**
         * @brief Stop streaming media file to channel.
         *
         * This will clear the flags #CLIENT_STREAM_AUDIO
         * and/or #CLIENT_STREAM_VIDEO.
         *
         * @see TeamTalk4.StartStreamingMediaFileToChannel() */
        public bool StopStreamingMediaFileToChannel()
        {
            return TT4.TT_StopStreamingMediaFileToChannel(m_ttInst);
        }

        /**
         * @brief Get the properties of a media file.
         *
         * Use this function to determine the audio and video properties of
         * a media file, so the user knows what can be streamed.
         *
         * @see TeamTalk4.StartStreamingMediaFileToChannel() */
        public static bool GetMediaFileInfo(string szMediaFilePath,
                                            out MediaFileInfo pMediaFileInfo)
        {
            return TT4.TT_GetMediaFileInfo(szMediaFilePath, out pMediaFileInfo);
        }

        /** @} */

        /** @addtogroup desktopshare
         * @{ */

        /**
         * @brief Transmit a desktop window (bitmap) to users in the same
         * channel.
         *
         * When TeamTalk4.SendDesktopWindow() is called the first time a new
         * desktop session will be started. To update the current desktop
         * session call TeamTalk4.SendDesktopWindow() again once the previous
         * desktop transmission has finished. Tracking progress of the
         * current desktop transmission is done by checking for the
         * TeamTalk4.OnDesktopTransferUpdate() event. While the desktop
         * transmission is active the flag #ClientFlag ::CLIENT_TX_DESKTOP will be set
         * on the local client instance.
         *
         * If the desktop window (bitmap) changes size (width/height) or
         * format a new desktop session will be started. Also if the user
         * changes channel a new desktop session will be started. Check @c
         * nSessionID of #BearWare.DesktopWindow to see if a new desktop session is
         * started or the TeamTalk4.OnUserDesktopWindow() event.
         *
         * Remote users will get the TeamTalk4.OnUserDesktopWindow() event
         * and can call TeamTalk4.GetUserDesktopWindow() to retrieve the desktop
         * window.
         * 
         * @note
         * Requires server version 4.3.0.1490 or later.
         *
         * @param lpBitmap Pointer to bitmap buffer.
         * @param nBitmapSize Size of bitmap buffer @c lpBitmap in bytes. The size 
         * of the bitmap can be calculated using the #BearWare.DesktopWindow-struct's 
         * @c nBytesPerLine multiplied by the @c nHeight.
         * @param lpDesktopWindow Properties of the bitmap. Set the @c nSessionID 
         * property to 0.
         * @param nConvertBmpFormat Before transmission convert the bitmap to this 
         * format.
         * @return TRUE if desktop window is queued for transmission. FALSE if 
         * @c nBitmapSize is invalid or if a desktop transmission is already 
         * active.
         * @return -1 on error. 0 if bitmap has no changes. Greater than 0 on 
         * success.
         * @see TeamTalk4.CloseDesktopWindow()
         * @see TeamTalk4.SendDesktopCursorPosition() */
        public int SendDesktopWindow(System.IntPtr lpBitmap,
                                                   int nBitmapSize,
                                                   DesktopWindow lpDesktopWindow,
                                                   BitmapFormat nConvertBmpFormat)
        {
            return TT4.TT_SendDesktopWindow(m_ttInst, lpBitmap, nBitmapSize, ref lpDesktopWindow, nConvertBmpFormat);
        }

        /**
         * @brief Close the current desktop session.
         *
         * Closing the desktop session will cause the users receiving the
         * current desktop session to see the desktop session ID change to
         * 0 in the TeamTalk4.OnUserDesktopWindow() event.
         */
        public bool CloseDesktopWindow()
        {
            return TT4.TT_CloseDesktopWindow(m_ttInst);
        }

        /**
         * @brief Get RGB values of the palette for the bitmap format.
         *
         * This currently only applies to bitmaps of format #BitmapFormat ::BMP_RGB8_PALETTE.
         *
         * Note that the pointer returned is non-const which means the
         * palette can be overwritten with a custom palette. The custom
         * palette will then be used internally during bitmap
         * conversion.
         *
         * @param nBmpPalette The bitmap format. Currently only #BitmapFormat ::BMP_RGB8_PALETTE
         * is supported.
         * @param nIndex The index in the color table of the RGB values to 
         * extract.
         * @return Pointer to RGB colors. First byte is Red, second Blue and 
         * third Green. Returns NULL if the color-index is invalid. */
        public static System.Drawing.Color Palette_GetColorTable(BitmapFormat nBmpPalette,
                                                          int nIndex)
        {
            IntPtr ptr = TT4.TT_Palette_GetColorTable(nBmpPalette, nIndex);
            switch(nBmpPalette)
            {
                case BitmapFormat.BMP_RGB8_PALETTE:
                    return Color.FromArgb(Marshal.ReadInt32(ptr));
            }
            return new Color();
        }

        /**
         * @brief Transmit the specified window in a desktop session.
         *
         * Same as TeamTalk4.SendDesktopWindow() except the properties for the
         * #BearWare.DesktopWindow are extracted automatically.
         * 
         * @param hWnd Windows handle for the window to transmit.
         * @param nBitmapFormat Bitmap format to use for the transmitted image.
         * @param nDesktopProtocol The protocol to use for transmitting the image.
         * @return See TeamTalk4.SendDesktopWindow(). */
        public int SendDesktopWindowFromHWND(System.IntPtr hWnd,
                                               BitmapFormat nBitmapFormat,
                                               DesktopProtocol nDesktopProtocol)
        {
            return TT4.TT_SendDesktopWindowFromHWND(m_ttInst, hWnd, nBitmapFormat, nDesktopProtocol);
        }

        /**
         * @brief Paint user's desktop window using a Windows' DC (device
         * context).
         *
         * Same as calling TT_PaintDesktopWindowEx() like this:
         *
           @verbatim
           TT_PaintDesktopWindowEx(nUserID, hDC, XDest, YDest, nDestWidth,
                                   nDestHeight, 0, 0, 
                                   'src_bmp_width', 'src_bmp_height');
           @endverbatim
         *
         * @c src_bmp_width and @c src_bmp_height are extracted internally
         * from the source image. */
        public bool PaintDesktopWindow(int nUserID,
                                       System.IntPtr hDC,
                                       int XDest,
                                       int YDest,
                                       int nDestWidth,
                                       int nDestHeight)
        {
            return TT4.TT_PaintDesktopWindow(m_ttInst, nUserID, hDC, XDest, YDest, nDestWidth, nDestHeight);
        }

        /**
         * @brief Paint user's desktop window using a Windows' DC (device
         * context).
         *
         * An application can either paint a bitmap by using
         * TeamTalk4.GetUserDesktopWindow() which provides a pointer to a bitmap
         * or the application can ask the client instance to paint the
         * image using this function.
         *
         * Typically this paint operation will be called in the WM_PAINT
         * message. Here is how the client instance paints internally:
         *
           @verbatim
           StretchDIBits(hDC, nPosX, nPosY, nWidth, nHeight, XSrc, YSrc, 
                         nSrcWidth, nSrcHeight, frame_buf, &bmi,
                         DIB_RGB_COLORS, SRCCOPY);
           @endverbatim 
         * 
         * @param nUserID The user's ID.
         * @param hDC The handle to the Windows device context.
         * @param XDest Coordinate of left corner where to start painting.
         * @param YDest Coordinate or top corner where to start painting.
         * @param nDestWidth The width of the image.
         * @param nDestHeight The height of the image.
         * @param XSrc The left coordinate in the source bitmap of where
         * to start reading.
         * @param YSrc The top left coordinate in the source bitmap of where
         * to start reading.
         * @param nSrcWidth The number of width pixels to read from source bitmap.
         * @param nSrcHeight The number of height pixels to read from source bitmap.
         * @return TRUE on success. FALSE on error, e.g. if user doesn't exist.
         * @see TeamTalk4.GetUserDesktopWindow() */
        public bool PaintDesktopWindowEx(int nUserID,
                                         System.IntPtr hDC,
                                         int XDest,
                                         int YDest,
                                         int nDestWidth,
                                         int nDestHeight,
                                         int XSrc,
                                         int YSrc,
                                         int nSrcWidth,
                                         int nSrcHeight)
        {
            return TT4.TT_PaintDesktopWindowEx(m_ttInst, nUserID, hDC, XDest, 
                                               YDest, nDestWidth, nDestHeight, 
                                               XSrc, YSrc, nSrcWidth, nSrcHeight);
        }

        /**
         * @brief Send the position of mouse cursor to users in the same channel.
         *
         * It's only possible to send the mouse cursor position if there's
         * a desktop session which is currently active.
         * 
         * @note
         * Requires server version 4.3.0.1490 or later.
         *
         * @param nUserID For now set to 0.
         * @param nPosX X coordinate of mouse cursor. Max is 65535.
         * @param nPosY Y coordinate of mouse cursor. Max is 65535.
         * @see TeamTalk4.SendDesktopWindow() */
        public bool SendDesktopCursorPosition(int nUserID,
                                              int nPosX,
                                              int nPosY)
        {
            return TT4.TT_SendDesktopCursorPosition(m_ttInst, nUserID, nPosX, nPosY);
        }

        /** 
         * @brief Send a mouse or keyboard event to a shared desktop window.
         *
         * If a user is sharing a desktop window it's possible for a
         * remote user to take control of mouse and keyboard input on the
         * remote computer. Read section @ref txdesktopinput on how to
         * transmit desktop input to a shared window.
         *
         * When the remote user receives the issued #BearWare.DesktopInput the
         * event #BearWare.TeamTalk4.OnUserDesktopInput is posted to the client
         * instance sharing the desktop window and
         * #BearWare.TeamTalk4.GetUserDesktopInput can be used to retrieve the
         * transmitted #BearWare.DesktopInput structs.
         *
         * Requires TeamTalk v. 4.6+ client and server.
         *
         * @param nUserID The user's ID who owns the shared desktop window
         * and should receive desktop input.
         * @param lpDesktopInputs An array of #BearWare.DesktopInput structs which
         * should be transmitted to the user. Internally in the client
         * instance each user ID has an internal queue which can contain a
         * maximum of 100 #BearWare.DesktopInput structs.
         * @return FALSE If user doesn't exist or if desktop input queue is full or
         * if @c nUserID doesn't subscribe to desktop input. */
        public bool SendDesktopInput(int nUserID,
                                     DesktopInput[] lpDesktopInputs)
        {
            return TT4.TT_SendDesktopInput(m_ttInst, nUserID, lpDesktopInputs, lpDesktopInputs.Length);
        }

        /**
         * @brief Get a user's desktop window (bitmap image).
         *
         * A user's desktop window can be extracted when the 
         * TeamTalk4.OnUserDesktopWindow() is received.
         *
         * A desktop window is simply a bitmap image. This method is used for
         * copying the user's bitmap image to a pre-allocated buffer.
         *
         * To know the properties of the bitmap call this method with @c
         * lpBitmap set to NULL and extract the properties in @c
         * lpDesktopWindow.  The size of the buffer to allocate will be @c
         * nBytesPerLine multiplied by @c nHeight in the #BearWare.DesktopWindow.
         *
         * For #BitmapFormat ::BMP_RGB8_PALETTE bitmaps check out TeamTalk4.Palette_GetColorTable().
         *
         * @param nUserID The user's ID.
         * @param lpBitmap Pointer to a pre-allocated buffer where the bitmap
         * data will be copied to. Pass NULL to query the byte size of the 
         * bitmap, so it can be written to @c lpnBitmapSize and @c 
         * lpDesktopWindow.
         * @param lpnBitmapSize Size of the allocated bitmap buffer @c
         * lpBitmap. If @c lpBitmap is NULL the size of the bitmap will be
         * written to this parameter.
         * @param lpDesktopWindow The properties of the shared desktop window.
         * Pass NULL to @c lpBitmap to query the properties of the desktop 
         * window.
         * @return FALSE if there's no active desktop window for this user.
         * @see TeamTalk4.SendDesktopWindow() */
        public bool GetUserDesktopWindow(int nUserID,
                                         System.IntPtr lpBitmap,
                                         ref int lpnBitmapSize,
                                         ref DesktopWindow lpDesktopWindow)
        {
            return TT4.TT_GetUserDesktopWindow(m_ttInst, nUserID, lpBitmap, ref lpnBitmapSize, ref lpDesktopWindow);
        }

        /**
         * @brief Get the mouse cursor position of a user.
         *
         * The mouse cursor position will be available when the
         * TeamTalk4.OnUserDesktopCursorPosition() is received and there's an
         * active desktop session.
         *
         * A mouse cursor position is transmitted using 
         * TeamTalk4.SendDesktopCursorPosition().
         *
         * @param nSrcUserID The owner of the cursor.
         * @param nDestUserID The owner of the desktop session where the cursor
         * is pointing to.
         * @param lpnPosX Output parameter for X coordinate.
         * @param lpnPosY Output parameter for Y coordinate.
         * @return FALSE if no coordinates were written to @c lpnPosX and 
         * @c lpnPosY. */
        public bool GetUserDesktopCursor(int nSrcUserID,
                                         int nDestUserID,
                                         ref int lpnPosX,
                                         ref int lpnPosY)
        {
            return TT4.TT_GetUserDesktopCursor(m_ttInst, nSrcUserID, nDestUserID, ref lpnPosX, ref lpnPosY);
        }

        /**
         * @brief Get desktop (mouse or keyboard) input received from a user.
         *
         * Desktop input is used in combination with a shared desktop
         * window, see @ref desktopshare.
         *
         * If the local client instance is subscribing to desktop input
         * (i.e. ::SUBSCRIBE_DESKTOPINPUT) then a remote user can send
         * desktop input to the local client instance. Once the client
         * instance receives desktop input it posts the message
         * TeamTalk4.OnUserDesktopInput() #BearWare.TeamTalk4.GetUserDesktopInput() can
         * be used to obtain the #BearWare.DesktopInput structs send by the remote
         * user. See @ref rxdesktopinput for more information on receiving
         * desktop input.
         *
         * @param nSrcUserID The user's ID who sent desktop input, see
         * #BearWare.TeamTalk4.OnUserDesktopInput().
         * @param lpDesktopInputs A pre-allocated array which will receive
         * the #BearWare.DesktopInput stucts.
         * @return FALSE if no desktop input could be retrieved.
         * @see TeamTalk4.SendDesktopInput()
         * @see WindowsDesktopInput.KeyTranslate()  */
        public bool GetUserDesktopInput(int nSrcUserID,
                                        out DesktopInput[] lpDesktopInputs)
        {
            DesktopInput[] inputs = new DesktopInput[TeamTalk4.TT_DESKTOPINPUT_MAX];
            int nCount = TeamTalk4.TT_DESKTOPINPUT_MAX;
            if (TT4.TT_GetUserDesktopInput(m_ttInst, nSrcUserID, inputs, ref nCount))
            {
                lpDesktopInputs = new DesktopInput[nCount];
                for (int i = 0; i < nCount; i++)
                    lpDesktopInputs[i] = inputs[i];
                return true;
            }
            else
            {
                lpDesktopInputs = new DesktopInput[0];
                return false;
            }
        }

        /** @} */

        /** @addtogroup connectivity
         * @{ */

        /**
         * @brief Connect to a server. 
         * 
         * This is a non-blocking call (but may block due to DNS lookup)
         * so the user application must wait for the event
         * #OnConnectSuccess to be posted once the connection has
         * been established or #OnConnectFailed if connection could
         * not be established. If the connection could not be establish
         * ensure to call #Disconnect to close open connections in the
         * client instance before trying again.
         *
         * Once connected call #DoLogin to log on to the server.
         * 
         * @param szHostAddress The IP-address or hostname of the server.
         * @param nTcpPort The host port of the server (TCP).
         * @param nUdpPort The audio/video port of the server (UDP).
         * @param nLocalTcpPort The local TCP port which should be used. 
         * Setting it to 0 makes OS select a port number (recommended).
         * @param nLocalUdpPort The local UDP port which should be used. 
         * Setting it to 0 makes OS select a port number (recommended).
         * @see OnConnectSuccess
         * @see OnConnectFailed
         * @see TeamTalk4.DoLogin */
        public bool Connect(string szHostAddress,
                    int nTcpPort,
                    int nUdpPort,
                    int nLocalTcpPort,
                    int nLocalUdpPort)
        {
            return TT4.TT_Connect(m_ttInst, szHostAddress, nTcpPort, nUdpPort, nLocalTcpPort, nLocalUdpPort);
        }

        /**
         * @brief Bind to specific IP-address priot to connecting to server.
         *
         * Same as Connect() except that this also allows which IP-address
         * to bind to on the local interface.
         *
         * @param szHostAddress The IP-address or hostname of the server.
         * @param nTcpPort The host port of the server (TCP).
         * @param nUdpPort The audio/video port of the server (UDP).
         * @param szBindIPAddr The IP-address to bind to on the local interface
         * in dotted decimal format, e.g. 192.168.1.10.
         * @param nLocalTcpPort The local TCP port which should be used. 
         * Setting it to 0 makes OS select a port number (recommended).
         * @param nLocalUdpPort The local UDP port which should be used. 
         * Setting it to 0 makes OS select a port number (recommended).
         * @see Connect */
        public bool ConnectEx(string szHostAddress,
                    int nTcpPort,
                    int nUdpPort,
                    string szBindIPAddr,
                    int nLocalTcpPort,
                    int nLocalUdpPort)
        {
            return TT4.TT_ConnectEx(m_ttInst, szHostAddress, nTcpPort, nUdpPort,
        szBindIPAddr, nLocalTcpPort, nLocalUdpPort);
        }

        /**
         * @brief Connect to non-encrypted TeamTalk server.
         *
         * This function is only useful in the Professional edition of the
         * TeamTalk SDK. It enabled the encrypted TeamTalk client to
         * connect to non-encrypted TeamTalk servers. The default
         * behaviour of TeamTalk4.Connect() and TeamTalk4.ConnectEx() in the
         * Professional SDK is to connect to encrypted servers.  */
        public bool ConnectNonEncrypted(string szHostAddress,
                                        int nTcpPort,
                                        int nUdpPort,
                                        string szBindIPAddr,
                                        int nLocalTcpPort,
                                        int nLocalUdpPort)
        {
            return TT4.TT_ConnectNonEncrypted(m_ttInst, szHostAddress, nTcpPort, nUdpPort, szBindIPAddr, nLocalTcpPort, nLocalUdpPort);
        }
        /**
         * @brief Disconnect from the server.
         * 
         * This will clear the flag #ClientFlag ::CLIENT_CONNECTED and #ClientFlag ::CLIENT_CONNECTING. */
        public bool Disconnect()
        {
            return TT4.TT_Disconnect(m_ttInst);
        }
        /**
         * @brief Query the maximum size of UDP data packets to the user
         * or server.
         *
         * The TeamTalk4.OnMaxPayloadUpdate() event is posted when
         * the query has finished.
         *
         * @param nUserID The ID of the user to query or 0 for querying 
         * server. Currently only @c nUserID = 0 is supported. */
        public bool QueryMaxPayload(int nUserID)
        {
            return TT4.TT_QueryMaxPayload(m_ttInst, nUserID);
        }
        /** 
         * @brief Set how often the client should ping the server on its TCP
         * and UDP connection.
         *
         * Ensure that both the TCP or the UDP ping interval is less than
         * the server's user-timeout specified by @a nUserTimeout in
         * #BearWare.ServerProperties. UDP keep-alive packets also updates the
         * ping-time to the server in @a nUdpPingTimeMs of
         * #BearWare.ClientStatistics.
         *
         * @param nTcpPingIntervalSec Seconds between issuing the ping-command.
         * Passing 0 will make the client instance use default settings.
         * @param nUdpPingIntervalSec Seconds between sending UDP keepalive
         * packets to the server (and p2p users). Passing 0 will make the 
         * client instance use default settings.
         * @see BearWare.ServerProperties */
        public bool SetKeepAliveInterval(int nTcpPingIntervalSec, int nUdpPingIntervalSec)
        {
            return TT4.TT_SetKeepAliveInterval(m_ttInst, nTcpPingIntervalSec, nUdpPingIntervalSec);
        }
        /** 
         * @brief Gets how often the client is sending keep-alive information to the 
         * server.
         *
         * @param lpnTcpPingIntervalSec The number of seconds between issuing the 
         * 'ping' command.
         * @param lpnUdpPingIntervalSec The number of seconds between sending UDP
         * keepalive packets.
         * @see TeamTalk4.SetKeepAliveInterval */
        public bool GetKeepAliveInterval(out int lpnTcpPingIntervalSec,
                                  out int lpnUdpPingIntervalSec)
        {
            return TT4.TT_GetKeepAliveInterval(m_ttInst, out lpnTcpPingIntervalSec, out lpnUdpPingIntervalSec);
        }
        /**
         * @brief Set server timeout for the client instance.
         * 
         * Set the number of seconds the client should allow the server
         * not to respond to keepalive requests
         *
         * @param nTimeoutSec Seconds before dropping connection if server hasn't 
         * replied. Passing 0 will use default (180 seconds). */
        public bool SetServerTimeout(int nTimeoutSec)
        {
            return TT4.TT_SetServerTimeout(m_ttInst, nTimeoutSec);
        }
        /**
         * @brief Get the server timeout for the client instance.
         * 
         * Get the number of seconds the client should allow the server
         * not to respond to keepalive requests
         *
         * @return The number of seconds. On error -1. */
        public int GetServerTimeout()
        {
            return TT4.TT_GetServerTimeout(m_ttInst);
        }
        /**
         * @brief Enable/disable peer to peer data transmission.
         * 
         * Peer to peer data transmission will reduce data transmission time,
         * since packets will not be forwarded through the server. This, however,
         * increases the bandwidth usage for clients since a separate data packet
         * will be sent to each user in a channel, instead of just sending a 
         * single packet to the server which would then broadcast the packet.
         *
         * Note that peer to peer data transmission is very unreliable and will 
         * only work with simple NAT-devices. Once a peer to peer connection 
         * succeeds or fails the event #OnConnectionP2P is posted.
         *
         * If the client instance is unable to connect with peer to
         * peer to a user it will try and forward the data packet
         * through the server if #UserRight ::USERRIGHT_FORWARD_AUDIO or
         * #UserRight ::USERRIGHT_FORWARD_VIDEO are enabled.
         *
         * @see TeamTalk4.GetServerProperties
         * @see BearWare.UserRights */
        public bool EnablePeerToPeer(TransmitType mask,
                             bool bEnable)
        {
            return TT4.TT_EnablePeerToPeer(m_ttInst, mask, bEnable);
        }
        /**
         * @brief Retrieve client statistics of bandwidth usage and
         * response times.
         *
         * @see BearWare.ClientStatistics */
        public bool GetStatistics(out ClientStatistics lpStats)
        {
            return TT4.TT_GetStatistics(m_ttInst, out lpStats);
        }
        /** 
         * @brief Get the number of bytes in a packet with the specified
         * audio codec. 
         *
         * Note that this is only an estimate which doesn't include
         * headers of underlying protocols.
         *
         * @param lpCodec The codec settings to test for packet size
         * @see BearWare.AudioCodec */
        public int GetPacketSize(AudioCodec lpCodec)
        {
            return TT4.TT_GetPacketSize(ref lpCodec);
        }
        /** @} */

        /** @addtogroup commands
         * @{ */

        /**
         * @brief Logon to a server.
         * 
         * Once connected to a server call this function to logon. If
         * the login is successful #OnCmdMyselfLoggedIn is
         * posted, otherwise #OnCmdError. Once logged on it's
         * not possible to talk to other users until the client instance
         * joins a channel. Call #DoJoinChannel to join a channel.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_INCORRECT_CHANNEL_PASSWORD
         * - #ClientError ::CMDERR_INVALID_ACCOUNT
         * - #ClientError ::CMDERR_MAX_SERVER_USERS_EXCEEDED
         * - #ClientError ::CMDERR_SERVER_BANNED
         * - #ClientError ::CMDERR_ALREADY_LOGGEDIN
         *
         * @param szNickname The nickname to use.
         * @param szServerPassword The server's password. Leave empty if user has 
         * account on the server.
         * @param szUsername If #UserRight ::USERRIGHT_GUEST_LOGIN is disabled a username and 
         * password must be specified in order to login. Leave blank if guest 
         * logins are ok.
         * @param szPassword The password of the user account on the server. Leave 
         * blank if no account is needed on the server.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see UserRight ::USERRIGHT_GUEST_LOGIN
         * @see BearWare.ServerProperties
         * @see TeamTalk4.DoJoinChannel
         * @see OnCmdMyselfLoggedIn
         * @see OnCmdError */
        public int DoLogin(string szNickname, string szServerPassword,
                    string szUsername, string szPassword)
        {
            return TT4.TT_DoLogin(m_ttInst, szNickname, szServerPassword,
                            szUsername, szPassword);
        }
        /**
         * @brief Logout of the server.
         *
         * If successful the event #OnCmdMyselfLoggedOut
         * will be posted.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         *
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see OnCmdMyselfLoggedOut */
        public int DoLogout()
        {
            return TT4.TT_DoLogout(m_ttInst);
        }
        /**
         * @brief Create a new channel and join it. This command requires
         * that the flag #UserRight ::USERRIGHT_CHANNEL_CREATION is set in @a uUserRights
         * of #BearWare.ServerProperties.
         *
         * This function can also be used to join an existing channel and
         * in this case the parameters @a szTopic and @a szOpPassword are
         * ignored.
         *
         * When #DoJoinChannel is used to create channels it works
         * similar to IRC. If the client instance tres to join a channel
         * which does not exist it will be created as a new channel. If
         * the client instance is the last user to leave a channel the
         * channel will be removed on the server. Only administrators can
         * create static (persistent) channels, namely by using
         * #DoMakeChannel.
         *
         * If the channel is created successfully the event
         * #OnCmdChannelNew will be sent, followed by
         * #OnCmdMyselfJoinedChannel.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_INCORRECT_CHANNEL_PASSWORD
         * - #ClientError ::CMDERR_MAX_CHANNEL_USERS_EXCEEDED
         * - #ClientError ::CMDERR_ALREADY_IN_CHANNEL
         * - #ClientError ::CMDERR_AUDIOCODEC_BITRATE_LIMIT_EXCEEDED
         * - #ClientError ::CMDERR_UNKNOWN_AUDIOCODEC If the server doesn't support the audio
         *   codec. Introduced in version 4.1.0.1264.
         *
         * @param lpChannel The channel to join or create if it doesn't already
         * exist.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoLeaveChannel
         * @see DoMakeChannel
         * @see OnCmdChannelNew
         * @see OnCmdMyselfJoinedChannel */
        public int DoJoinChannel(Channel lpChannel)
        {
            return TT4.TT_DoJoinChannel(m_ttInst, ref lpChannel);
        }
        /**
         * @brief Join an existing channel.
         * 
         * This command basically calls #DoJoinChannel but omits the
         * unnecessary parameters for only joining a channel and not
         * creating a new one.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_INCORRECT_CHANNEL_PASSWORD
         * - #ClientError ::CMDERR_MAX_CHANNEL_USERS_EXCEEDED
         * - #ClientError ::CMDERR_ALREADY_IN_CHANNEL
         *
         * @param nChannelID The ID of the channel to join.
         * @param szPassword The password for the channel to join.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoLeaveChannel
         * @see DoMakeChannel
         * @see OnCmdChannelNew
         * @see OnCmdMyselfJoinedChannel */
        public int DoJoinChannelByID(int nChannelID, string szPassword)
        {
            return TT4.TT_DoJoinChannelByID(m_ttInst, nChannelID, szPassword);
        }
        /**
         * @brief Leave the current channel.
         *
         * Note that #DoLeaveChannel() doesn't take any parameters
         * since a user can only participate in one channel at the time.
         * If command is successful the event #OnCmdMyselfLeftChannel
         * will be posted.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_NOT_IN_CHANNEL
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         *
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoJoinChannel
         * @see OnCmdMyselfLeftChannel */
        public int DoLeaveChannel()
        {
            return TT4.TT_DoLeaveChannel(m_ttInst);
        }
        /**
         * @brief Change the client instance's nick name.
         *
         * The event #OnCmdUserUpdate will be posted if the
         * update was successful.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         *
         * @param szNewNick is the new nick name to use.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see OnCmdUserUpdate */
        public int DoChangeNickname(string szNewNick)
        {
            return TT4.TT_DoChangeNickname(m_ttInst, szNewNick);
        }
        /**
         * @brief Change the client instance's currect status
         *
         * The event #OnCmdUserUpdate will be posted if the update
         * was successful.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         *
         * @param nStatusMode The value for the status mode.
         * @param szStatusMessage The user's message associated with the status 
         * mode.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see OnCmdUserUpdate */
        public int DoChangeStatus(int nStatusMode, string szStatusMessage)
        {
            return TT4.TT_DoChangeStatus(m_ttInst, nStatusMode, szStatusMessage);
        }
        /**
         * @brief Send a text message to either a user or a channel. 
         *
         * Can also be a broadcast message which is received by all users
         * on the server.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED 
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param lpTextMessage A preallocated text-message struct.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see UserRight ::USERRIGHT_CLIENT_BROADCAST */
        public int DoTextMessage(TextMessage lpTextMessage)
        {
            return TT4.TT_DoTextMessage(m_ttInst, ref lpTextMessage);
        }
        /**
         * @brief Make another user operator of a channel. 
         * 
         * Requires that the client instance must already be operator of
         * the channel or is logged in as an administrator.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param nUserID The user who should become channel operator. 
         * @param nChannelID The channel where the user should become operator
         * @param bMakeOperator Whether user should be op'ed or deop'ed
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoChannelOpEx */
        public int DoChannelOp(int nUserID, int nChannelID, bool bMakeOperator)
        {
            return TT4.TT_DoChannelOp(m_ttInst, nUserID, nChannelID, bMakeOperator);
        }
        /**
         * @brief Make another user operator of a channel using the 
         * @a szOpPassword of #BearWare.Channel.
         * 
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         * - #ClientError ::CMDERR_INCORRECT_OP_PASSWORD
         *
         * @note
         * Requires server version 4.1.0.994 or later.
         *
         * @param nUserID The user who should become channel operator. 
         * @param nChannelID The channel where the user should become operator.
         * @param szOpPassword The @a szOpPassword of #BearWare.Channel.
         * @param bMakeOperator Whether user should be op'ed or deop'ed.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoChannelOp */
        public int DoChannelOpEx(int nUserID,
                          int nChannelID,
                          string szOpPassword,
                          bool bMakeOperator)
        {
            return TT4.TT_DoChannelOpEx(m_ttInst, nUserID, nChannelID, szOpPassword, bMakeOperator);
        }
        /**
         * @brief Kick user out of channel. 
         *
         * Only a channel operator or administration can kick users. To
         * ban a user call #DoBanUser before #DoKickUser.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param nUserID The ID of the user to kick. 
         * @param nChannelID The channel where the user is participating. If
         * local instance is admin and @a nChannelID is 0, the user will be kicked
         * off the server. 
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoBanUser */
        public int DoKickUser(int nUserID, int nChannelID)
        {
            return TT4.TT_DoKickUser(m_ttInst, nUserID, nChannelID);
        }
        /**
         * @brief Send a file to the specified channel. 
         *
         * If user is logged on as an admin the file can be located in any 
         * channel. If the user is not an admin the file must be located in 
         * the same channel as the user is currently participating in.
         * The file being uploaded
         * must have a file size which is less than the disk quota of the channel, 
         * minus the sum of all the files in the channel. The disk quota of a
         * channel can be obtained in the nDiskQuota of the #BearWare.Channel struct passed 
         * to #GetChannel.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_MAX_DISKUSAGE_EXCEEDED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_FILETRANSFER_NOT_FOUND
         * - #ClientError ::CMDERR_OPENFILE_FAILED
         * - #ClientError ::CMDERR_FILE_NOT_FOUND
         * - #ClientError ::CMDERR_FILE_ALREADY_EXISTS
         * - #ClientError ::CMDERR_FILESHARING_DISABLED
         *
         * @param nChannelID The ID of the channel of where to put the file. Only 
         * admins can upload in channel other then their own.
         * @param szLocalFilePath The path of the file to upload, e.g. C:\\myfile.txt.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see Channel
         * @see GetChannel */
        public int DoSendFile(int nChannelID, string szLocalFilePath)
        {
            return TT4.TT_DoSendFile(m_ttInst, nChannelID, szLocalFilePath);
        }
        /**
         * @brief Download a file from the specified channel. 
         *
         * If user is logged on as an admin the file can be located in any 
         * channel. If the user is not an admin the file must be located in 
         * the same channel as the user is currently participating in.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_FILETRANSFER_NOT_FOUND
         * - #ClientError ::CMDERR_OPENFILE_FAILED
         * - #ClientError ::CMDERR_FILE_NOT_FOUND
         * - #ClientError ::CMDERR_FILESHARING_DISABLED
         *
         * @param nChannelID The ID of the channel of where to get the file. Only 
         * admins can download in channel other then their own.
         * @param nFileID The ID of the file which is passed by #OnCmdFileNew.
         * @param szLocalFilePath The path of where to store the file, e.g. 
         * C:\\myfile.txt.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see OnCmdFileNew
         * @see GetChannelFiles */
        public int DoRecvFile(int nChannelID, int nFileID, string szLocalFilePath)
        {
            return TT4.TT_DoRecvFile(m_ttInst, nChannelID, nFileID, szLocalFilePath);
        }
        /**
         * @brief Delete a file from a channel. 
         *
         * A user is allowed to delete a file from a channel if either the
         * user is an admin, operator of the channel or owner of the
         * file. To be owner of the file a user must have a #BearWare.UserAccount
         * on the server.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED 
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_FILE_NOT_FOUND
         *
         * @param nChannelID The ID of the channel where the file is located.
         * @param nFileID The ID of the file to delete. The ID of the file which 
         * is passed by #OnCmdFileNew.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see OnCmdFileNew
         * @see GetChannelFiles */
        public int DoDeleteFile(int nChannelID, int nFileID)
        {
            return TT4.TT_DoDeleteFile(m_ttInst, nChannelID, nFileID);
        }
        /**
         * @brief Subscribe to user events and/or data.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED 
         * - #ClientError ::CMDERR_SUBSCRIPTIONS_DISABLED
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param nUserID The ID of the user this should affect.
         * @param uSubscriptions Union of #Subscription to subscribe to.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see Subscription */
        public int DoSubscribe(int nUserID, Subscription uSubscriptions)
        {
            return TT4.TT_DoSubscribe(m_ttInst, nUserID, uSubscriptions);
        }
        /**
         * @brief Unsubscribe to user events/data. This can be used to ignore messages
         * or voice data from a specific user.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED 
         * - #ClientError ::CMDERR_SUBSCRIPTIONS_DISABLED
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param nUserID The ID of the user this should affect.
         * @param uSubscriptions Union of #Subscription to unsubscribe.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see Subscription */
        public int DoUnsubscribe(int nUserID, Subscription uSubscriptions)
        {
            return TT4.TT_DoUnsubscribe(m_ttInst, nUserID, uSubscriptions);
        }
        /**
         * @brief Make a static (persistent) channel.
         * 
         * This command only works for admins since it creates a persistent
         * channel on the server which will be stored in the server's
         * config file.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_ALREADY_EXISTS
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND If channel's combined path is longer than
         *   #BearWare.TeamTalk4.TT_STRLEN.
         * - #ClientError ::CMDERR_INCORRECT_CHANNEL_PASSWORD If the password is longer than
         *   #BearWare.TeamTalk4.TT_STRLEN.
         * - #ClientError ::CMDERR_UNKNOWN_AUDIOCODEC If the server doesn't support the audio
         *   codec. Introduced in version 4.1.0.1264.
         *
         * @param lpChannel A Channel-structure containing information about
         * the channel being created. The Channel's member @a nChannelID is ignored.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoJoinChannel */
        public int DoMakeChannel(Channel lpChannel)
        {
            return TT4.TT_DoMakeChannel(m_ttInst, ref lpChannel);
        }
        /**
         * @brief Update a channel's properties.
         *
         * Admin and operators of channel can update a channel's
         * properties. Note that a channel's #BearWare.AudioCodec cannot be changed
         * if there's currently users in the channel.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED 
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_CHANNEL_HAS_USERS
         *
         * @param lpChannel A Channel-structure containing information about
         * the channel being modified. The channel member's \a nParentID
         * and \a szName are ignored.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoMakeChannel */
        public int DoUpdateChannel(Channel lpChannel)
        {
            return TT4.TT_DoUpdateChannel(m_ttInst, ref lpChannel);
        }
        /**
         * @brief Remove a channel from a server. 
         *
         * This command only applies to admins.
         *
         * If there's any users in the channel they will be kicked and
         * subchannels will be deleted as well.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         *
         * @param nChannelID The ID of the channel to remove.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoMakeChannel */
        public int DoRemoveChannel(int nChannelID)
        {
            return TT4.TT_DoRemoveChannel(m_ttInst, nChannelID);
        }
        /**
         * @brief Issue command to move a user from one channel to
         * another.
         *
         * This command only applies to admins.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param nUserID User to be moved.
         * @param nChannelID Channel where user should be put into.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoMoveUserByID */
        public int DoMoveUser(int nUserID, int nChannelID)
        {
            return TT4.TT_DoMoveUser(m_ttInst, nUserID, nChannelID);
        }
        /**
         * @brief Update server properties.
         *
         * This command only applies to admins.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_SERVER_HAS_USERS
         *
         * @param lpServerProperties A structure holding the information to be set 
         * on the server.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see GetServerProperties */
        public int DoUpdateServer(ServerProperties lpServerProperties)
        {
            return TT4.TT_DoUpdateServer(m_ttInst, ref lpServerProperties);
        }
        /**
         * @brief Issue command to list user accounts on the server.
         *
         * User accounts can be used to create different user types like
         * e.g. administrators.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         *
         * @param nIndex Index of first user to display.
         * @param nCount The number of users to retrieve.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see UserAccount
         * @see UserType
         * @see GetUserAccounts */
        public int DoListUserAccounts(int nIndex, int nCount)
        {
            return TT4.TT_DoListUserAccounts(m_ttInst, nIndex, nCount);
        }
        /**
         * @brief Issue command to create a new user account on the
         * server.
         *
         * Check out section @ref useradmin to see how the server handles
         * users.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_INVALID_ACCOUNT
         *
         * @param lpUserAccount The properties of the user account to create.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoListUserAccounts
         * @see DoDeleteUserAccount
         * @see UserAccount
         * @see UserType */
        public int DoNewUserAccount(UserAccount lpUserAccount)
        {
            return TT4.TT_DoNewUserAccount(m_ttInst, ref lpUserAccount);
        }
        /**
         * @brief Issue command to delete a user account on the server.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_ACCOUNT_NOT_FOUND
         *
         * @param szUsername The username of the user account to delete.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoListUserAccounts
         * @see DoNewUserAccount
         * @see BearWare.UserAccount
         * @see BearWare.UserType */
        public int DoDeleteUserAccount(string szUsername)
        {
            return TT4.TT_DoDeleteUserAccount(m_ttInst, szUsername);
        }
        /**
         * @brief Issue a ban command on a user in a specific channel. 
         *
         * The ban applies to the user's IP-address. Call #DoKickUser
         * to kick the user off the server. Only admins can ban users.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_CHANNEL_NOT_FOUND
         * - #ClientError ::CMDERR_USER_NOT_FOUND
         *
         * @param nUserID The ID of the user to ban.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoKickUser
         * @see DoListBans */
        public int DoBanUser(int nUserID)
        {
            return TT4.TT_DoBanUser(m_ttInst, nUserID);
        }
        /**
         * @brief Issue a ban command on an IP-address user. 
         *
         * Same as TeamTalk4.DoBanUser() except this command applies to IP-addresses
         * and therefore doesn't require a user to be logged in.
         *
         * Possible errors:
         * - #CMDERR_NOT_LOGGEDIN
         * - #CMDERR_NOT_AUTHORIZED
         *
         * @param szIpAddress The IP-address to ban.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see TeamTalk4.DoKickUser
         * @see TeamTalk4.DoListBans */
        public int DoBanIPAddress(string szIpAddress)
        {
            return TT4.TT_DoBanIPAddress(m_ttInst, szIpAddress);
        }

        /**
         * @brief Unban the user with the specified IP-address.
         *
         * Only admins can remove a ban.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_BAN_NOT_FOUND
         *
         * @param szIpAddress The IP-address to unban.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoBanUser
         * @see DoListBans
         * @see DoBanIPAddress */
        public int DoUnBanUser(string szIpAddress)
        {
            return TT4.TT_DoUnBanUser(m_ttInst, szIpAddress);
        }
        /**
         * @brief Issue a command to list the banned users.
         *
         * Only admins can list bans. Once completed call the function
         * #GetBannedUsers to get the list of users.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         *
         * @param nIndex Index of first ban to display.
         * @param nCount The number of bans to display.
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see TeamTalk4.GetBannedUsers */
        public int DoListBans(int nIndex, int nCount)
        {
            return TT4.TT_DoListBans(m_ttInst, nIndex, nCount);
        }
        /**
         * @brief Save the server's current state to its settings file (typically 
         * the server's .xml file).
         *
         * Note that the server only saves channels with the flag
         * #ChannelType ::CHANNEL_STATIC.
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         *
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error.
         * @see DoUpdateServer */
        public int DoSaveConfig()
        {
            return TT4.TT_DoSaveConfig(m_ttInst);
        }
        /**
         * @brief Get the server's current statistics obtained through
         * TeamTalk4.GetServerStatistics().
         *
         * Possible errors:
         * - #ClientError ::CMDERR_NOT_LOGGEDIN
         * - #ClientError ::CMDERR_NOT_AUTHORIZED
         * - #ClientError ::CMDERR_UNKNOWN_COMMAND
         *
         * @note
         * Requires server version 4.1.0.1089 or later.
         *
         * @return Returns command ID which will be passed in 
         * OnCmdProcessing() event when the server is processing the 
         * command. -1 is returned in case of error. */
        public int DoQueryServerStats()
        {
            return TT4.TT_DoQueryServerStats(m_ttInst);
        }
        /**
         * @brief Quit from server. 
         *
         * Possible errors:
         * - none
         *
         * This will generate a #OnConnectionLost since the server
         * will drop the client.
         *
         * @return Returns command ID which will be passed in 
         * #OnCmdProcessing event when the server is processing the 
         * command. -1 is returned in case of error. */
        public int DoQuit()
        {
            return TT4.TT_DoQuit(m_ttInst);
        }
        /** @} */

        /**
         * @addtogroup server
         * @{ */

        /** 
         *
         * @brief Get the server's properties.
         *
         * @param lpProperties A struct to hold the server's properties. */
        public bool GetServerProperties(out ServerProperties lpProperties)
        {
            return TT4.TT_GetServerProperties(m_ttInst, out lpProperties);
        }
        /**
         * @brief Get the server's statistics, i.e. bandwidth usage etc.
         *
         * @see DoQueryServerStatistics()
         * @param lpStatistics A struct to hold the server's statistics. */
        public bool GetServerStatistics(out ServerStatistics lpStatistics)
        {
            return TT4.TT_GetServerStatistics(m_ttInst, out lpStatistics);
        }
        /**
         * @brief Get the IDs of all the users on the server.
         *
         * Extracts the user IDs of all the users on the server. If
         * only users in a specific channel is needed call
         * GetChannelUsers()
         *
         * @param lpUserIDs An output array of integer which will receive the
         * user IDs of the users in the channel.
         * @see GetChannelUsers
         * @see GetUser */
        public bool GetServerUsers(out int[] lpUserIDs)
        {
            int count = 0;
            bool b = TT4.TT_GetServerUsers_NULL(m_ttInst, IntPtr.Zero, ref count);
            int[] users = new int[count];
            b = TT4.TT_GetServerUsers(m_ttInst, users, ref count);
            lpUserIDs = b ? users : null;
            return b;
        }
        /** @} */

        /** @addtogroup channels
         * @{ */

        /**
         * @brief Get the root channel's ID
         *
         * @return Returns the ID of the root channel. If 0 is returned no root 
         * channel exists.
         * @see GetMyChannelID
         * @see GetChannelPath */
        public int GetRootChannelID()
        {
            return TT4.TT_GetRootChannelID(m_ttInst);
        }
        /**
         * @brief Get the channel which the local client instance is
         * currently participating in.
         *
         * @return Returns the ID of the current channel. If 0 is returned the 
         * user is not participating in a channel. */
        public int GetMyChannelID()
        {
            return TT4.TT_GetMyChannelID(m_ttInst);
        }
        /**
         * @brief Same as GetMyChannelID().
         */
        public int ChannelID { get { return GetMyChannelID(); } }
        /**
         * @brief Get the channel with a specific ID.
         *
         * @param nChannelID The ID of the channel to get information about.
         * @param lpChannel A preallocated struct which will receive the 
         * channel's properties.
         * @return FALSE if unable to retrieve channel otherwise TRUE. */
        public bool GetChannel(int nChannelID, out Channel lpChannel)
        {
            return TT4.TT_GetChannel(m_ttInst, nChannelID, out lpChannel);
        }
        /**
         * @brief Get the channel's path. Channels are separated by '/'.
         *
         * @param nChannelID The channel's ID.
         * @param szChannelPath Will receive the channel's path.
         * @return Returns TRUE if channel exists. */
        public bool GetChannelPath(int nChannelID, out string szChannelPath)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            bool b = TT4.TT_GetChannelPath(m_ttInst, nChannelID, ptr);
            szChannelPath = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return b;
        }
        /**
         * @brief Get the channel ID of the supplied path. Channels are
         * separated by '/'
         *
         * @param szChannelPath Will receive the channel's path.
         * @return The channel's ID or 0 on error. */
        public int GetChannelIDFromPath(string szChannelPath)
        {
            return TT4.TT_GetChannelIDFromPath(m_ttInst, szChannelPath);
        }
        /**
         * @brief Get the IDs of all users in a channel.
         *
         * @param nChannelID The channel's ID.
         * @param lpUserIDs An output array of integer which will receive the
         * user IDs of the users in the channel.
         * @see BearWare.User 
         * @see GetChannel */
        public bool GetChannelUsers(int nChannelID, out int[] lpUserIDs)
        {
            int count = 0;
            bool b = TT4.TT_GetChannelUsers_NULL(m_ttInst, nChannelID, IntPtr.Zero, ref count);
            int[] users = new int[count];
            b = TT4.TT_GetChannelUsers(m_ttInst, nChannelID, users, ref count);
            lpUserIDs = b ? users : null;
            return b;
        }
        /**
         * @brief Get the list of the files in a channel which can be
         * downloaded.
         *
         * @param nChannelID The ID of the channel to extract the files from.
         * @param lpFileInfos An output array which will receive the file IDs.
         * @see GetChannelFileInfo */
        public bool GetChannelFiles(int nChannelID, out FileInfo[] lpFileInfos)
        {
            int count = 0;
            bool b = TT4.TT_GetChannelFiles_NULL(m_ttInst, nChannelID, IntPtr.Zero, ref count);
            FileInfo[] files = new FileInfo[count];
            b = TT4.TT_GetChannelFiles(m_ttInst, nChannelID, files, ref count);
            lpFileInfos = b ? files : null;
            return b;
        }
        /**
         * @brief Get information about a file which can be downloaded.
         *
         * Typically this is called after receiving
         * #OnCmdFileNew.
         *
         * @param nChannelID The ID of the channel to extract the file from.
         * @param nFileID The ID of the file.
         * @param lpFileInfo A preallocated struct which will receive 
         * file information. */
        public bool GetChannelFileInfo(int nChannelID, int nFileID, out FileInfo lpFileInfo)
        {
            return TT4.TT_GetChannelFileInfo(m_ttInst, nChannelID, nFileID, out lpFileInfo);
        }
        /**
         * @brief Check whether user is operator of a channel
         *
         * @param nUserID the ID of the user to check.
         * @param nChannelID the ID of the channel to check whether user
         * is operator of. */
        public bool IsChannelOperator(int nUserID, int nChannelID)
        {
            return TT4.TT_IsChannelOperator(m_ttInst, nUserID, nChannelID);
        }
        /** 
         * @brief Get the IDs of all the channels on the server.
         *
         * Use TT_GetChannel() to get more information about each of the
         * channels. 
         * @see TeamTalk4.GetServerUsers() */
        public bool GetServerChannels(out int[] lpChannelIDs)
        {
            int count = 0;
            bool b = TT4.TT_GetServerChannels_NULL(m_ttInst, IntPtr.Zero, ref count);
            int[] channels = new int[count];
            b = TT4.TT_GetServerChannels(m_ttInst, channels, ref count);
            lpChannelIDs = b ? channels : null;
            return b;
        }
        /** @} */

        /** @addtogroup users
         * @{ */

        /**
         * @brief Get the local client instance's user ID. 
         *
         * This information can be retrieved after the
         * #OnCmdMyselfLoggedIn event.
         *
         * @return Returns the user ID assigned to the current user on the server.
         * -1 is returned if no ID has been assigned to the user. */
        public int GetMyUserID()
        {
            return TT4.TT_GetMyUserID(m_ttInst);
        }

        /**
         * @brief Get the local client instance's #BearWare.UserAccount.
         *
         * This information can be retrieved after
         * TeamTalk4.OnCmdMyselfLoggedIn() event.
         *
         * @note
         * Requires server version 4.0.1.970 or later.
         * 
         * @param lpUserAccount The local client's user account registered on
         * the server. Note that the @a szPassword field of #BearWare.UserAccount
         * will not be set.
         * @see TeamTalk4.DoLogin */
        public bool GetMyUserAccount(out UserAccount lpUserAccount)
        {
            return TT4.TT_GetMyUserAccount(m_ttInst, out lpUserAccount);
        }

        /**
         * @brief Same as GetMyUserID().
         */
        public int UserID { get { return GetMyUserID(); } }

        /**
         * @brief Get the client instance's user type. 
         *
         * This information can be retrieved after
         * #OnCmdMyselfLoggedIn event.
         * 
         * @return A bitmask of the type of user based on #UserType.
         * @see DoLogin
         * @see BearWare.UserType
         * @see GetMyUserAccount */
        public UserType GetMyUserType()
        {
            return TT4.TT_GetMyUserType(m_ttInst);
        }

        /**
         * @brief Same as GetMyUserType().
         */
        public BearWare.UserType UserType { get { return GetMyUserType(); } }

        /**
         * @brief If an account was used in TeamTalk4.DoLogin() then this value will 
         * return the @a nUserData from the #BearWare.UserAccount.
         *
         * This information can be retrieved after
         * #OnCmdMyselfLoggedIn event.
         *
         * @return If set, @a nUserData from #BearWare.UserAccount,
         * otherwise 0. */
        public int GetMyUserData()
        {
            return TT4.TT_GetMyUserData(m_ttInst);
        }

        /**
         * @brief Same as GetMyUserData().
         */
        public int UserData { get { return GetMyUserData(); } }

        /**
         * @brief Get the user with the specified ID.
         *
         * @param nUserID The ID of the user to extract.
         * @param lpUser A preallocated #BearWare.User struct.
         * @see GetUserByUsername */
        public bool GetUser(int nUserID, out User lpUser)
        {
            return TT4.TT_GetUser(m_ttInst, nUserID, out lpUser);
        }
        /**
         * @brief Get statistics for data and packet reception from a user.
         *
         * @param nUserID The ID of the user to extract.
         * @param lpStats A preallocated #BearWare.UserStatistics struct. */
        public bool GetUserStatistics(int nUserID, out UserStatistics lpStats)
        {
            return TT4.TT_GetUserStatistics(m_ttInst, nUserID, out lpStats);
        }
        /**
         * @brief Get the user with the specified username.
         *
         * Remember to take into account that multiple users can log in
         * with the same account if ::USERRIGHT_DOUBLE_LOGIN is specified.
         * 
         * @param szUsername The user's username (from #BearWare.UserAccount).
         * @param lpUser A preallocated #BearWare.User struct. */
        public bool GetUserByUsername(string szUsername, out User lpUser)
        {
            return TT4.TT_GetUserByUsername(m_ttInst, szUsername, out lpUser);
        }

        /**
         * @brief Get a text-message sent by a user.
         *
         * The client instance uses a cyclic buffer in order not to drain
         * resources and can hold a maximum of 65535 text-messages.
         *
         * @param nMsgID The ID of the message to extract.
         * @param bRemoveMsg Clear the text-message to release resources.
         * @param lpTextMessage Will receive the content of the message extracted. 
         * Pass NULL if only called to remove the message. */
        public bool GetTextMessage(int nMsgID, bool bRemoveMsg, out TextMessage lpTextMessage)
        {
            return TT4.TT_GetTextMessage(m_ttInst, nMsgID, bRemoveMsg, out lpTextMessage);
        }
        /** @} */

        /** @addtogroup sounddevices
         * @{ */

        /**
         * @brief Set the volume of a user.
         *
         * Note that it's a virtual volume which is being set since the
         * master volume affects the user volume.
         * 
         * @param nUserID The ID of the user whose volume will be changed.
         * @param nVolume Must be between #BearWare.SoundLevel.SOUND_VOLUME_MIN and #BearWare.SoundLevel.SOUND_VOLUME_MAX.
         * @see SetSoundOutputVolume */
        public bool SetUserVolume(int nUserID, int nVolume)
        {
            return TT4.TT_SetUserVolume(m_ttInst, nUserID, nVolume);
        }
        /**
         * @brief Get the volume of a user. 
         *
         * Note that it's a virtual volume which is being set since the
         * master volume affects the user volume.
         *
         * @param nUserID The ID of the user.
         * @return A value between #BearWare.SoundLevel.SOUND_VOLUME_MIN and #BearWare.SoundLevel.SOUND_VOLUME_MAX */
        public int GetUserVolume(int nUserID)
        {
            return TT4.TT_GetUserVolume(m_ttInst, nUserID);
        }
        /**
         * @brief Use software to gain a user's volume.
         *
         * @param nUserID The ID of the user who should have sound gained.
         * @param nGainLevel The gain level for the user. A value from
         * #BearWare.SoundLevel.SOUND_GAIN_MIN to #BearWare.SoundLevel.SOUND_GAIN_MAX
         * @see GetUserGainLevel
         * @see SoundLevel.SOUND_GAIN_DEFAULT */
        public bool SetUserGainLevel(int nUserID, int nGainLevel)
        {
            return TT4.TT_SetUserGainLevel(m_ttInst, nUserID, nGainLevel);
        }
        /**
         * @brief Get the software gain level for a user.
         *
         * @param nUserID The ID of the user whose gain level should be retrieved.
         * @see SetUserGainLevel
         * @return The gain level for the user. A value from #BearWare.SoundLevel.SOUND_GAIN_MIN to 
         * #BearWare.SoundLevel.SOUND_GAIN_MAX */
        public int GetUserGainLevel(int nUserID)
        {
            return TT4.TT_GetUserGainLevel(m_ttInst, nUserID);
        }
        /**
         * @brief Mute a user.
         *
         * To stop receiving audio from a user call TeamTalk4.DoUnsubscribe().
         *
         * @param nUserID The user ID of the user to mute (or unmute).
         * @param bMute TRUE will mute, FALSE will unmute.
         * @see SetSoundOutputMute */
        public bool SetUserMute(int nUserID, bool bMute)
        {
            return TT4.TT_SetUserMute(m_ttInst, nUserID, bMute);
        }
        /**
         * @brief Set the delay of when a user should be considered to no
         * longer be talking.
         *
         * When a user starts talking the TeamTalk4.OnUserTalking() is
         * triggered with its parameter set to active. A user will remain
         * in this active state until no packets are received from this
         * user, plus a delay (due to network interruptions). This delay
         * is by default set to 500 msec but can be changed by calling
         * TeamTalk4.SetUserStoppedTalkingDelay().
         *
         * @see GetUserStoppedTalkingDelay */
        public bool SetUserStoppedTalkingDelay(int nUserID, 
                                               int nDelayMSec)
        {
            return TT4.TT_SetUserStoppedTalkingDelay(m_ttInst, nUserID, nDelayMSec);
        }
        /**
         * @brief Get the delay of when a user should no longer be
         * considered as talking.
         *
         * @return The delay in miliseconds. -1 on error.
         * @see SetUserStoppedTalkingDelay */
        public int GetUserStoppedTalkingDelay(int nUserID)
        {
            return TT4.TT_GetUserStoppedTalkingDelay(m_ttInst, nUserID);
        }

        /**
         * @brief Set the position of a user.
         *
         * This can only be done using DirectSound (#SoundSystem ::SOUNDSYSTEM_DSOUND)
         * and with sound duplex mode (#ClientFlag ::CLIENT_SNDINOUTPUT_DUPLEX)
         * disabled.
         *
         * @param nUserID ID of user.
         * @param x Distance in meters to user (left/right).
         * @param y Distance in meters to user (back/forward).
         * @param z Distance in meters to user (up/down). */
        public bool SetUserPosition(int nUserID, float x, float y, float z)
        {
            return TT4.TT_SetUserPosition(m_ttInst, nUserID, x, y, z);
        }
        /**
         * @brief Get a user's position.
         *
         * @param nUserID ID of user.
         * @param x Distance in meters to user (left/right).
         * @param y Distance in meters to user (back/forward).
         * @param z Distance in meters to user (up/down). */
        public bool GetUserPosition(int nUserID, out float x, out float y, out float z)
        {
            return TT4.TT_GetUserPosition(m_ttInst, nUserID, out x, out y, out z);
        }
        /**
         * @brief Set whether a user should speak in the left, right or
         * both speakers. This function only works if #BearWare.AudioCodec has been
         * set to use stereo.
         *
         * @param nUserID ID of user.
         * @param bLeftSpeaker TRUE if user should be played in left speaker.
         * @param bRightSpeaker TRUE if user should be played in right speaker.
         * @see GetUserStereo */
        public bool SetUserStereo(int nUserID, bool bLeftSpeaker, bool bRightSpeaker)
        {
            return TT4.TT_SetUserStereo(m_ttInst, nUserID, bLeftSpeaker, bRightSpeaker);
        }
        /**
         * @brief Check what speaker a user is outputting to.
         *
         * @param nUserID ID of user.
         * @param lpbLeftSpeaker TRUE if playing in left speaker.
         * @param lpbRightSpeaker TRUE if playing in right speaker.
         * @see SetUserStereo */
        public bool GetUserStereo(int nUserID, out bool lpbLeftSpeaker, out bool lpbRightSpeaker)
        {
            return TT4.TT_GetUserStereo(m_ttInst, nUserID, out lpbLeftSpeaker, out lpbRightSpeaker);
        }
        /**
         * @brief Store audio conversations to disk.
         * 
         * Set the path of where to store audio from a channel to disk. To
         * store in MP3 format instead of .wav format ensure that the LAME
         * MP3 encoder file lame_enc.dll is placed in the same directory
         * as the SDKs DLL files. To stop recording set @a szFolderPath
         * to an empty string and @a uAFF to #AudioFileFormat ::AFF_NONE.
         *
         * To store audio of other channels than the client instance check
         * out the section @ref spying.
         *
         * @param nUserID The ID of the #BearWare.User which should
         * store audio to disk.
         * @param szFolderPath The path on disk to where files should be stored.
         * This value will be stored in @a szAudioFolder of #BearWare.User.
         * @param szFileNameVars The file name used for audio files
         * can consist of the following variables: \%nickname\%,
         * \%username\%, \%userid\%, \%counter\% and a specified time
         * based on @c strftime (google @c 'strftime' for a
         * description of the format. The default format used by
         * TeamTalk is: '\%Y\%m\%d-\%H\%M\%S #\%userid\%
         * \%username\%'.  The \%counter\% variable is a 9 digit
         * integer which is incremented for each audio file. The file
         * extension is automatically appended based on the file type
         * (.wav for #AudioFileFormat ::AFF_WAVE_FORMAT and .mp3 for
         * AFF_MP3_*_FORMAT). Pass NULL or empty string to revert to
         * default format.
         * @param uAFF The #AudioFileFormat to use for storing audio files. Passing
         * #AudioFileFormat ::AFF_NONE will cancel/reset the current recording.
         * @return FALSE if path is invalid, otherwise TRUE.
         * @see BearWare.User
         * @see OnUserAudioFile */
        public bool SetUserAudioFolder(int nUserID, string szFolderPath, string szFileNameVars,
                                       AudioFileFormat uAFF)
        {
            return TT4.TT_SetUserAudioFolder(m_ttInst, nUserID, szFolderPath, szFileNameVars, uAFF);
        }
        /**
         * @brief Change the amount of media data which can be buffered
         * in the user's playback queue.
         * 
         * Increasing the media buffer size is especially important when
         * the user is currently streaming a media file using
         * TeamTalk4.StartStreamingMediaFileToChannel(). Once streaming has finished
         * it is recommended to reset the media buffer, i.e. setting it
         * to zero.
         *
         * A simple way to notify users that the client instance is streaming
         * a media file is to change the status of the local client instance
         * using TeamTalk4.DoChangeStatus() or to send a #BearWare.TextMessage using
         * TeamTalk4.DoTextMessage().
         *
         * @param nUserID The ID of the user who should have changed
         * the size of the media buffer.
         * @param nMSec The number of miliseconds of audio and video which
         * should be allowed to be stored in the media buffer. 2000 - 3000 
         * msec is a good size for a media buffer. Set the media
         * buffer size to 0 msec to reset the media buffer to its default value.
         */
        public bool SetUserMediaBufferSize(int nUserID, int nMSec)
        {
            return TT4.TT_SetUserMediaBufferSize(m_ttInst, nUserID, nMSec);
        }
        /** @brief Extract the raw audio from a user who has been talking.
         *
         * To enable access to user's raw audio first call
         * TeamTalk4.EnableAudioBlockEvent(). Whenever new audio becomes
         * available the event TeamTalk4.OnUserAudioBlock() is generated and 
         * TeamTalk4.AcquireUserAudioBlock() can be called to extract the audio.
         *
         * The #BearWare.AudioBlock contains shared memory with the local client
         * instance therefore always remember to call
         * TeamTalk4.ReleaseUserAudioBlock() to release the shared memory.
         *
         * If TeamTalk4.AcquireUserAudioBlock() is called multiple times without
         * calling TeamTalk4.ReleaseUserAudioBlock() then the same #BearWare.AudioBlock
         * will be retrieved again.
         *
         * @see TeamTalk4.ReleaseUserAudioBlock()
         * @see TeamTalk4.EnableAudioBlockEvent()
         * @see TeamTalk4.OnUserAudioBlock() */
        public bool AcquireUserAudioBlock(int nUserID, ref AudioBlock lpAudioBlock)
        {
            return TT4.TT_AcquireUserAudioBlock(m_ttInst, nUserID, ref lpAudioBlock); 
        }
        /** 
         * @brief Release the shared memory of an #BearWare.AudioBlock.
         *
         * All #BearWare.AudioBlock-structures extracted through
         * TeamTalk4.AcquireUserAudioBlock() must be released again since they
         * share memory with the local client instance.
         *
         * Never access @c lpRawAudio after releasing its
         * #BearWare.AudioBlock. This will cause the application to crash with a
         * memory exception.
         *
         * @see TeamTalk4.AcquireUserAudioBlock()
         * @see TeamTalk4.OnUserAudioBlock() */
        public bool ReleaseUserAudioBlock(int nUserID)
        {
            return TT4.TT_ReleaseUserAudioBlock(m_ttInst, nUserID);
        }
        /**
         * @brief Release all audio blocks of the local client instance.
         *
         * This function is only for convenience to ensure that no memory
         * is leaked. Normally TeamTalk4.ReleaseUserAudioBlock() should be used
         * to release #BearWare.AudioBlock-data.
         * @see TeamTalk4.ReleaseUserAudioBlock()
         * @see TeamTalk4.OnUserAudioBlock() */
        public bool ReleaseAllUserAudioBlocks()
        {
            return TT4.TT_ReleaseAllAudioBlocks(m_ttInst);
        }
        /** @} */

        /** @ingroup channels
         * @brief Get information about an active file transfer.  
         *
         * An active file transfer is one which has been posted through the
         * event #OnFileTransfer.
         *
         * @param nTransferID The ID of the file transfer to investigate. Transfer
         * ID is passed by #OnFileTransfer.
         * @param lpFileTransfer A preallocated struct which will receive the file 
         * transfer information.
         * @see CancelFileTransfer */
        public bool GetFileTransferInfo(int nTransferID, out FileTransfer lpFileTransfer)
        {
            return TT4.TT_GetFileTransferInfo(m_ttInst, nTransferID, out lpFileTransfer);
        }

        /** @ingroup channels
         * @brief Cancel an active file transfer. 
         *
         * An active file transfer is one which has been post through the
         * event #OnFileTransfer.
         *
         * @param nTransferID The ID of the file transfer to investigate. Transfer 
         * ID is passed by #OnFileTransfer. */
        public bool CancelFileTranfer(int nTransferID)
        {
            return TT4.TT_CancelFileTransfer(m_ttInst, nTransferID);
        }

        /** @ingroup server
         * @brief Get the list of banned users.
         * 
         * After the command #DoListBans has completed, this function
         * can be called to retrieve the list of banned users. The list of
         * banned users can only be retrieved once after which the
         * internal representation of the users is deleted (to save
         * memory).
         *
         * @param lpBannedUsers An output array which will receive the banned 
         * users.
         * @see DoBanUser  */
        public bool GetBannedUsers(out BannedUser[] lpBannedUsers)
        {
            int count = 0;
            bool b = TT4.TT_GetBannedUsers_NULL(m_ttInst, IntPtr.Zero, ref count);
            BannedUser[] banned = new BannedUser[count];
            b = TT4.TT_GetBannedUsers(m_ttInst, banned, ref count);
            lpBannedUsers = b ? banned : null;
            return b;
        }
        /** @ingroup server
         * @brief Get the list of user accounts.
         * 
         * After the command #DoListUserAccounts has competed, this
         * function can be called to retrieve the list of user accounts on
         * the server. The list of user accounts can only be retrieved
         * once after which the internal representation of the users is
         * deleted (to save memory).
         *
         * @param lpUserAccounts An output array which will receive the
         * user accounts on the server.
         * @see DoListUserAccounts
         * @see UserAccount */
        public bool GetUserAccounts(out UserAccount[] lpUserAccounts)
        {
            int count = 0;
            bool b = TT4.TT_GetUserAccounts_NULL(m_ttInst, IntPtr.Zero, ref count);
            UserAccount[] accounts = new UserAccount[count];
            b = TT4.TT_GetUserAccounts(m_ttInst, accounts, ref count);
            lpUserAccounts = b ? accounts : null;
            return b;
        }

        /** @ingroup errorhandling
         * @brief Get textual discription of an error message.
         * 
         * Get a description of an error code posted by either
         * #OnCmdError or #OnInternalError.
         *
         * @param nError The number of the error.
         * @return A text description of the error.
         * @see OnCmdError
         * @see OnInternalError */
        public static string GetErrorMessage(ClientError nError)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            TT4.TT_GetErrorMessage((int)nError, ptr);
            string s = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return s;
        }

        /** @addtogroup hotkey
         * @{ */

        /**
         * @brief Register a global hotkey. 
         *
         * When the hotkey becomes active or inactive it will post the event
         * #OnHotKeyToggle.
         *
         * A hotkey can e.g. be used as a push-to-talk key
         * combination. When the hotkey becomes active call
         * #EnableTransmission.
         *
         * Note that having a hotkey enabled makes the Visual Studio
         * debugger really slow to respond, so when debugging it's best
         * not to have hotkeys enabled.
         *
         * @param nHotKeyID The ID of the hotkey to register. It will be
         * passed as the event #OnHotKeyToggle when the hotkey becomes either active or inactive.
         * @param lpnVKCodes An array of virtual key codes which constitute the
         * hotkey. This document outlines the virtual key codes:
         * http://msdn.microsoft.com/en-us/library/ms645540(VS.85).aspx
         * A hotkey consisting of Left Control+A would have the array consist of 
         * [162, 65].
         * @see HotKey_Unregister
         * @see HotKey_InstallTestHook */
        public bool HotKey_Register(int nHotKeyID, int[] lpnVKCodes)
        {
            return TT4.TT_HotKey_Register(m_ttInst, nHotKeyID, lpnVKCodes, lpnVKCodes.Length);
        }
        /**
         * @brief Unregister a registered hotkey.
         *
         * @param nHotKeyID is the ID of the hotkey to unregister.
         * @see HotKey_Register */
        public bool HotKey_Unregister(int nHotKeyID)
        {
            return TT4.TT_HotKey_Unregister(m_ttInst, nHotKeyID);
        }
        /**
         * @brief Check whether hotkey is active.
         *
         * @param nHotKeyID is the ID of the registered hotkey. 
         * @return 1 if active, 0 if inactive, -1 if hotkey-ID is invalid */
        public int HotKey_IsActive(int nHotKeyID)
        {
            return TT4.TT_HotKey_IsActive(m_ttInst, nHotKeyID);
        }
        /**
         * @brief Install a test hook which calls the event #OnHotKeyTest
         * whenever a key or mouse button is pressed.
         * 
         * This can be used to determine a key-combination, like e.g.
         * a push-to-talk key.
         * 
         * @see HotKey_RemoveTestHook
         * @see OnHotKeyTest */
        public bool HotKey_InstallTestHook()
        {
            if (m_eventhandler != null)
            {
#if PocketPC
                IntPtr hWnd = m_eventhandler.Hwnd;
#else
                IntPtr hWnd = m_eventhandler.Handle;
#endif
                return TT4.TT_HotKey_InstallTestHook(m_ttInst, hWnd);
            }
            return false;
        }
        /**
         * @brief Remove the test hook again so the #OnHotKeyTest event
         * will no longer be notified.
         *
         * @see HotKey_InstallTestHook */
        public bool HotKey_RemoveTestHook()
        {
            return TT4.TT_HotKey_RemoveTestHook(m_ttInst);
        }
        /**
         * @brief Get a string description of the virtual-key code.
         *
         * @param nVKCode The virtual key code passed in #OnHotKeyTest.
         * @param szKeyName Will receive key description in local language.
         * @see HotKey_InstallTestHook */
        public bool HotKey_GetKeyString(int nVKCode, out string szKeyName)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            bool b = TT4.TT_HotKey_GetKeyString(m_ttInst, nVKCode, ptr);
            szKeyName = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return b;
        }
        /** @} */

        /** @addtogroup events
         * @{ */

        /**
         * @brief Delegate for events #OnConnectSuccess, 
         * #OnConnectFailed and #OnConnectionLost. */
        public delegate void Connection();
        /**
         * @brief Connected successfully to the server.
         *
         * This event is posted if #Connect was successful.
         *
         * #DoLogin can now be called in order to logon to the
         * server.
         * 
         * @see Connection Delegate for this event.
         * @see Connect
         * @see DoLogin */
        public event Connection OnConnectSuccess;
        /** 
         * @brief Failed to connect to server.
         *
         * This event is posted if #Connect fails. Ensure to call
         * #Disconnect before calling #Connect again.
         *
         * @see Connection Delegate for this event.
         * @see Connect */
        public event Connection OnConnectFailed;
        /** 
         * @brief Connection to server has been lost.
         *
         * The server is not responding to requests from the local
         * client instance and the connection has been dropped. To
         * change when the client instance should regard the server as
         * unavailable call #SetServerTimeout. 
         *
         * GetStatistics() can be used to check when data was
         * last received from the server.
         *
         * Ensure to call #Disconnect before calling #Connect
         * again.
         *
         * @see Connection Delegate for this event.
         * @see Connect */
        public event Connection OnConnectionLost;

        /** @brief Delegate for event #OnConnectionP2P
         * @param nUserID User ID
         * @param bSuccess TRUE if P2P connection was successful, FALSE if P2P
         * connection failed. */
        public delegate void ConnectionP2P(int nUserID, bool bSuccess);
        /** 
         * @brief Peer to peer (p2p) status changed.
         *
         * Audio and video packets can be sent to the user without
         * forwarding through the server. This is done by enabling P2P
         * network using #EnablePeerToPeer. Once the client instance
         * joins a new channel it will try and create a peer to peer
         * connection to each user in the channel. The client instance
         * will try to create a peer to peer connection for 5 seconds.
         *
         * Read section @ref txforward and section @ref txp2p on the
         * different ways of transmitting data.
         * 
         * @see ConnectionP2P Delegate for this event.
         * @see UserRight ::USERRIGHT_FORWARD_AUDIO
         * @see UserRight ::USERRIGHT_FORWARD_VIDEO */
        public event ConnectionP2P OnConnectionP2P;

        /**
         * @brief Delegate for event #OnCmdProcessing.
         * @param nCmdID Command ID being processed (returned by @c Do* 
         * commands)
         * @param bComplete Is 0 if command ID started processing and 1
         * if the command has finished processing. */
        public delegate void CommandProcessing(int nCmdID, bool bComplete);
        /** 
         * @brief A command issued by @c Do* methods is being
         * processed.
         *
         * Read section @ref cmdprocessing on how to use command
         * processing in the user application.
         * 
         * @see CommandProcessing Delegate for this event. */
        public event CommandProcessing OnCmdProcessing;
        /** @brief Delegate for event #OnCmdMyselfLoggedIn
         * @param nMyUserID The client instance's user ID, i.e. what can now 
         * be retrieved through #GetMyUserID. */
        public delegate void MyselfLoggedIn(int nMyUserID);
        /** 
         * @brief The client instance successfully logged on to
         * server.
         *
         * The call to #DoLogin was successful and all channels on
         * the server will be posted in the event
         * #OnCmdChannelNew immediately following this
         * event. If #UserRight ::USERRIGHT_VIEW_ALL_USERS is enabled the client
         * instance will also receive the events
         * #OnCmdUserLoggedIn and
         * #OnCmdUserJoinedChannel for every user on the server.
         * 
         * @see MyselfLoggedIn Delegate for this event.
         * @see DoLogin */
        public event MyselfLoggedIn OnCmdMyselfLoggedIn;
        /** @brief Delegate for event #OnCmdMyselfLoggedOut */
        public delegate void MyselfLoggedOut();
        /** 
         * @brief The client instance logged out of a server.  
         *
         * A response to #DoLogout.
         * 
         * @see MyselfLoggedOut Delegate for this event.
         * @see DoLogout */
        public event MyselfLoggedOut OnCmdMyselfLoggedOut;
        /** @brief Delegate for event #OnCmdMyselfJoinedChannel
         * @param nChannelID Channel's ID */
        public delegate void MyselfJoinedChannel(int nChannelID);
        /** 
         * @brief The client instance has joined a new channel. 
         *
         * Result of command #DoJoinChannel or
         * #DoJoinChannelByID. Can also be a result of an
         * administrator calling #DoMoveUser.
         *
         * If #UserRight ::USERRIGHT_VIEW_ALL_USERS is disabled the client instance
         * will afterwards receive the #OnCmdUserJoinedChannel
         * event for each of the users in the channel.
         * 
         * @see MyselfJoinedChannel Delegate for this event.
         * @see OnCmdMyselfLeftChannel */
        public event MyselfJoinedChannel OnCmdMyselfJoinedChannel;
        /** @brief Delegate for event #OnCmdMyselfLeftChannel.
         * @param nChannelID Channel's ID */
        public delegate void MyselfLeftChannel(int nChannelID);
        /** 
         * @brief The client instance left a channel. 
         * 
         * @see MyselfLeftChannel Delegate for this event.
         * @see OnCmdMyselfJoinedChannel
         * @see DoLeaveChannel */
        public event MyselfLeftChannel OnCmdMyselfLeftChannel;
        /** @brief Delegate for event #OnCmdMyselfKickedFromChannel.
         * @param nUserID User ID of the kicker. */
        public delegate void MyselfKickedFromChannel(int nUserID);
        /** 
         * @brief The client instance was kicked from a channel.
         * 
         * @see MyselfKickedFromChannel Delegate for this event. */
        public event MyselfKickedFromChannel OnCmdMyselfKickedFromChannel;
        /** @brief Delegate for events #OnCmdUserLoggedIn and #OnCmdUserLoggedOut.
         * @param nUserID The user's ID. */
        public delegate void UserAuth(int nUserID);
        /**
         * @brief A new user logged on to the server.
         *
         * Use #GetUser to get the properties of the user.
         *
         * @see UserAuth Delegate for this event.
         * @see DoLogin
         * @see GetUser To retrieve user.
         * @see OnCmdUserLoggedOut */
        public event UserAuth OnCmdUserLoggedIn;
        /**
         * @brief A client logged out of the server. 
         *
         * This event is called when a user logs out with
         * #DoLogout or disconnects with #Disconnect.
         *
         * @see UserAuth Delegate for this event.
         * @see DoLogout
         * @see Disconnect
         * @see OnCmdUserLoggedIn */
        public event UserAuth OnCmdUserLoggedOut;
        /** @brief Delegate for events #OnCmdUserUpdate, #OnCmdUserJoinedChannel and #OnCmdUserLeftChannel.
         * @param nUserID User's ID
         * @param nChannelID Channel ID. 0 for no channel. 
         * @see GetUser To retrieve user. */
        public delegate void UserUpdate(int nUserID, int nChannelID);
        /**
         * @brief User changed properties.
         *
         * 
         * @see UserUpdate Delegate for this event. */
        public event UserUpdate OnCmdUserUpdate;
        /** 
         * @brief A user has joined a channel.
         *
         * @see UserUpdate Delegate for this event. */
        public event UserUpdate OnCmdUserJoinedChannel;
        /** 
         * @brief User has left a channel.
         *
         * @see UserUpdate Delegate for this event. */
        public event UserUpdate OnCmdUserLeftChannel;
        /** @brief Delegate for events #OnCmdUserTextMessage.
         * 
         * @param nUserID The user's user ID.
         * @param nMsgID The message's ID.
         *
         * @see GetTextMessage To retrieve message.
         * @see GetUser To retrieve user.
         * @see DoTextMessage() To send a text message. */
        public delegate void UserTextMessage(int nUserID, int nMsgID);
        /** 
         * @brief A user has sent a text-message.
         *
         * @see UserTextMessage Delegate for this event. */
        public event UserTextMessage OnCmdUserTextMessage;
        /** @brief Delegate for #OnCmdChannelNew, #OnCmdChannelUpdate and #OnCmdChannelRemove.
         * 
         * @param nChannelID Channel's ID.
         * @see GetChannel To retrieve channel. */
        public delegate void ChannelUpdate(int nChannelID);
        /** 
         * @brief A new channel has been created.
         * 
         * @see ChannelUpdate Delegate for this event. */
        public event ChannelUpdate OnCmdChannelNew;
        /** 
         * @brief A channel's properties has been updated.
         * 
         * @see ChannelUpdate Delegate for this event. */
        public event ChannelUpdate OnCmdChannelUpdate;
        /** 
         * @brief A channel has been removed.
         *
         * Note that calling the #GetChannel with the channel ID
         * will fail because the channel is no longer there.
         * 
         * @see ChannelUpdate Delegate for this event. */
        public event ChannelUpdate OnCmdChannelRemove;
        /** @brief Delegate for event #OnCmdServerUpdate.
         * @see GetServerProperties To retrieve new server properties. */
        public delegate void ServerUpdate();
        /** 
         * @brief Server has updated its settings (server name, MOTD,
         * etc.)
         * 
         * Use #GetServerProperties to get the new server
         * properties.
         * 
         * @see ServerUpdate Delegate for this event. */
        public event ServerUpdate OnCmdServerUpdate;
        /** @brief A delegate for events #OnCmdFileNew and #OnCmdFileRemove.
         * @param nFileID File ID.
         * @param nChannelID Channel ID.
         * @see GetChannelFileInfo To retrieve file. */
        public delegate void FileUpdate(int nFileID, int nChannelID);
        /** 
         * @brief A new file is added to a channel. 
         *
         * Use #GetChannelFileInfo to get information about the
         * file.
         *
         * @see FileUpdate Delegate for this event. */
        public event FileUpdate OnCmdFileNew;
        /** 
         * @brief A file has been removed from a channel.
         *
         * @see FileUpdate Delegate for this event. */
        public event FileUpdate OnCmdFileRemove;
        /** @brief Delegate for event #OnCmdError.
         * 
         * @param nErrorNo Error number
         * @param nCmdID The command ID returned from the @c Do* commands. */
        public delegate void CommandError(ClientError nErrorNo, int nCmdID);
        /** 
         * @brief The server rejected a command issued by the local
         * client instance.
         *
         * To figure out which command failed use the command ID
         * returned by the @c Do* command. Section @ref cmdprocessing
         * explains how to use command ID.
         *
         * @see CommandError Delegate for this event.
         * @see GetErrorMessage */
        public event CommandError OnCmdError;
        /** @brief Delegate for event #CommandSuccess.
         * 
         * @param nCmdID The command ID returned from the @c Do* commands. */
        public delegate void CommandSuccess(int nCmdID);
        /**
         * @brief The server successfully processed  a command issued by the local
         * client instance.
         *
         * To figure out which command succeeded use the command ID
         * returned by the @c Do* command. Section @ref cmdprocessing
         * explains how to use command ID.
         *
         * @see CommandError Delegate for this event. */
        public event CommandSuccess OnCmdSuccess;
        /** @brief Delegate for event #OnUserTalking.
         * 
         * @param nUserID User's ID.
         * @param bTalking TRUE if talking otherwise FALSE.
         * @see IsTransmitting To see if "myself" is transmitting. */
        public delegate void UserTalking(int nUserID, bool bTalking);
        /**
         * @brief A user is talking.
         *
         * Playback using the sound output device has started for a
         * user.
         * 
         * @see SetUserStoppedTalkingDelay()
         * @see UserTalking Delegate for this event. */
        public event UserTalking OnUserTalking;
        /** @brief A delegate for the event #OnUserVideoFrame.
         * 
         * @param nUserID User's ID.
         * @param nFrameQueueSize Number of video frames currently in queue for
         * display. The client uses a cyclic buffer for video frame in
         * order to prevent resources from being drained. Therefore
         * the #OnUserVideoFrame event might be posted
         * more times than there actually are frames available. So use
         * the frame count to ensure the lastest frame is always
         * displayed.
         * 
         * @see GetUserVideoFrame To retrieve video frame. */
        public delegate void UserVideoFrame(int nUserID, int nFrameQueueSize);
        /** 
         * @brief A new video frame was received from a user.
         *
         * Use #GetUserVideoFrame to display the image.
         * 
         * @see UserVideoFrame Delegate for this event. */
        public event UserVideoFrame OnUserVideoFrame;
        /** @brief Delegate for event #OnUserAudioFile.
         * 
         * @param nUserID The user's ID.
         * @param nStatus The status of the audio file. */
        public delegate void UserAudioFile(int nUserID, AudioFileStatus nStatus);
        /** 
         * @brief An audio file recording has changed status.
         *
         * SetUserAudioFolder() makes the client instance store all
         * audio from a user to a specified folder. Every time an
         * audio file is being processed this event is posted.
         *
         * @see UserAudioFile Delegate for this event. */
        public event UserAudioFile OnUserAudioFile;
        /** @brief Delegate for event #OnInternalError.
         * @param nErrorNo An error number based on #ClientError. The value
         * will be of the type #ClientError @c INTERR_*. */
        public delegate void ErrorOccured(ClientError nErrorNo);
        /** 
         * @brief A sound device failed to initialize. 
         *
         * This can e.g. happen if a new user joins a channel and
         * there is no sound output channels
         * available. nMaxOutputChannels of #BearWare.SoundDevice struct tells
         * how many streams can be active simultaneously.
         *
         * @see ErrorOccured Delegate for this event.
         * @see OnCmdMyselfJoinedChannel If sound input device fails it
         * will be when joining a new channel.
         * @see GetSoundOutputDevices
         * @see SoundDevice */
        public event ErrorOccured OnInternalError;
        /** @brief Delegate for event #OnVoiceActivation.
         * @param bVoiceActive TRUE if enabled, FALSE if disabled. */
        public delegate void VoiceActivation(bool bVoiceActive);
        /** 
         * @brief Voice activation has triggered transmission.
         *
         * @see VoiceActivation Delegate for this event.
         * @see SetVoiceActivationLevel
         * @see ClientFlag ::CLIENT_SNDINPUT_VOICEACTIVATED
         * @see EnableTransmission */
        public event VoiceActivation OnVoiceActivation;
        /** @brief Delegate for event #OnStreamAudioFileUser.
         * @param nUserID User's ID of where the audio file is streamed to.
         * @param nStatus The status of the audio file. */
        public delegate void StreamUserAudioFile(int nUserID, AudioFileStatus nStatus);
        /** 
         * @brief An audio file being streamed to a user is
         * processing.
         *
         * This event is called as a result of
         * #StartStreamingAudioFileToUser. 
         * @see StreamUserAudioFile Delegate for this event. */
        public event StreamUserAudioFile OnStreamAudioFileUser;
        /** @brief Delegate for event #OnStreamAudioFileChannel.
         * 
         * @param nChannelID Channel's ID of where the audio file is being 
         * streamed to.
         * @param nStatus The status of the audio file. */
        public delegate void StreamChannelAudioFile(int nChannelID, AudioFileStatus nStatus);
        /** 
         * @brief Audio file being stream to a channel is processing.
         *
         * This event is called as a result of
         * #StartStreamingAudioFileToChannel.
         *
         * @see StreamChannelAudioFile Delegate for this event.
         * @see StartStreamingAudioFileToChannel */
        public event StreamChannelAudioFile OnStreamAudioFileChannel;
        /** @brief Delegate for event #OnHotKeyToggle. 
         *
         * @param nHotKeyID The hotkey ID passed to #HotKey_Register
         * @param bActive TRUE when hotkey is active and FALSE when 
         * it becomes inactive. */
        public delegate void HotKeyToggle(int nHotKeyID, bool bActive);
        /** 
         * @brief A hotkey has been acticated or deactivated.
         *
         * @see HotKeyToggle Delegate for this event.
         * @see HotKey_Register
         * @see HotKey_Unregister */
        public event HotKeyToggle OnHotKeyToggle;
        /** @brief Delegate for event #OnHotKeyTest.
         * @param nVkCode The virtual key code. Look here for a list of virtual
         * key codes: http://msdn.microsoft.com/en-us/library/ms645540(VS.85).aspx
         * @param bActive TRUE when key is down and FALSE when released. */
        public delegate void HotKeyTest(int nVkCode, bool bActive);
        /**
         * @brief A button was pressed or released on the user's
         * keyboard or mouse.
         * 
         * When #HotKey_InstallTestHook is called a hook is
         * installed in Windows which intercepts all keyboard and
         * mouse presses. Every time a key or mouse is pressed or
         * released this event is posted.
         *
         * Use #HotKey_GetKeyString to get a key description of the 
         * pressed key.
         * 
         * @see HotKeyTest Delegate for this event.
         * @see HotKey_InstallTestHook */
        public event HotKeyTest OnHotKeyTest;
        /** @brief Delegate for event #OnFileTransfer.
         * @param nTransferID Transfer ID
         * @param nStatus The #BearWare.FileTransfer's status. */
        public delegate void FileTransferUpdate(int nTransferID, FileTransferStatus nStatus);
        /**
         * @brief A file transfer is processing. 
         *
         * Use #GetFileTransferInfo to get information about the
         * file transfer. Ensure to check if the file transfer is
         * completed, because the file transfer instance will be
         * removed from the client instance when the user application
         * reads the #BearWare.FileTransfer object and it has completed the
         * transfer.
         *
         * @see FileTransferUpdate Delegate for this event.
         * @see GetFileTransferInfo To retrieve #BearWare.FileTransfer. */
        public event FileTransferUpdate OnFileTransfer;

        /** @brief Delegate for event #OnUserAudioBlock.
         * @param nUserID The user ID of the user who has an #BearWare.AudioBlock. */
        public delegate void NewAudioBlock(int nUserID);
        /**
         * @brief A new audio block can be extracted.
         *
         * This event is only generated if TeamTalk4.EnableAudioBlockEvent()
         * is first called.
         *
         * Call TeamTalk4.AcquireUserAudioBlock() to extract the #BearWare.AudioBlock. */
        public event NewAudioBlock OnUserAudioBlock;

        /** @brief Delegate for event #OnUserDesktopWindow.
         *
         * @param nUserID The user's ID.
         * @param nSessionID The ID of the desktop window's session. If
         * this ID changes it means the user has started a new
         * session. If the session ID becomes 0 it means the desktop
         * session has been closed by the user.
         * @see TeamTalk4.SendDesktopWindow() */
        public delegate void NewDesktopWindow(int nUserID, int nSessionID);
        /**
         * @brief A new or updated desktop window has been received
         * from a user.
         *
         * Use TeamTalk4.GetUserDesktopWindow() to retrieve the bitmap of the
         * desktop window.
         */
        public event NewDesktopWindow OnUserDesktopWindow;

        /** 
         * @brief Delegate for event OnDesktopWindowTransfer().
         *
         * @param nSessionID The desktop session's ID. If the desktop session ID
         * becomes 0 it means the desktop session has been closed and/or
         * cancelled.
         * @param nTransferRemaining The number of bytes remaining before transmission
         * of last desktop window completes. When remaining bytes is 0 
         * TeamTalk4.SendDesktopWindow() can be called again. */
        public delegate void DesktopTransferUpdate(int nSessionID, int nTransferRemaining);

        /**
         * @brief Used for tracking when a desktop window has been
         * transmitted to the server.
         *
         * When the transmission has completed the flag
         * #ClientFlag ::CLIENT_TX_DESKTOP will be cleared from the
         * local client instance. */
        public event DesktopTransferUpdate OnDesktopTransferUpdate;

        /** @brief Delegate for event OnUserDesktopCursorPosition().
         *
         * @param nSrcUserID The user ID of the owner of the mouse cursor.
         * @param nDestUserID The owner of the desktop session the mouse cursor
         * is pointing to. */
        public delegate void UserDesktopCursorPosition(int nSrcUserID, int nDestUserID);
        /**
         * @brief A user has sent the position of the mouse cursor.
         *
         * Use TeamTalk4.SendDesktopCursorPosition() to send the position of
         * the mouse cursor. */
        public event UserDesktopCursorPosition OnUserDesktopCursorPosition;

        /** @brief Delegate for event OnMaxPayloadUpdate().
         * 
         * @param nUserID The user's ID. 0 means server's maximum payload
         * size.
         * @param nMaxPayload The maximum size in bytes of the payload data which
         * is put in UDP packets. 0 means the max payload query failed.
         * @see TeamTalk4.QueryMaxPayload() */
        public delegate void MaxPayloadUpdate(int nUserID, int nMaxPayload);

        /**
         * @brief The maximum size of the payload put into UDP packets
         * has been updated.
         *
         * @see TeamTalk4.QueryMaxPayload() */
        public event MaxPayloadUpdate OnMaxPayloadUpdate;

        /** @brief Delegate for event #OnStreamMediaFileChannel.
         * 
         * @param nStatus The status of the audio file. */
        public delegate void StreamChannelMediaFile(AudioFileStatus nStatus);

        /**
         * @brief Media file being streamed to a channel is processing.
         *
         * This event is called as a result of
         * TeamTalk4.StartStreamingMediaFileToChannel(). */
        public event StreamChannelMediaFile OnStreamMediaFileChannel;

        /** @brief Delegate for event #OnUserDesktopInput.
         * 
         * @param nUserID The User ID of the who sent desktop input. */
        public delegate void UserDesktopInput(int nUserID);

        /**
         * @brief Desktop input (mouse or keyboard input) has been
         * received from a user.
         *
         * This event is generated if a remote user has called
         * #BearWare.TeamTalk4.SendDesktopInput(). In order for the local client
         * instance to received desktop input it must have enabled the
         * subscription ::SUBSCRIBE_DESKTOPINPUT.
         *
         * Call #BearWare.TeamTalk4.GetUserDesktopInput() to obtain the mouse or
         * keyboard key-codes.
         */
        public event UserDesktopInput OnUserDesktopInput;

        /** @} */

        //TT4 instance
        private IntPtr m_ttInst;
        //TT4 event handler
        c_tt.MyEventHandler m_eventhandler;
    }

    /** @addtogroup desktopshare
     * @{ */

    /**
     * @brief Translate to and from TeamTalk's intermediate key-codes (TTKEYCODE).
     * @see TeamTalk4.OnUserDesktopInput() */
    public enum TTKeyTranslate : uint
    {
        /** @brief Perform no translation. */
        TTKEY_NO_TRANSLATE                  = 0,
        /** @brief Translate from Windows scan-code to TTKEYCODE. The
         * Windows scan-code can be retrieved in Windows' @c
         * WM_KEYDOWN and @c WM_KEYUP event. */
        TTKEY_WINKEYCODE_TO_TTKEYCODE       = 1,
        /** @brief Translate from TTKEYCODE to Windows scan-code. */
        TTKEY_TTKEYCODE_TO_WINKEYCODE       = 2,
        /** @brief Translate from Mac OS X Carbon @c kVK_* key-code to
         * TTKEYCODE. The Mac OS X key-codes are defined in Carbon's
         * API. */
        TTKEY_MACKEYCODE_TO_TTKEYCODE       = 3,
        /** @brief Translate from TTKEYCODE to Mac OS X Carbon @c
         * kVK_* key-code. */
        TTKEY_TTKEYCODE_TO_MACKEYCODE       = 4
    }

    /** @} */

    /** @addtogroup mixer
     * @{ */

    /**
     * @brief The Windows mixer controls which can be queried by the
     * #BearWare.WindowsMixer class.
     *
     * Wave-In devices which are not in the enum-structure can be
     * accessed by WindowsMixer.GetWaveInControlCount() which allows the user to
     * query selection based on an index.
     *
     * Note that Windows Vista has deprecated mixer controls.
     *
     * @see WindowsMixer.SetWaveOutMute
     * @see WindowsMixer.SetWaveOutVolume
     * @see WindowsMixer.SetWaveInSelected
     * @see WindowsMixer.SetWaveInVolume
     * @see WindowsMixer.GetWaveInControlName
     * @see WindowsMixer.SetWaveInControlSelected */
    public enum MixerControl : uint
    {
        WAVEOUT_MASTER,
        WAVEOUT_WAVE,
        WAVEOUT_MICROPHONE,

        WAVEIN_MICROPHONE,
        WAVEIN_LINEIN,
        WAVEIN_WAVEOUT,
    }

    /**
     * @brief A wrapper for the Windows Mixer API.
     */
    public class WindowsMixer
    {
        /**
         * @brief Get the number of Windows Mixers available.
         *
         * The index from 0 to GetMixerCount() - 1 should be passed to the
         * WindowsMixer.* functions.
         * @see WindowsMixer.GetMixerName */
        public static int GetMixerCount()
        {
            return TT4.TT_Mixer_GetMixerCount();
        }
        /**
         * @brief Get the name of a Windows Mixer based on its name.
         *
         * @param nMixerIndex The index of the mixer. Ranging from 0 to 
         * GetMixerCount() - 1.
         * @param szMixerName The output string receiving the name of the device. */
        public static bool GetMixerName(int nMixerIndex,
            out string szMixerName)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            bool b = TT4.TT_Mixer_GetMixerName(nMixerIndex, ptr);
            szMixerName = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return b;
        }
        /**
         * @brief Get the name of the mixer associated with a wave-in device.
         * 
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param szMixerName The output string receiving the name of the device. 
         * @see TeamTalk4.GetSoundInputDevices */
        public static bool GetWaveInName(int nWaveDeviceID,
                                  out string szMixerName)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            bool b = TT4.TT_Mixer_GetWaveInName(nWaveDeviceID, ptr);
            szMixerName = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return b;
        }
        /**
         * @brief Get the name of the mixer associated with a wave-out device.
         * 
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param szMixerName The output string receiving the name of the device. 
         * @see TeamTalk4.GetSoundOutputDevices */
        public static bool GetWaveOutName(int nWaveDeviceID,
                                  out string szMixerName)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            bool b = TT4.TT_Mixer_GetWaveOutName(nWaveDeviceID, ptr);
            szMixerName = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return b;
        }

        /**
         * @brief Mute or unmute a Windows Mixer Wave-Out device from the
         * 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @param bMute True if device should be muted, False if it should be
         * unmuted.
         * @see WindowsMixer.GetWaveOutMute */
        public static bool SetWaveOutMute(int nWaveDeviceID, MixerControl nControl, bool bMute)
        {
            return TT4.TT_Mixer_SetWaveOutMute(nWaveDeviceID, nControl, bMute);
        }
        /**
         * @brief Get the mute state of a Windows Mixer Wave-Out device
         * from the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @return TRUE if mute, FALSE if unmuted, -1 on error.
         * @see WindowsMixer.SetWaveOutMute */
        public static int GetWaveOutMute(int nWaveDeviceID, MixerControl nControl)
        {
            return TT4.TT_Mixer_GetWaveOutMute(nWaveDeviceID, nControl);
        }
        /**
         * @brief Set the volume of a Windows Mixer Wave-Out device from
         * the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @param nVolume A value ranging from 0 to 65535. */
        public static bool SetWaveOutVolume(int nWaveDeviceID, MixerControl nControl, int nVolume)
        {
            return TT4.TT_Mixer_SetWaveOutVolume(nWaveDeviceID, nControl, nVolume);
        }
        /**
         * @brief Get the volume of a Windows Mixer Wave-Out device from
         * the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @return A value ranging from 0 to 65535, or -1 on error. */
        public static int GetWaveOutVolume(int nWaveDeviceID, MixerControl nControl)
        {
            return TT4.TT_Mixer_GetWaveOutVolume(nWaveDeviceID, nControl);
        }
        /**
         * @brief Set the selected state of a Windows Mixer Wave-In
         * device from the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control. */
        public static bool SetWaveInSelected(int nWaveDeviceID, MixerControl nControl)
        {
            return TT4.TT_Mixer_SetWaveInSelected(nWaveDeviceID, nControl);
        }
        /**
         * @brief Get the selected state of a Windows Mixer Wave-In device
         * from the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @return TRUE if mute, FALSE if unmuted, -1 on error. */
        public static int GetWaveInSelected(int nWaveDeviceID, MixerControl nControl)
        {
            return TT4.TT_Mixer_GetWaveInSelected(nWaveDeviceID, nControl);
        }
        /**
         * @brief Set the volume of a Windows Mixer Wave-In device from
         * the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @param nVolume A value ranging from 0 to 65535. */
        public static bool SetWaveInVolume(int nWaveDeviceID, MixerControl nControl, int nVolume)
        {
            return TT4.TT_Mixer_SetWaveInVolume(nWaveDeviceID, nControl, nVolume);
        }
        /**
         * @brief Get the volume of a Windows Mixer Wave-In device from
         * the 'enum' of devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControl A mixer control.
         * @return A value ranging from 0 to 65535, or -1 on error. */
        public static int GetWaveInVolume(int nWaveDeviceID, MixerControl nControl)
        {
            return TT4.TT_Mixer_GetWaveInVolume(nWaveDeviceID, nControl);
        }
        /**
         * @brief Enable and disable microphone boost.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param bEnable TRUE to enable, FALSE to disable. */
        public static bool SetWaveInBoost(int nWaveDeviceID, bool bEnable)
        {
            return TT4.TT_Mixer_SetWaveInBoost(nWaveDeviceID, bEnable);
        }
        /**
         * @brief See if microphone boost is enabled.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @return TRUE if boost is enabled, FALSE if disabled, -1 on error. */
        public static int GetWaveInBoost(int nWaveDeviceID)
        {
            return TT4.TT_Mixer_GetWaveInBoost(nWaveDeviceID);
        }
        /**
         * @brief Mute/unmute microphone input.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param bEnable TRUE to enable, FALSE to disable. */
        public static bool SetWaveInMute(int nWaveDeviceID, bool bEnable)
        {
            return TT4.TT_Mixer_SetWaveInMute(nWaveDeviceID, bEnable);
        }
        /**
         * @brief See if microphone is muted.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @return TRUE if mute is enabled, FALSE if disabled, -1 on error. */
        public static int GetWaveInMute(int nWaveDeviceID)
        {
            return TT4.TT_Mixer_GetWaveInMute(nWaveDeviceID);
        }
        /**
         * @brief Get the number of Windows Mixer Wave-In devices.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @see WindowsMixer.GetWaveInControlName
         * @see WindowsMixer.SetWaveInControlSelected
         * @see WindowsMixer.GetWaveInControlSelected */
        public static int GetWaveInControlCount(int nWaveDeviceID)
        {
            return TT4.TT_Mixer_GetWaveInControlCount(nWaveDeviceID);
        }
        /**
         * @brief Get the name of the Wave-In device with the specified
         * index.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControlIndex The index of the control. Randing from 0 to 
         * GetWaveInControlCount() - 1.
         * @param szDeviceName The output string of the name of the device.
         * @see WindowsMixer.GetWaveInControlCount */
        public static bool GetWaveInControlName(int nWaveDeviceID, int nControlIndex,
            out string szDeviceName)
        {
            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN * 2);
            bool b = TT4.TT_Mixer_GetWaveInControlName(nWaveDeviceID, nControlIndex, ptr);
            szDeviceName = Marshal.PtrToStringUni(ptr);
            Marshal.FreeHGlobal(ptr);
            return b;
        }
        /**
         * @brief Set the selected state of a Wave-In device in the
         * Windows Mixer.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControlIndex The index of the device. Randing from 0 to 
         * GetWaveInControlCount() - 1.
         * @see WindowsMixer.GetWaveInControlCount */
        public static bool SetWaveInControlSelected(int nWaveDeviceID, int nControlIndex)
        {
            return TT4.TT_Mixer_SetWaveInControlSelected(nWaveDeviceID, nControlIndex);
        }
        /**
         * @brief Get the selected state of a Wave-In device in the
         * Windows Mixer.
         *
         * @param nWaveDeviceID The @a nWaveDeviceID from the #BearWare.SoundDevice
         * struct.
         * @param nControlIndex The index of the device. Randing from 0 to  
         * GetWaveInControlCount() - 1.
         * @see WindowsMixer.GetWaveInControlCount */
        public static bool GetWaveInControlSelected(int nWaveDeviceID, int nControlIndex)
        {
            return TT4.TT_Mixer_GetWaveInControlSelected(nWaveDeviceID, nControlIndex);
        }
    }
    /** @} */

    /** @addtogroup firewall
     * @{ */

    /**
     * @brief A wrapper for the Windows Firewall API.
     */
    public class WindowsFirewall
    {
        /**
         * @brief Check if the Windows Firewall is currently enabled.
         *
         * This function does not invoke UAC on Windows Vista/7.
         * @see WindowsFirewall.Enable() */
        public static bool IsEnabled()
        {
            return TT4.TT_Firewall_IsEnabled();
        }
        /**
         * @brief Enable/disable the Windows Firewall.
         *
         * The Windows Firewall was introduced in Windows XP SP2.
         *
         * On Windows XP (SP2+) the user calling this function is assumed
         * to have administrator rights. On Windows Vista/7 UAC is invoked
         * to ask the user for administrator rights.
         * @see WindowsFirewall.IsEnabled() */
        public static bool Enable(bool bEnable)
        {
            return TT4.TT_Firewall_Enable(bEnable);
        }
        /**
         * @brief Check if an executable is already in the Windows
         * Firewall exception list.
         *
         * This function does not invoke UAC on Windows Vista/7.
         * @see WindowsFirewall.AddAppException() */
        public static bool AppExceptionExists(string szExecutable)
        {
            return TT4.TT_Firewall_AppExceptionExists(szExecutable);
        }
        /**
         * @brief Add an application to the Windows Firewall exception
         * list.
         *
         * On Windows XP (SP2+) the user calling this function is assumed
         * to have administrator rights. On Windows Vista/7 UAC is invoked
         * to ask the user for administrator rights.
         * @see WindowsFirewall.AppExceptionExists()
         * @see WindowsFirewall.RemoveAppException() */
        public static bool AddAppException(string szName, string szExecutable)
        {
            return TT4.TT_Firewall_AddAppException(szName, szExecutable);
        }
        /**
         * @brief Remove an application from the Windows Firewall exception
         * list.
         *
         * On Windows XP (SP2+) the user calling this function is assumed
         * to have administrator rights. On Windows Vista/7 UAC is invoked
         * to ask the user for administrator rights.
         * @see WindowsFirewall.AppExceptionExists()
         * @see WindowsFirewall.AddAppException() */
        public static bool RemoveAppException(string szExecutable)
        {
            return TT4.TT_Firewall_RemoveAppException(szExecutable);
        }
    }
    /** @} */

    /** ingroup desktopshare
     * @brief A struct which describes the properties of a window
     * which can be shared. */
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public struct ShareWindow
    {
        /** @brief The Windows handle of the window. */
        public IntPtr hWnd;
        /** @brief X coordinate of the window relative to the Windows desktop. */
        public int nWndX;
        /** @brief Y coordinate of the window relative to the Windows desktop. */
        public int nWndY;
        /** @brief The width in pixels of the window. */
        public int nWidth;
        /** @brief The height in pixels of the window. */
        public int nHeight;
        /** @brief The title of the window. */
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = TeamTalk4.TT_STRLEN)]
        public string szWindowTitle;
    }

    /** ingroup desktopshare
     * @brief Helper class for extracting Windows HWND handles and
     * controlling mouse and keyboard. */
    public class WindowsHelper
    {
        /** @brief Get the handle (HWND) of the window which is currently
         * active (focused) on the Windows desktop. */
        public static System.IntPtr GetDesktopActiveHWND()
        {
            return TT4.TT_Windows_GetDesktopActiveHWND();
        }

        /** @brief Get the handle (HWND) of the Windows desktop (full desktop). */
        public static System.IntPtr GetDesktopHWND()
        {
            return TT4.TT_Windows_GetDesktopHWND();
        }

        /** @brief Get the all the handles (HWND) of visible windows. */
        public static bool GetDesktopWindowHWND(int nIndex,
                                                ref System.IntPtr lpHWnd)
        {
            return TT4.TT_Windows_GetDesktopWindowHWND(nIndex, ref lpHWnd);
        }

        /**
         * @brief Get the properties of a window from its window handle (HWND). */
        public static bool GetWindow(System.IntPtr hWnd,
                                     ref ShareWindow lpShareWindow)
        {
            return TT4.TT_Windows_GetWindow(hWnd, ref lpShareWindow);
        }

        /**
         * @brief Translate platform key-code to and from TeamTalk's
         * intermediate format.
         *
         * Section @ref keytranslate has a table which shows how the keys on a US
         * 104-keyboard are translated to TeamTalk's intermediate format.
         *
         * Section @ref transdesktopinput explains how to transmit key-codes.
         *
         * @param nTranslate The key-code format to translate to and from.
         * @param lpDesktopInputs An array of #BearWare.DesktopInput structs to translate.
         * @param lpTranslatedDesktopInputs A pre-allocated array of #BearWare.DesktopInput
         * struct to hold the translated desktop input.
         * @return The number of translated #BearWare.DesktopInput stucts. If value
         * is different from @c nDesktopInputCount then some @c uKeyCode
         * values could not be translated and have been assigned the value
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_IGNORE.
         * @see TeamTalk4.SendDesktopInput()
         * @see WindowsHelper.Execute() */
        public static int KeyTranslate(TTKeyTranslate nTranslate,
                                       DesktopInput[] lpDesktopInputs,
                                       out DesktopInput[] lpTranslatedDesktopInputs)
        {
            lpTranslatedDesktopInputs = new DesktopInput[lpDesktopInputs.Length];
            return TT4.TT_DesktopInput_KeyTranslate(nTranslate, lpDesktopInputs,
                                                    lpTranslatedDesktopInputs,
                                                    lpDesktopInputs.Length);
        }

        /**
         * @brief Execute desktop (mouse or keyboard) input.
         *
         * When executed either a key-press, key-release or mouse move
         * will take place on the computer running the client
         * instance. Remember to calculate the offsets for the mouse
         * cursor prior to this call. The mouse position will be relative
         * to the screen resolution.
         *
         * The content of the #BearWare.DesktopInput struct must been translated to
         * the platform's key-code format prior to this
         * call. I.e. uKeyCode must be a either a Windows scan-code, Mac
         * OS X Carbon key-code or one of the mouse buttons:
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_LMOUSEBTN,
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_RMOUSEBTN,
         * #BearWare.TeamTalk4.TT_DESKTOPINPUT_KEYCODE_MMOUSEBTN.
         *
         * @param lpDesktopInputs The mouse or keyboard inputs.
         * @see WindowsHelper.KeyTranslate() */
        public static bool Execute(DesktopInput[] lpDesktopInputs)
        {
            return TT4.TT_DesktopInput_Execute(lpDesktopInputs, lpDesktopInputs.Length);
        }

    }

    /*
    class Class1
    {
        [STAThread]
        static void Main(string[] args)
        {
            TeamTalk4 tt = new TeamTalk4(true);
            bool ret;
            int len, id;
            ret = c_tt.TT4.TT_SetLicenseInformation("", 0);
            IntPtr ttInst = c_tt.TT4.TT_InitTeamTalkPoll();
            ClientFlag flag = c_tt.TT4.TT_GetFlags(ttInst);

            int devin, devout, devcount;
            ret = c_tt.TT4.TT_GetDefaultSoundDevices(ttInst, out devin, out devout);
            SoundDevice[] devs = null;
            ret = c_tt.TT4.TT_GetSoundInputDevices(ttInst, devs, out devcount);
            devs = new SoundDevice[devcount];
            ret = c_tt.TT4.TT_GetSoundInputDevices(ttInst, devs, out devcount);

            ret = c_tt.TT4.TT_GetSoundOutputDevices(ttInst, devs, out devcount);
            devs = new SoundDevice[devcount];
            ret = c_tt.TT4.TT_GetSoundOutputDevices(ttInst, devs, out devcount);

            ret = c_tt.TT4.TT_InitSoundInputDevice(ttInst, devin);
            flag = c_tt.TT4.TT_GetFlags(ttInst);
            ret = c_tt.TT4.TT_InitSoundOutputDevice(ttInst, devout);
            flag = c_tt.TT4.TT_GetFlags(ttInst);
            ret = c_tt.TT4.TT_EnableVoiceActivation(ttInst, true);

            VideoCaptureDevice[] viddevs = null;
            ret = c_tt.TT4.TT_GetVideoCaptureDevices(ttInst, viddevs, out devcount);
            viddevs = new VideoCaptureDevice[devcount];
            len = Marshal.SizeOf(viddevs[0]);
            ret = c_tt.TT4.TT_GetVideoCaptureDevices(ttInst, viddevs, out devcount);
            VideoCodec vidcodec = new VideoCodec();
            vidcodec.nCodec = Codec.THEORA_CODEC;
            vidcodec.theora.nQuality = 10;
            vidcodec.theora.nTargetBitRate = 0;
            ret = c_tt.TT4.TT_InitVideoCaptureDevice(ttInst, viddevs[0].szDeviceID, 
                ref viddevs[0].captureFormats[0], ref vidcodec);
            flag = c_tt.TT4.TT_GetFlags(ttInst);
            ret = c_tt.TT4.TT_Connect(ttInst, "teamtalk.dyndns.dk", 10333, 10333, 0, 0);

            c_tt.TTMessage ttmsg;
            int maxms = 5000;
            ret = c_tt.TT4.TT_GetMessage(ttInst, out ttmsg, out maxms);
            while((c_tt.TT4.TT_GetFlags(ttInst) & ClientFlag ::CLIENT_CONNECTED) != ClientFlag ::CLIENT_CONNECTED)
                Thread.Sleep(100);

            ClientStatistics stats;
            c_tt.TT4.TT_GetStatistics(ttInst, out stats);

            AudioCodec audcodec = new AudioCodec();
            audcodec.nCodec = Codec.SPEEX_CODEC;
            audcodec.speex.bStereoPlayback = false;
            audcodec.speex.bUseJitterBuffer = false;
            audcodec.speex.nBandmode = 0;
            audcodec.speex.nMSecPerPacket = 45;
            audcodec.speex.nQuality = 10;
            len = c_tt.TT4.TT_GetPacketSize(ref audcodec);

            id = c_tt.TT4.TT_DoLogin(ttInst, "BBB", "", "admin", "pillow");

            ret = c_tt.TT4.TT_GetMessage(ttInst, out ttmsg, out maxms);

            while ((c_tt.TT4.TT_GetFlags(ttInst) & ClientFlag ::CLIENT_AUTHORIZED) != ClientFlag ::CLIENT_AUTHORIZED)
                Thread.Sleep(100);
            flag = c_tt.TT4.TT_GetFlags(ttInst);

            //TextMessage txtmsg = new TextMessage();
            //txtmsg.nMsgType = TextMsgType.MSGTYPE_USER;
            //txtmsg.nToUserID = 15160;
            //txtmsg.szMessage = "Heoollle";
            //c_tt.TT4.TT_DoTextMessage(ttInst, ref txtmsg);

            IntPtr ptr = Marshal.AllocHGlobal(TeamTalk4.TT_STRLEN*2);
            c_tt.TT4.TT_GetErrorMessage(1000, ptr);
            string s = Marshal.PtrToStringUni(ptr);

            c_tt.TT4.TT_GetChannelPath(ttInst, 1, ptr);
            s = Marshal.PtrToStringUni(ptr);
            c_tt.TT4.TT_Mixer_GetMixerName(0, ptr);
            s = Marshal.PtrToStringUni(ptr);

            Marshal.FreeHGlobal(ptr);
            

            int[] keys = {163};
            c_tt.TT4.TT_HotKey_Register(ttInst, 1, keys, 1);

            c_tt.TT4.TT_HotKey_Unregister(ttInst, 1);

            ret = c_tt.TT4.TT_CloseTeamTalk(ttInst);
        }
    }
    */
}
